#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "guiutils.h"

#include "about_dlg.h"
#include "me_types.h"
#include "manedit.h"
#include "config.h"

#include "images/icon_ok_20x20.xpm"
#include "images/icon_manedit_editor_48x48.xpm"   
#include "images/manedit_banner_400x150.xpm"


static void AboutDialogOKCB(GtkWidget *widget, gpointer data);
static gint AboutDialogDeleteEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
);

about_dialog_struct *AboutDialogNew(me_core_struct *core);
void AboutDialogMap(about_dialog_struct *ad);
void AboutDialogUnmap(about_dialog_struct *ad);
void AboutDialogDelete(about_dialog_struct *ad);


/*
 *	Ok callback.
 */
static void AboutDialogOKCB(GtkWidget *widget, gpointer data)
{
	about_dialog_struct *ad = ABOUT_DIALOG(data);
	if(ad == NULL)
	    return;

	AboutDialogUnmap(ad);
}

/*
 *	"delete_event" signal callback.
 */
static gint AboutDialogDeleteEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
)
{
	AboutDialogOKCB(widget, data);
	return(TRUE);
}


/*
 *	Creates a new About Dialog.
 */
about_dialog_struct *AboutDialogNew(me_core_struct *core)
{
	const gint border_minor = 2;
	gchar *s;
	GdkFont *font;
	GdkWindow *window;
	GtkStyle *style;
	GtkAccelGroup *accelgrp;
	GtkWidget	*w,
			*parent, *parent2, *parent3,
			*toplevel;
	about_dialog_struct *ad;

	if(core == NULL)
	    return(NULL);

	ad = ABOUT_DIALOG(g_malloc0(
	    sizeof(about_dialog_struct)
	));
	if(ad == NULL)
	    return(NULL);

	ad->accelgrp = accelgrp = gtk_accel_group_new();
	ad->toplevel = toplevel = gtk_window_new(GTK_WINDOW_DIALOG);

	ad->freeze_count++;

	/* Toplevel GtkWindow */
	w = toplevel;
	gtk_widget_realize(w);
	window = w->window;
	if(window != NULL)
	{
	    gdk_window_set_decorations(
		window,
		GDK_DECOR_BORDER | GDK_DECOR_TITLE | GDK_DECOR_MENU |
		GDK_DECOR_MINIMIZE
	    );
	    gdk_window_set_functions(
		window,
		GDK_FUNC_MOVE | GDK_FUNC_MINIMIZE | GDK_FUNC_CLOSE
	    );
	    GUISetWMIcon(
		window,
		(guint8 **)icon_manedit_editor_48x48_xpm
	    );
	}
	style = gtk_widget_get_style(w);
	gtk_signal_connect(
	    GTK_OBJECT(w), "delete_event",
	    GTK_SIGNAL_FUNC(AboutDialogDeleteEventCB), ad
	);
	gtk_window_add_accel_group(GTK_WINDOW(w), accelgrp);
	parent = w;


	/* Main GtkVBox */
	w = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(parent), w);
	gtk_widget_show(w);
	parent = w;


	/* Hbox to hold banners and labels */
	w = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(parent), w, TRUE, TRUE, 5);
	gtk_widget_show(w);
	parent2 = w;

	/* Vbox inside of hbox for alignment */
	w = gtk_vbox_new(FALSE, 5);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, TRUE, 5);
	gtk_widget_show(w);
	parent2 = w;

	/* Heading image */
	w = gtk_alignment_new(
	    0.5f, 0.5f,
	    1.0f, 1.0f
	);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	w = gtk_pixmap_new_from_xpm_d(
	    window,
	    NULL,
	    (guint8 **)manedit_banner_400x150_xpm
	);
	gtk_container_add(GTK_CONTAINER(parent3), w);
	gtk_widget_show(w);

	/* Version banner */
	s = g_strdup_printf(
	    "Version %s",
	    PROG_VERSION
	);
	font = gdk_font_load(MEDIT_FONT_NAME_HEADING3_REV);
        if(font == NULL)
            font = GDK_FONT_REF(style->font);
	w = GUIBannerCreate(
	    s,
	    font,
            &style->bg[GTK_STATE_NORMAL],
            &style->text[GTK_STATE_NORMAL],
            &style->fg[GTK_STATE_NORMAL],
	    GTK_JUSTIFY_CENTER,
            TRUE,				/* Expand */
            border_minor,			/* Inside border */
            border_minor			/* Outside border */
	);
	g_free(s);
	GDK_FONT_UNREF(font);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	/* Copyright label */
	w = gtk_label_new(PROG_COPYRIGHT);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_LEFT);
	gtk_widget_show(w);


	/* Horizontal rule */
	w = gtk_hseparator_new();
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	/* Hbox to hold buttons homogeniously */
	w = gtk_hbox_new(TRUE, 0);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 5);
	gtk_widget_show(w);
	parent2 = w;

	/* OK GtkButton */
	ad->ok_btn = w = GUIButtonPixmapLabelH(
	    (guint8 **)icon_ok_20x20_xpm,
	    "OK",
	    NULL
	);
	gtk_widget_set_usize(
	    w,
	    GUI_BUTTON_HLABEL_WIDTH_DEF, GUI_BUTTON_HLABEL_HEIGHT_DEF
	);
	GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(AboutDialogOKCB), ad
	);
        gtk_accel_group_add(
            accelgrp,
            GDK_o, GDK_CONTROL_MASK,
            GTK_ACCEL_VISIBLE,
            GTK_OBJECT(w),
            "clicked"
        );
        GUIButtonLabelUnderline(w, GDK_o);
	gtk_widget_show(w);

	ad->freeze_count--;

	return(ad);
}


/*
 *	Maps the About Dialog.
 */
void AboutDialogMap(about_dialog_struct *ad)
{
	GtkWidget *w;

	if(ad == NULL)
	    return;

	w = ad->ok_btn;
	if(w != NULL)
	{
	    gtk_widget_grab_focus(w);
	    gtk_widget_grab_default(w);
	}

	gtk_widget_show(ad->toplevel);
}

/*
 *	Unmaps the About Dialog.
 */
void AboutDialogUnmap(about_dialog_struct *ad)
{
	if(ad == NULL)
	    return;

	gtk_widget_hide(ad->toplevel);
}

/*
 *	Deletes the About Dialog.
 */
void AboutDialogDelete(about_dialog_struct *ad)
{
	if(ad == NULL)
	    return;

	ad->freeze_count++;

	GTK_WIDGET_DESTROY(ad->toplevel);
	GTK_ACCEL_GROUP_UNREF(ad->accelgrp);

	ad->freeze_count--;

	g_free(ad);
}
