#ifndef CA_STRING_H
#define CA_STRING_H

#include <glib.h>


/*
 *	Character Attribute String:
 *
 *	A 2 bytes per character null-terminated string type, the first
 *	byte specifies the gchar character value and the second byte
 *	specifies the CAStringAttributes attribute.
 *
 *	The end of the string is denoted with the null-terminating
 *	gchar character value '\0'.
 */
#define CAString			guint16


#define CA_STRING_TARGET_TYPE		"CAString"


/*
 *	Get/Set Macros:
 */
#define CA_STRING_CHARACTER(_cas_)	((gchar)(	\
	(*((CAString *)(_cas_)) & 0x00FF)		\
))
#define CA_STRING_SET_CHARACTER(_cas_,_c_)		\
	(((guint8 *)(_cas_))[0]) = (guint8)(_c_)


#define CA_STRING_ATTRIBUTE(_cas_)	((CAStringAttributes)(	\
	((*((CAString *)(_cas_)) & 0xFF00) >> 8)		\
))
#define CA_STRING_SET_ATTRIBUTE(_cas_,_c_)		\
	(((guint8 *)(_cas_))[1]) = (guint8)(_c_)
#define CA_STRING_ADD_ATTRIBUTE(_cas_,_a_)		\
	(((guint8 *)(_cas_))[1]) |= (guint8)(_a_)
#define CA_STRING_REMOVE_ATTRIBUTE(_cas_,_a_)		\
	(((guint8 *)(_cas_))[1]) &= ~((guint8)(_a_))


/*
 *	Special Characters:
 */
#define CA_STRING_CHARACTER_LINE_END	0x0A	/* Use Newline */
#define CA_STRING_CHARACTER_LINE_BREAK	0x0D	/* Use Carrage Return */


/*
 *	CAStringAttributes:
 *
 *	Flags for the second byte.
 */
typedef enum {
	CA_STRING_ATTRIBUTE_NORMAL	= 0,		/* No attributes */
	CA_STRING_ATTRIBUTE_SELECTED	= (1 << 0),
	CA_STRING_ATTRIBUTE_BOLD	= (1 << 1),
	CA_STRING_ATTRIBUTE_UNDERLINE	= (1 << 2),
	CA_STRING_ATTRIBUTE_HEADING	= (1 << 3)
} CAStringAttributes;


/* New, Copy, Delete */
extern CAString *CAStringNewFromString(
	const gchar *s,
	const gint len
);
extern CAString *CAStringCopy(const CAString *ca_string);
extern void CAStringDelete(CAString *ca_string);

/* Convert To Regular Single-Byte String */
extern gchar *CAStringToRegularString(
	const CAString *ca_string,
	const gint len
);

/* Measuring */
extern gint CAStringLengthPointers(
	const CAString *ca_string1_ptr,
	const CAString *ca_string2_ptr
);
extern gint CAStringLength(const CAString *ca_string);
extern gboolean CAStringIsEmpty(const CAString *ca_string);

/* Seek */
extern CAString *CAStringSeekToIndex(
	CAString *ca_string,
	const gint i
);
extern CAString *CAStringSeekToCharacter(
	CAString *ca_string,
	const gchar c
);
extern CAString *CAStringSeekToNextCharacter(
	CAString *ca_string,
	const gchar c
);
extern CAString *CAStringSeekToNextCharacters(
        CAString *ca_string,
        const gchar *characters_list,
        const gint ncharacters
);
extern CAString *CAStringSeekToNextAttribute(CAString *ca_string);

extern CAString *CAStringNextLine(CAString *ca_string);
extern CAString *CAStringPreviousLine(
	CAString *ca_string,
	CAString *ca_string_i
);
extern CAString *CAStringCurrentLineStart(
	CAString *ca_string,
	CAString *ca_string_i
);

/* Add/Remove Attribute */
extern void CAStringSegmentAddAttribute(
	CAString *ca_string,
	const gint start_i, const gint end_i,
	const CAStringAttributes attr
);
extern void CAStringSegmentRemoveAttribute(
	CAString *ca_string,
	const gint start_i, const gint end_i,
	const CAStringAttributes attr
);

/* Select/Unselect */
extern void CAStringSelect(
	CAString *ca_string,
	const gint selection_start_pos,
	const gint selection_end_pos,
	const gboolean clear_existing_selection
);
extern void CAStringUnselectAll(CAString *ca_string);

/* Insert/Remove Character */
extern CAString *CAStringInsertCharacter(
	CAString *ca_string,
	const gint i,
	const gchar c,
	const CAStringAttributes attr
);
extern CAString *CAStringAppendCharacter(
        CAString *ca_string,
        const gchar c,
        const CAStringAttributes attr
);
extern CAString *CAStringInsertText(
	CAString *ca_string,
	const gint i,
	const CAString *ca_string2
);
extern CAString *CAStringRemoveCharacter(
	CAString *ca_string,
	const gint i
);
extern CAString *CAStringRemoveText(
	CAString *ca_string,
	const gint i,
	const gint len
);

/* Remove End Lines */
extern void CAStringClearLineEnds(CAString *ca_string);

/* Remove Spaces */
extern CAString *CAStringChug(CAString *ca_string);
extern CAString *CAStringChomp(CAString *ca_string);
extern CAString *CAStringStrip(CAString *ca_string);


#endif	/* CA_STRING_H */
