/*
				ManPage Editor
 */

#ifndef EDITOR_H
#define EDITOR_H

#include <gtk/gtk.h>
#include "me_types.h"


typedef struct _editor_struct		editor_struct;
#define EDITOR(p)			((editor_struct *)(p))

typedef struct _editor_item_struct	editor_item_struct;
#define EDITOR_ITEM(p)			((editor_item_struct *)(p))

typedef struct _editor_fip_struct	editor_fip_struct;
#define EDITOR_FIP(p)			((editor_fip_struct *)(p))


/*
 *	Operation IDs:
 */
typedef enum {
	EDITOR_OPID_NEW,
	EDITOR_OPID_OPEN,
	EDITOR_OPID_SAVE,
	EDITOR_OPID_SAVE_AS,
	EDITOR_OPID_CLOSE_MANUAL,
	EDITOR_OPID_UNDO,
	EDITOR_OPID_REDO,
	EDITOR_OPID_CUT,
	EDITOR_OPID_COPY,
	EDITOR_OPID_PASTE,
	EDITOR_OPID_NEW_HEADER,
	EDITOR_OPID_NEW_SECTION,
	EDITOR_OPID_REMOVE,			/* Remove header or section */
	EDITOR_OPID_PREVIEW
} editor_opid;

/*
 *	Editor Item Types:
 */
typedef enum {
	EDITOR_ITEM_FILE,			/* Toplevel node */
	EDITOR_ITEM_HEADER,
	EDITOR_ITEM_SECTION
} editor_item_type;


/*
 *	Editor:
 */
struct _editor_struct {

	GtkWidget	*toplevel;
	GtkAccelGroup	*accelgrp;
	gint		freeze_count,
			busy_count;
	gboolean	syntax_highlighting;
	GdkColormap	*colormap;

	me_core_struct	*core;
	gint		viewer_num;

	GtkWidget	*main_vbox,
			*menu_bar_dock,
			*menu_bar,
			*tool_bar_dock,
			*tool_bar,
			*main_paned,
			*layout_ctree;

	/* GdkColors */
	GdkColor	normal_text_color,
			deliminator_text_color,
			tag_text_color,
			symbol_text_color;

	/* GList of GtkWidget * toplevel widgets that contain widgets
	 * to edit each EditorItemType
	 */
	GList		*contents_toplevels_list;

	GtkWidget	*header_name_entry,
			*header_section_number_entry,
			*header_version_entry,
			*header_author_entry,
			*header_catagory_entry,
			*header_comments_text;

	GtkWidget	*section_name_entry,
			*section_text;

	/* Syntax highlight timeout callback values */
	guint		synhl_timeout_cb_id;	/* None if (guint)(-1) */
	gint		synhl_timeout_cb_cursor_pos;
	gint		synhl_timeout_cb_start_pos;
	gint		synhl_timeout_cb_end_pos;

	/* Important menu items */
	GtkWidget	*file_mh,
			*new_mi,
			*new_from_template_mi,
			*open_mi,
			*save_mi,
			*save_as_mi,
			*revert_mi,
			*close_manual_mi,
			*close_mi,
			*exit_mi,

			*edit_mh,
			*undo_mi,
			*undo_milabel,
			*redo_mi,
			*redo_milabel,
			*cut_mi,
			*copy_mi,
			*paste_mi,
			*select_all_mi,
			*unselect_all_mi,
			*find_in_pages_mi,

			*view_mh,
			*preview_mi,
			*syntax_highlight_cmi,

			*layout_mh,
			*add_header_mi,
			*add_section_mi,
			*remove_mi,
			*properties_mi,

			*fmt_mh,
			*fmt_pl_mi,
			*fmt_pr_mi,
			*fmt_pi_mi,
			*fmt_li_mi,
			*fmt_b_mi,
			*fmt_u_mi,
			*fmt_br_mi,
			*fmt_amp_mi,
			*fmt_lt_mi,
			*fmt_gt_mi,
			*fmt_strip_tags_mi,

			*settings_mh,
			*settings_options_mi,

			*windows_mh,
			*windows_new_editor,
			*windows_new_viewer;

	/* Formatting Bar Buttons */
	GtkWidget	*fmt_bar_dock,
			*fmt_bar,
			*fmt_pl_btn,
			*fmt_pr_btn,
			*fmt_pi_btn,
			*fmt_li_btn,
			*fmt_b_btn,
			*fmt_u_btn,
			*fmt_br_btn,
			*fmt_amp_btn,
			*fmt_lt_btn,
			*fmt_gt_btn,
			*fmt_strip_tags_btn;

	/* Search Bar */
	GtkWidget	*find_bar_dock,
			*find_bar,
			*find_combo,
			*find_btn,
			*replace_combo,
			*replace_btn,
			*replace_all_btn;

	/* Find bar button find menu */
	GtkWidget	*find_bar_find_menu,
			*find_bar_find_in_pages_mi;

	/* Find bar button replace menu */
	GtkWidget       *find_bar_replace_menu,
			*find_bar_replace_entire_page_mi;

	/* Right-click menus */
	GtkWidget	*layout_menu,
			*edit_panel_menu;

	/* Important menu items on right-click menus */
	GtkWidget	*layout_menu_expand_mi,
			*layout_menu_preview_mi,
			*layout_menu_add_header_mi,
			*layout_menu_add_section_mi,
			*layout_menu_remove_mi,
			*layout_menu_properties_mi,
			*layout_menu_new_manual_mi,
			*layout_menu_new_from_template_mi,
			*layout_menu_open_mi,
			*layout_menu_save_mi,
			*layout_menu_save_as_mi,
			*layout_menu_revert_mi,
			*layout_menu_close_manual_mi;

	GtkWidget	*edit_panel_undo_mi,
			*edit_panel_undo_milabel,
			*edit_panel_redo_mi,
			*edit_panel_redo_milabel,
			*edit_panel_cut_mi,
			*edit_panel_copy_mi,
			*edit_panel_paste_mi,
			*edit_panel_select_all_mi,
			*edit_panel_unselect_all_mi;
	/* Format submenu on edit panel menu */
	GtkWidget	*edit_panel_fmt_sub_menu,
			*edit_panel_fmt_pl_mi,
			*edit_panel_fmt_pr_mi,
			*edit_panel_fmt_pi_mi,
			*edit_panel_fmt_li_mi,
			*edit_panel_fmt_b_mi,
			*edit_panel_fmt_u_mi,
			*edit_panel_fmt_br_mi,
			*edit_panel_fmt_amp_mi,
			*edit_panel_fmt_lt_mi,
			*edit_panel_fmt_gt_mi,
			*edit_panel_fmt_strip_tags_mi;

	/* Status bar */
	GtkWidget	*status_bar_dock,
			*status_bar_toplevel,	/* An hbox */
			*status_bar_progress,
			*status_bar_label,
			*status_bar_cursor_label;

	/* Last progress position value from 0.0 to 1.0 */
	gfloat		status_bar_progress_pos_last;

	/* Undo/redo structures */
	gpointer	*undo;
	gint		total_undos;

	gpointer	*redo;
	gint		total_redos;

	gint		max_undos;	/* Maximum undos and maximum redos */


	/* Last open file path */
	gchar		*last_open_path;

	/* Last save as file path */
	gchar		*last_save_as_path;


	/* Editor Find In Pages dialog */
	editor_fip_struct *fip_dlg;

};

#define EDITOR_TOPLEVEL(p)		(((p) != NULL) ? \
 (EDITOR(p)->toplevel) : NULL)
#define EDITOR_LAYOUT_CTREE(p)		(((p) != NULL) ? \
 (EDITOR(p)->layout_ctree) : NULL)
#define EDITOR_HEADER_TEXT(p)		(((p) != NULL) ? \
 (EDITOR(p)->header_text) : NULL)
#define EDITOR_SECTION_TEXT(p)		(((p) != NULL) ? \
 (EDITOR(p)->section_text) : NULL)
#define EDITOR_LAST_OPEN_PATH(p)	(((p) != NULL) ? \
 (EDITOR(p)->last_open_path) : NULL)
#define EDITOR_LAST_SAVE_AS_PATH(p)	(((p) != NULL) ? \
 (EDITOR(p)->last_save_as_path) : NULL)


/*
 *	EditorItem:
 */
struct _editor_item_struct {

	editor_item_type	type;
	editor_struct	*editor;
	GtkCTreeNode	*node;			/* GtkCTreeNode that this is on */

	gboolean	has_changes;

	/* Members for type EDITOR_ITEM_FILE */
	gchar		*name,			/* Name of manual page */
			*path;			/* Full path to manual page
						 * or NULL if not saved yet */

	/* Members for type EDITOR_ITEM_HEADER */
	gchar		*header_name,
			*header_section_number,
			*header_version,
			*header_author,
			*header_catagory;

	/* Members for type EDITOR_ITEM_SECTION */
	gchar		*section_name;

	/* Lines List */
	gchar		**lines_list;
	gint		nlines;

};


/* EditorItem */
extern editor_item_struct *EditorItemNew(void);
extern editor_item_struct *EditorItemNewWithValues(
	editor_item_type type,
	editor_struct *editor,
	GtkCTreeNode *node
);
extern void EditorItemSetLines(
	editor_item_struct *item,
	gchar **lines_list, const gint nlines
);
extern editor_item_struct *EditorItemCopy(editor_item_struct *item);
extern void EditorItemDelete(editor_item_struct *item);

/* Select */
extern GtkCTreeNode *EditorCurrentSelectedNode(editor_struct *editor);
extern editor_item_struct *EditorCurrentSelectedItem(editor_struct *editor);
extern void EditorSelect(
	editor_struct *editor,
	GtkCTreeNode *node
);

/* Set/Get EditorItem */
extern void EditorBranchSetData(
	GtkCTree *ctree,
	GtkCTreeNode *node,
	editor_item_struct *item, GtkDestroyNotify destroy
);
extern editor_item_struct *EditorBranchGetData(
	GtkCTree *ctree,
	GtkCTreeNode *node
);
extern void EditorSetNodeHasChanges(
	editor_struct *editor,
	GtkCTreeNode *node,
	const gboolean has_changes
);

/* Get Node */
extern GtkCTreeNode *EditorGetFirstToplevelNode(editor_struct *editor);
extern GtkCTreeNode *EditorGetToplevelNode(GtkCTreeNode *node);

/* New, Open, Save As */
extern gint EditorOpenQuery(
	editor_struct *editor,
	GtkCTreeNode *insert_node,
	const gboolean verbose
);
extern gint EditorNewTemplateQuery(
	editor_struct *editor,
	GtkCTreeNode *insert_node,
	const gboolean verbose
);
extern gint EditorSaveAsQuery(
	editor_struct *editor,
	GtkCTreeNode *node,
	const gboolean verbose
);

/* Editor */
extern editor_struct *EditorNew(me_core_struct *core);
extern void EditorRecordConfiguration(editor_struct *editor);
extern void EditorUpdateMenus(editor_struct *editor);
extern void EditorSetStatusPosition(
	editor_struct *editor,
	const gint row, const gint column
);
extern void EditorSetStatusMessage(
	editor_struct *editor,
	const gchar *msg
);
extern void EditorSetStatusProgress(
	editor_struct *editor,
	const gfloat v
);
extern void EditorClearUndosList(editor_struct *editor);
extern void EditorSetBusy(
	editor_struct *editor,
	const gboolean busy
);
extern gboolean EditorIsMapped(editor_struct *editor);
extern void EditorMap(editor_struct *editor);
extern void EditorUnmap(editor_struct *editor);
extern void EditorDelete(editor_struct *editor);


#endif	/* EDITOR_H */
