#include <string.h>
#include <errno.h>
#include <sys/stat.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "../include/string.h"
#include "../include/disk.h"

#include "guiutils.h"
#include "tool_bar.h"
#include "cfg_win.h"
#include "cdialog.h"

#include "editor.h"
#include "editor_cb.h"
#include "editor_values.h"
#include "editor_syntax_highlight.h"
#include "editor_dnd.h"
#include "editor_fio.h"
#include "editor_preview.h"
#include "editor_find_in_pages.h"
#include "editor_find_replace.h"
#include "editor_undo.h"
#include "options_win.h"
#include "manedit.h"
#include "me_emit.h"
#include "me_op.h"

#include "me_cfg_list.h"
#include "config.h"


void EditorItemDestroyCB(gpointer data);

gint EditorKeyEventCB(GtkWidget *widget, GdkEventKey *key, gpointer data);
gint EditorTextButtonEventCB(
	GtkWidget *widget, GdkEventButton *button, gpointer data
);

void EditorToolBarItemActivateCB(
	ToolBarItem *item, gint id, gpointer data
);

gint EditorDeleteEventCB(GtkWidget *widget, GdkEvent *event, gpointer data);
void EditorCloseCB(GtkWidget *widget, gpointer data);
void EditorExitCB(GtkWidget *widget, gpointer data);

void EditorManualNewCB(GtkWidget *widget, gpointer data);
void EditorManualOpenCB(GtkWidget *widget, gpointer data);
void EditorManualNewFromTemplateCB(GtkWidget *widget, gpointer data);
void EditorManualSaveCB(GtkWidget *widget, gpointer data);
void EditorManualSaveAsCB(GtkWidget *widget, gpointer data);
void EditorManualRevertCB(GtkWidget *widget, gpointer data);
void EditorManualCloseCB(GtkWidget *widget, gpointer data);

void EditorManualAddHeaderCB(GtkWidget *widget, gpointer data);
void EditorManualAddSectionCB(GtkWidget *widget, gpointer data);
void EditorManualRemoveSectionCB(GtkWidget *widget, gpointer data);
void EditorManualPropertiesCB(GtkWidget *widget, gpointer data);
void EditorManualPreviewCB(GtkWidget *widget, gpointer data);
void EditorManualSyntaxHighlightToggleCB(GtkWidget *widget, gpointer data);

void EditorUndoCB(GtkWidget *widget, gpointer data);
void EditorRedoCB(GtkWidget *widget, gpointer data);
void EditorCutCB(GtkWidget *widget, gpointer data);
void EditorCopyCB(GtkWidget *widget, gpointer data);
void EditorPasteCB(GtkWidget *widget, gpointer data);
void EditorDeleteCB(GtkWidget *widget, gpointer data);
void EditorSelectAllCB(GtkWidget *widget, gpointer data);
void EditorUnselectAllCB(GtkWidget *widget, gpointer data);
void EditorOptionsCB(GtkWidget *widget, gpointer data);

void EditorFmtParagraphLeftCB(GtkWidget *widget, gpointer data);
void EditorFmtParagraphRightCB(GtkWidget *widget, gpointer data);
void EditorFmtParagraphIndentedCB(GtkWidget *widget, gpointer data);
void EditorFmtListItemCB(GtkWidget *widget, gpointer data);
void EditorFmtBoldCB(GtkWidget *widget, gpointer data);
void EditorFmtUnderlineCB(GtkWidget *widget, gpointer data);
void EditorFmtLineBreakCB(GtkWidget *widget, gpointer data);
void EditorFmtAmpersandCB(GtkWidget *widget, gpointer data);
void EditorFmtLessThanCB(GtkWidget *widget, gpointer data);
void EditorFmtGreaterThanCB(GtkWidget *widget, gpointer data);
void EditorStripTagsCB(GtkWidget *widget, gpointer data);

void EditorMapFIPCB(GtkWidget *widget, gpointer data);

void EditorFindComboActivateCB(GtkWidget *widget, gpointer data);
void EditorReplaceAllCB(GtkWidget *widget, gpointer data);
void EditorReplaceEntirePageCB(GtkWidget *widget, gpointer data);
static gint EditorReplaceEntirePageCBIterate(
	editor_struct *editor,
	GtkCTreeNode *sel_node,
	gchar *needle,
	const gchar *replacement,
	const gboolean case_sensitive
);
void EditorReplaceComboActivateCB(GtkWidget *widget, gpointer data);

static void EditorFindCBNexus(
	editor_struct *editor,
	const gboolean case_sensitive
);
static void EditorReplaceCBNexus(
	editor_struct *editor,
	const gboolean case_sensitive
);

gint EditorMenuMapCB(GtkWidget *widget, GdkEvent *event, gpointer data);

void EditorExpandCB(GtkWidget *widget, gpointer data);
void EditorLayoutCTreeSelectCB(
	GtkCTree *ctree, GtkCTreeNode *node, gint column,
	gpointer data
);
void EditorLayoutCTreeUnselectCB(
	GtkCTree *ctree, GtkCTreeNode *node, gint column,
	gpointer data
);
void EditorLayoutCTreeExpandCB(
	GtkCTree *ctree, GtkCTreeNode *node, gpointer data
);

gint EditorSyntaxHighlightTimeoutCB(gpointer data);

gint EditorTextKeyEventCB(
	GtkEditable *editable, GdkEventKey *key, gpointer data
);
void EditorTextChangeCB(GtkEditable *editable, gpointer data);
void EditorTextInsertCB(
	GtkEditable *editable, const gchar *text, gint length,
	gint *position, gpointer data
);
void EditorTextDeleteCB(
	GtkEditable *editable, gint start_pos, gint end_pos, gpointer data
);

void EditorReconfiguredCB(editor_struct *editor);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : TRUE)


/*
 *	Set to the pointer of the editor that is currently having
 *	an undo or redo operation done on it, this is to keep track 
 *	of and discard signals like "insert_text" and "delete_text"
 *	which would be sent if text was being reinserted or redeleted
 *	during and undo or redo process.
 */
static editor_struct *editor_processing_undo = NULL;


/*
 *	Destroys the EditorItem.
 */
void EditorItemDestroyCB(gpointer data)
{
	EditorItemDelete(EDITOR_ITEM(data));
}

/*
 *	Toplevel GtkWindow "key_press_event" or "key_release_event"
 *	signal callback.
 */
gint EditorKeyEventCB(GtkWidget *widget, GdkEventKey *key, gpointer data)
{
	gint status = FALSE;
	gboolean press;
	me_core_struct *core;
	editor_struct *editor = EDITOR(data);
	if((key == NULL) || (editor == NULL))
	    return(status);

	if(editor->freeze_count > 0)
	    return(status);

	press = (key->type == GDK_KEY_PRESS) ? TRUE : FALSE;
	core = editor->core;

	switch(key->keyval)
	{
	  case GDK_F3:
	    if(press)
	    {
		EditorManualOpenCB(widget, data);
	    }
	    status = TRUE;
	    break;
	}

	return(status);
}

/*
 *	Text "button_press_event" or "button_release_event" signal
 *	callback.
 */
gint EditorTextButtonEventCB(
	GtkWidget *widget, GdkEventButton *button, gpointer data
)
{
	gint status = FALSE;
	GtkText *text;
	editor_struct *editor = EDITOR(data);
	if((widget == NULL) || (button == NULL) || (editor == NULL))
	    return(status);

	if(editor->freeze_count > 0)
	    return(status);

	if(!GTK_IS_TEXT(widget))
	    return(status);

	text = GTK_TEXT(widget);

	switch((gint)button->type)
	{
	  case GDK_BUTTON_PRESS:
	    switch(button->button)
	    {
	      case GDK_BUTTON3:         /* Popup menu */
		/* Map the popup menu */
		gtk_menu_popup(
		    GTK_MENU(editor->edit_panel_menu),
		    NULL, NULL, NULL, editor,
		    button->button, button->time
		);
		/* Need to mark the text widget's button as 0 or
		 * else it will keep marking.
		 */
		text->button = 0;
		gtk_grab_remove(widget);
		gtk_signal_emit_stop_by_name(
		    GTK_OBJECT(widget), "button_press_event"
		);
		status = TRUE;
		break;

	      case GDK_BUTTON4:         /* Scroll up */
		if(text->vadj != NULL)  
		{
		    GtkAdjustment *adj = text->vadj;
		    const gfloat inc = MAX(
			(0.25f * adj->page_size),
			adj->step_increment
		    );
		    gfloat v = adj->value - inc;
		    if(v > (adj->upper - adj->page_size))
			v = adj->upper - adj->page_size;
		    if(v < adj->lower)
			v = adj->lower;
		    gtk_adjustment_set_value(adj, v);
		}
		/* Need to mark the GtkText button as 0 or else it will
		 * keep marking
		 */
		text->button = 0;
		gtk_grab_remove(widget);
		gtk_signal_emit_stop_by_name(
		    GTK_OBJECT(widget), "button_press_event"
		);
		status = TRUE;
		break;

	      case GDK_BUTTON5:         /* Scroll down */
		if(text->vadj != NULL)  
		{
		    GtkAdjustment *adj = text->vadj;
		    const gfloat inc = MAX(
			(0.25f * adj->page_size),
			adj->step_increment
		    );
		    gfloat v = adj->value + inc;
		    if(v > (adj->upper - adj->page_size))
			v = adj->upper - adj->page_size;
		    if(v < adj->lower)
			v = adj->lower;
		    gtk_adjustment_set_value(adj, v);
		}
		/* Need to mark the GtkText button as 0 or else it will
		 * keep marking
		 */
		text->button = 0;
		gtk_grab_remove(widget);
		gtk_signal_emit_stop_by_name(
		    GTK_OBJECT(widget), "button_press_event"
		);
		status = TRUE; 
		break;
	    }
	    break;

	  case GDK_BUTTON_RELEASE:
	    switch(button->button)
	    {
	      case GDK_BUTTON3:         /* Popup menu */
		/* Need to mark the GtkText button as 0 or else it will
		 * keep marking
		 */
		text->button = 0;
		gtk_grab_remove(widget);
		gtk_signal_emit_stop_by_name(
		    GTK_OBJECT(widget), "button_release_event"
		);
		status = TRUE;
		break;

	      case GDK_BUTTON4:         /* Scroll up */ 
		/* Need to mark the GtkText button as 0 or else it will
		 * keep marking
		 */
		text->button = 0;
		gtk_grab_remove(widget);
		gtk_signal_emit_stop_by_name(
		    GTK_OBJECT(widget), "button_release_event"
		);
		status = TRUE;
		break;

	      case GDK_BUTTON5:         /* Scroll down */
		/* Need to mark the GtkText button as 0 or else it will
		 * keep marking
		 */
		text->button = 0;
		gtk_grab_remove(widget);
		gtk_signal_emit_stop_by_name(
		    GTK_OBJECT(widget), "button_release_event"
		);
		status = TRUE;
		break;
	    }
	    break;
	}

	return(status);
}


/*
 *	ToolBarItem "activate" signal callback.
 */
void EditorToolBarItemActivateCB(
	ToolBarItem *item, gint id, gpointer data
)
{
	GtkWidget *widget;
	editor_struct *editor = EDITOR(data);
	if((item == NULL) || (editor == NULL))
	    return;

	if(editor->freeze_count > 0)
	    return;

	widget = ToolBarItemGetWidget(item);

	switch((editor_opid)id)
	{
	  case EDITOR_OPID_NEW:
	    EditorManualNewCB(widget, data);
	    break;
	  case EDITOR_OPID_OPEN:
	    EditorManualOpenCB(widget, data);
	    break;
	  case EDITOR_OPID_SAVE:
	    EditorManualSaveCB(widget, data);
	    break;
	  case EDITOR_OPID_SAVE_AS:
	    EditorManualSaveAsCB(widget, data);
	    break;
	  case EDITOR_OPID_CLOSE_MANUAL:
	    EditorManualCloseCB(widget, data);
	    break;
	  case EDITOR_OPID_UNDO:
	    EditorUndoCB(widget, data);
	    break;
	  case EDITOR_OPID_REDO:
	    EditorRedoCB(widget, data);
	    break;
	  case EDITOR_OPID_CUT:
	    EditorCutCB(widget, data);
	    break;
	  case EDITOR_OPID_COPY:
	    EditorCopyCB(widget, data);
	    break;
	  case EDITOR_OPID_PASTE:
	    EditorPasteCB(widget, data);
	    break;
	  case EDITOR_OPID_NEW_HEADER:
	    EditorManualAddHeaderCB(widget, data);
	    break;
	  case EDITOR_OPID_NEW_SECTION:
	    EditorManualAddSectionCB(widget, data);
	    break;
	  case EDITOR_OPID_REMOVE:
	    EditorManualRemoveSectionCB(widget, data);
	    break;
	  case EDITOR_OPID_PREVIEW:
	    EditorManualPreviewCB(widget, data);
	    break;
	}
}


/*
 *	Toplevel GtkWindow "delete_event" signal callback.
 */
gint EditorDeleteEventCB(GtkWidget *widget, GdkEvent *event, gpointer data)
{
	gboolean yes_to_all = FALSE;
	GtkWidget *toplevel;
	GtkCTreeNode *node;
	editor_item_struct *item;
	me_core_struct *core;
	editor_struct *editor = EDITOR(data);
	if(editor == NULL)
	    return(TRUE);

	if(editor->freeze_count > 0)
	    return(TRUE);

	editor->freeze_count++;

	core = editor->core;
	toplevel = editor->toplevel;

	/* Check if any of the manual pages have unsaved changes */
	node = EditorGetFirstToplevelNode(editor);
	while(node != NULL)
	{
	    item = EditorBranchGetData(
		GTK_CTREE(editor->layout_ctree),
		node
	    );
	    if(item == NULL)
	    {
		node = GTK_CTREE_NODE_SIBLING(node);
		continue;
	    }

	    if(item->type != EDITOR_ITEM_FILE)
	    {
		node = GTK_CTREE_NODE_SIBLING(node);
		continue;
	    }

	    /* Does this manual page have any unsaved changes? */
	    if(item->has_changes)
	    {
		gint response;
		gchar *msg = g_strdup_printf(
		    "Save changes to manual page `%s'?",
		    item->name
		);
		if(yes_to_all)
		{
		    response = CDIALOG_RESPONSE_YES_TO_ALL;
		}
		else
		{
		    CDialogSetTransientFor(toplevel);
		    response = CDialogGetResponse(
"Save Changes",
			msg,
			NULL,
			CDIALOG_ICON_WARNING,
			CDIALOG_BTNFLAG_YES | CDIALOG_BTNFLAG_YES_TO_ALL |
			CDIALOG_BTNFLAG_NO | CDIALOG_BTNFLAG_CANCEL,
			CDIALOG_BTNFLAG_YES
		    );
		    CDialogSetTransientFor(NULL);
		}
		g_free(msg);

		/* Check response */
		switch(response)
		{
		  case CDIALOG_RESPONSE_YES_TO_ALL:
		    yes_to_all = TRUE;
		  case CDIALOG_RESPONSE_YES:
		  case CDIALOG_RESPONSE_OK:
		    /* Has this manual page never been saved before? */
		    if(STRISEMPTY(item->path))
		    {
			/* Query the user to save the manual page as and
			 * saves the manual page as
			 */
			if(EditorSaveAsQuery(
			    editor,
			    node,		/* Manual page */
			    TRUE		/* Verbose */
			) != 0)
			{
			    /* User aborted or save failed */
			    editor->freeze_count--;
			    return(TRUE);
			}
		    }
		    else
		    {
			/* Save */
			if(EditorFileSave(
			    editor,
			    node
			) != 0)
			{
			    editor->freeze_count--;
			    return(TRUE);
			}
		    }
		    break;

		  case CDIALOG_RESPONSE_NO:
		    break;

		  case CDIALOG_RESPONSE_CANCEL:
		  case CDIALOG_RESPONSE_NOT_AVAILABLE:
		    /* Response not available */
		    editor->freeze_count--;
		    return(TRUE);
		    break;
		}
	    }

	    node = GTK_CTREE_NODE_SIBLING(node);
	}

	/* Record the Editor values to the CfgList */
	EditorRecordConfiguration(editor);

	/* Unmap the Editor, the core's main timeout callback will
	 * later detect that the Editor has been unmapped and delete it
	 */
	EditorUnmap(editor);

	editor->freeze_count--;

	return(TRUE);
}

/*
 *	Close callback.
 */
void EditorCloseCB(GtkWidget *widget, gpointer data)
{
	EditorDeleteEventCB(widget, NULL, data);
}

/*
 *	Close all windows.
 */
void EditorExitCB(GtkWidget *widget, gpointer data)
{
	me_core_struct *core;
	editor_struct *editor = EDITOR(data);
	if(editor == NULL)
	    return;

	if(editor->freeze_count > 0)
	    return;

	core = editor->core;
	core->need_close_all_windows = TRUE;
}


/*
 *	New manual page callback.
 */
void EditorManualNewCB(GtkWidget *widget, gpointer data)
{
	gint		i,
			ncolumns;
	gchar		**strv,
			*title;
	GtkWidget *toplevel;
	GtkCList *clist;
	GtkCTree *ctree;
	GtkCTreeNode	*node,
			*insert_node,
			*parent_node;
	me_pixmaps_list_struct *pixmaps_list;
	me_core_struct *core;
	editor_struct *editor = EDITOR(data);
	if(editor == NULL)
	    return;

	if(editor->freeze_count > 0)
	    return;

	EditorSetBusy(editor, TRUE);
	EditorSetStatusMessage(editor, "Creating new manual page...");
	editor->freeze_count++;

	toplevel = editor->toplevel;
	ctree = GTK_CTREE(editor->layout_ctree);
	clist = GTK_CLIST(ctree);
	core = editor->core;
	pixmaps_list = &core->pixmaps_list;

	/* Get the selected toplevel node to insert at */
	insert_node = EditorGetToplevelNode(
	    EditorCurrentSelectedNode(editor)
	);

	/* Allocate the row cell values */
	ncolumns = MAX(clist->columns, 1);
	strv = (gchar **)g_malloc(ncolumns * sizeof(gchar *));
	if(strv == NULL)
	{
	    EditorSetStatusMessage(
		editor,
"Create new manual page failed"
	    );
	    editor->freeze_count--;
	    EditorSetBusy(editor, FALSE);
	    return;
	}

	for(i = 0; i < ncolumns; i++)
	    strv[i] = "";

	/* Increment the untitled count and set the new title */
	core->untitled_count++;
	title = g_strdup_printf(
	    "Untitled %i",
	    core->untitled_count
	);

	if(ncolumns > 0)
	    strv[0] = title;

	/* Create the toplevel node for the new manual page */
	node = gtk_ctree_insert_node(
	    ctree,
	    NULL,				/* Toplevel */
	    insert_node,			/* Insert before this sibling */
	    strv,
	    ME_LIST_ICON_TEXT_SPACING,
	    pixmaps_list->manual_closed_20x20,
	    pixmaps_list->manual_closed_20x20_mask,
	    pixmaps_list->manual_opened_20x20,
	    pixmaps_list->manual_opened_20x20_mask,
	    FALSE,				/* Not leaf */
	    TRUE				/* Expanded */
	);
	if(node != NULL)
	{
	    editor_item_struct *item = EditorItemNewWithValues(
		EDITOR_ITEM_FILE,
		editor,
		node
	    );
	    if(item != NULL)
	    {
		g_free(item->name);
		item->name = STRDUP(title);

		/* Do not set the path, implies not saved yet */

		item->has_changes = TRUE;

		EditorBranchSetData(
		    ctree,
		    node,
		    item, EditorItemDestroyCB
		);
	    }
	}
	parent_node = node;

	/* Create the header node */
	if(ncolumns > 0)
	    strv[0] = "Header";
	node = gtk_ctree_insert_node(
	    ctree,
	    parent_node,
	    NULL,
	    strv,
	    ME_LIST_ICON_TEXT_SPACING,
	    pixmaps_list->manpage_heading_20x20,
	    pixmaps_list->manpage_heading_20x20_mask,
	    pixmaps_list->manpage_heading_20x20,
	    pixmaps_list->manpage_heading_20x20_mask,
	    TRUE,				/* Is leaf */
	    FALSE				/* Not expanded */
	);
	if(node != NULL)
	{
	    editor_item_struct *item = EditorItemNewWithValues(
		EDITOR_ITEM_HEADER,
		editor,
		node
	    );
	    if(item != NULL)
	    {
		g_free(item->header_name);
		item->header_name = STRDUP(title);

		item->has_changes = TRUE;

		EditorBranchSetData(
		    ctree,
		    node,
		    item, EditorItemDestroyCB
		);
	    }
	}

	/* Create the first section node */
	if(ncolumns > 0)
	    strv[0] = "SECTION 1";
	node = gtk_ctree_insert_node(
	    ctree,
	    parent_node,
	    NULL,
	    strv,
	    ME_LIST_ICON_TEXT_SPACING,
	    pixmaps_list->manpage_section_20x20,
	    pixmaps_list->manpage_section_20x20_mask,
	    pixmaps_list->manpage_section_20x20,
	    pixmaps_list->manpage_section_20x20_mask,
	    TRUE,				/* Is leaf */
	    FALSE				/* Not expanded */
	);
	if(node != NULL)
	{
	    editor_item_struct *item = EditorItemNewWithValues(
		EDITOR_ITEM_SECTION,
		editor,
		node
	    );
	    if(item != NULL)
	    {
		g_free(item->section_name);
		item->section_name = STRDUP(strv[0]);

		item->has_changes = TRUE;

		EditorBranchSetData(
		    ctree,
		    node,
		    item, EditorItemDestroyCB
		);
	    }
	}

	g_free(strv);
	g_free(title);

	EditorUpdateMenus(editor);
	EditorSetStatusMessage(editor, "New manual page created");
	EditorSetStatusProgress(editor, 0.0f);

	editor->freeze_count--;
	EditorSetBusy(editor, FALSE);

	/* Select the new node */
	EditorSelect(
	    editor,
	    node
	);
}

/*
 *	Open manual page callback.
 */
void EditorManualOpenCB(GtkWidget *widget, gpointer data)
{
	GtkCTreeNode *insert_node;
	editor_struct *editor = EDITOR(data);
	if(editor == NULL)
	    return;

	if(editor->freeze_count > 0)
	    return;

	EditorSetBusy(editor, TRUE);
	editor->freeze_count++;

	/* Get the selected manual page to insert at */
	insert_node = EditorCurrentSelectedNode(editor);

	/* Query the user to open the manual page and open the manual
	 * page
	 */
	(void)EditorOpenQuery(
	    editor,
	    insert_node,			/* Insert at this manual page */
	    TRUE				/* Verbose */
	);

	EditorUpdateMenus(editor);

	editor->freeze_count--;
	EditorSetBusy(editor, FALSE);
}

/*
 *	New manual page from template.
 */
void EditorManualNewFromTemplateCB(GtkWidget *widget, gpointer data)
{
	GtkCTreeNode *insert_node;
	editor_struct *editor = EDITOR(data);
	if(editor == NULL)
	    return;

	if(editor->freeze_count > 0)
	    return;

	EditorSetBusy(editor, TRUE);
	editor->freeze_count++;

	/* Get the node to insert at */
	insert_node = EditorCurrentSelectedNode(editor);

	/* Query the user to open a manual page template and create
	 * the new manual page
	 */
	(void)EditorNewTemplateQuery(
	    editor,
	    insert_node,			/* Insert at this manual page */
	    TRUE				/* Verbose */
	);

	EditorUpdateMenus(editor);

	editor->freeze_count--;
	EditorSetBusy(editor, FALSE);
}

/*
 *	Save manual page callback.
 */
void EditorManualSaveCB(GtkWidget *widget, gpointer data)
{
	gint status;
	const gchar *path;
	GtkCTree *ctree;
	GtkCTreeNode *toplevel_node;
	GtkWidget *toplevel;
	editor_item_struct *item;
	editor_struct *editor = EDITOR(data);
	if(editor == NULL)
	    return;

	if(editor->freeze_count > 0)
	    return;

	EditorSetBusy(editor, TRUE);
	editor->freeze_count++;

	toplevel = editor->toplevel;
	ctree = GTK_CTREE(editor->layout_ctree);

	/* Get toplevel node of the selected manual page */
	toplevel_node = EditorGetToplevelNode(
	    EditorCurrentSelectedNode(editor)
	);
	if(toplevel_node == NULL)
	{
	    CDialogSetTransientFor(toplevel);
	    CDialogGetResponse(
		"Save Failed",
"No manual page selected.",
		NULL,
		CDIALOG_ICON_WARNING,
		CDIALOG_BTNFLAG_OK, 
		CDIALOG_BTNFLAG_OK
	    );
	    CDialogSetTransientFor(NULL);
	    editor->freeze_count--;
	    EditorSetBusy(editor, FALSE);
	    return;
	}

	/* Get the EditorItem and path of the selected manual page */
	item = EditorBranchGetData(
	    ctree,
	    toplevel_node
	);
	path = (item != NULL) ? item->path : NULL;

	/* If the selected manual page does not have a path set then
	 * it means it was never saved before, in which case we should
	 * save as
	 */
	if(STRISEMPTY(path))
	{
	    /* Thaw first before save as or else the callback will
	     * not respond
	     */
	    editor->freeze_count--;
	    EditorManualSaveAsCB(widget, data);
	    EditorSetBusy(editor, FALSE);
	    return;
	}

	/* Save */
	status = EditorFileSave(
	    editor,
	    toplevel_node
	);
	if(status != 0)
	{
	    /* Error occured while saving */
	    const gint error_code = (gint)errno;
	    gchar *msg;
	    switch(status)
	    {
	      /* User aborted */
	      case -5:
	      case -4:
		break;

	      default:
		msg = g_strdup_printf(
"%s:\n\
\n\
    %s",
		    g_strerror(error_code),
		    path
		);
		CDialogSetTransientFor(toplevel);
		CDialogGetResponse(
		    "Save Failed",
		    msg,
		    NULL,
		    CDIALOG_ICON_WARNING,
		    CDIALOG_BTNFLAG_OK,
		    CDIALOG_BTNFLAG_OK
		);
		g_free(msg);
		CDialogSetTransientFor(NULL);
		break;
	    }
	}
	else
	{
	    /* Save successful
	     *
	     * Remove the has changes flag on all the nodes for this
	     * manual page
	     */
	    GtkCTreeNode *node = GTK_CTREE_NODE_CHILD(toplevel_node);
	    while(node != NULL)
	    {
		item = EDITOR_ITEM(gtk_ctree_node_get_row_data(
		    ctree,
		    node
		));
		if(item != NULL)
		    item->has_changes = FALSE;

		node = GTK_CTREE_NODE_SIBLING(node);
	    }

	    /* Mark toplevel_node as no longer having changes */
	    item = EDITOR_ITEM(gtk_ctree_node_get_row_data(
		ctree,
		toplevel_node
	    ));
	    if(item != NULL)
		item->has_changes = FALSE;

	    /* Reget the values for the selected node */
	    EditorGetValues(
		editor,
		EditorCurrentSelectedNode(editor)
	    );
	}

	EditorUpdateMenus(editor);

	editor->freeze_count--;
	EditorSetBusy(editor, FALSE);
}

/*
 *	Save manual page as callback.
 */
void EditorManualSaveAsCB(GtkWidget *widget, gpointer data)
{
	GtkWidget *toplevel;
	GtkCTree *ctree;
	GtkCTreeNode *node;
	me_core_struct *core;
	editor_struct *editor = EDITOR(data);
	if(editor == NULL)
	    return;

	if(editor->freeze_count > 0)
	    return;

	EditorSetBusy(editor, TRUE);
	editor->freeze_count++;

	toplevel = editor->toplevel;
	ctree = GTK_CTREE(editor->layout_ctree);
	core = editor->core;

	/* Get the GtkCTreeNode of the selected manual page */
	node = EditorGetToplevelNode(
	    EditorCurrentSelectedNode(editor)
	);
	if(node == NULL)
	{
	    CDialogSetTransientFor(toplevel);
	    CDialogGetResponse(
		"Save Failed",
"No manual page selected.",
		NULL,
		CDIALOG_ICON_WARNING,
		CDIALOG_BTNFLAG_OK, 
		CDIALOG_BTNFLAG_OK
	    );
	    CDialogSetTransientFor(NULL);
	    editor->freeze_count--;
	    EditorSetBusy(editor, FALSE);
	    return;
	}

	/* Query the user to save the manual page as and saves the
	 * manual page as
	 */
	(void)EditorSaveAsQuery(
	    editor,
	    node,				/* Manual page */
	    TRUE				/* Verbose */
	);

	EditorUpdateMenus(editor);

	editor->freeze_count--;
	EditorSetBusy(editor, FALSE);
}

/*
 *	Revert callback.
 */
void EditorManualRevertCB(GtkWidget *widget, gpointer data)
{
	gint		error_code,
			status,
			response;
	gchar		*msg,
			*path;
	GtkCTree *ctree;
	GtkCTreeNode	*node,
			*prev_sibling_node = NULL;
	GtkWidget *toplevel;
	editor_item_struct *item;
	editor_struct *editor = EDITOR(data);
	if(editor == NULL)
	    return;

	if(editor->freeze_count > 0)
	    return;

	EditorSetBusy(editor, TRUE);
	editor->freeze_count++;

	toplevel = editor->toplevel;
	ctree = GTK_CTREE(editor->layout_ctree);

	node = EditorGetToplevelNode(
	    EditorCurrentSelectedNode(editor)
	);
	if(node == NULL)
	{
	    editor->freeze_count--;
	    EditorSetBusy(editor, FALSE);
	    return;
	}

	item = EditorBranchGetData(ctree, node);
	if(item == NULL)
	{
	    editor->freeze_count--;
	    EditorSetBusy(editor, FALSE);
	    return;
	}

	/* Has the selected manual page never been saved? */
	if(STRISEMPTY(item->path))
	{
	    CDialogSetTransientFor(toplevel);
	    CDialogGetResponse(
"Revert Failed",
"The selected manual page has never been saved and\n\
therefore has no file to revert to.",
		NULL,
		CDIALOG_ICON_WARNING,
		CDIALOG_BTNFLAG_OK,
		CDIALOG_BTNFLAG_OK
	    );
	    CDialogSetTransientFor(NULL);
	    editor->freeze_count--;
	    EditorSetBusy(editor, FALSE);
	    return;
	}

	/* Confirm revert */
	msg = g_strdup_printf(
"Revert manual page \"%s\"?",
	    item->name
	);
	CDialogSetTransientFor(toplevel);
	response = CDialogGetResponse(
"Confirm Revert",
	    msg,
	    NULL,
	    CDIALOG_ICON_QUESTION,
	    CDIALOG_BTNFLAG_YES | CDIALOG_BTNFLAG_NO,
	    CDIALOG_BTNFLAG_NO
	);
	CDialogSetTransientFor(NULL);
	g_free(msg);
	switch(response)
	{
	  case CDIALOG_RESPONSE_YES_TO_ALL:
	  case CDIALOG_RESPONSE_YES:
	  case CDIALOG_RESPONSE_OK:
	    break;

	  default:
	    editor->freeze_count--;
	    EditorSetBusy(editor, FALSE);
	    return;
	    break;
	}

	EditorSetStatusMessage(editor, "Reverting manual page...");

	/* Get/copy the manual page's path */
	path = g_strdup(item->path);

	/* Record/get this node's sibling node */
	prev_sibling_node = GTK_CTREE_NODE_SIBLING(node);

	/* Remove this manual page's node */
	gtk_ctree_remove_node(ctree, node);
/*	node = NULL; */

	/* Reopen the manual page */
	status = EditorFileOpen(
	    editor,
	    path,
	    prev_sibling_node,
	    FALSE				/* Not a template */
	);
	error_code = (int)errno;
	switch(status)
	{
	  case -5:
	  case -4:
	    break;

	  case 0:
	    break;

	  default:
	    msg = g_strdup_printf(
"%s:\n\
\n\
    %s",
		g_strerror(error_code),
		path
	    );
	    CDialogSetTransientFor(toplevel);
	    CDialogGetResponse(
		"Revert Failed",
		msg,
		NULL,
		CDIALOG_ICON_WARNING,
		CDIALOG_BTNFLAG_OK,
		CDIALOG_BTNFLAG_OK
	    );
	    g_free(msg);
	    CDialogSetTransientFor(NULL);
	    break;
	}

	g_free(path);

	EditorSetStatusMessage(editor, "Revert done");

	editor->freeze_count--;
	EditorSetBusy(editor, FALSE);
}

/*
 *	Close manual page callback.
 */
void EditorManualCloseCB(GtkWidget *widget, gpointer data)
{
	GtkCList *clist;
	GtkCTree *ctree;
	GtkCTreeNode *node;
	GtkWidget *toplevel;
	editor_item_struct *item;
	me_core_struct *core;
	editor_struct *editor = EDITOR(data);
	if(editor == NULL)
	    return;

	if(editor->freeze_count > 0)
	    return;

	EditorSetBusy(editor, TRUE);
	editor->freeze_count++;

	toplevel = editor->toplevel;
	ctree = GTK_CTREE(editor->layout_ctree);
	clist = GTK_CLIST(ctree);
	core = editor->core;

	/* Get the selected node */
	node = EditorCurrentSelectedNode(editor);
	if(node == NULL)
	{
	    editor->freeze_count--;
	    EditorSetBusy(editor, FALSE);
	    return;
	}

	/* Apply data on currently selected node to ensure that we
	 * have the most recent data.
	 */
	EditorSetValues(editor, node);

	/* Get the toplevel node of the specified node */
	node = EditorGetToplevelNode(node);
	if(node == NULL)
	{
	    editor->freeze_count--;
	    EditorSetBusy(editor, FALSE);
	    return;
	}
	/* Node should now be the toplevel node */

	/* Get the EditorItem */
	item = EditorBranchGetData(ctree, node);
	if(item == NULL)
	{
	    /* No EditorItem, just remove the GtKCTreeNode */
	    gtk_ctree_remove_node(ctree, node);
	    editor->freeze_count--;
	    EditorSetBusy(editor, FALSE);
	    return;
	}

	/* Check if this manual page has changes */
	if(item->has_changes)
	{
	    /* Has changes, prompt for save */
	    gint response;
	    gchar *msg = g_strdup_printf(
"Save changes to manual page `%s'?",
		item->name  
	    );
	    CDialogSetTransientFor(toplevel);
	    response = CDialogGetResponse(
		"Save Changes",
		msg,
		NULL,
		CDIALOG_ICON_WARNING,
		CDIALOG_BTNFLAG_YES | CDIALOG_BTNFLAG_NO |
		    CDIALOG_BTNFLAG_CANCEL,
		CDIALOG_BTNFLAG_YES
	    );
	    CDialogSetTransientFor(NULL);
	    g_free(msg);
	    switch(response)
	    {
	      case CDIALOG_RESPONSE_YES:
	      case CDIALOG_RESPONSE_YES_TO_ALL:
	      case CDIALOG_RESPONSE_OK:
		/* Request to save, check if the full path is set.
		 * If the full path is not set (regardless of if the
		 * name is set or not, we need to consider that
		 * this file was never saved before and thus we need
		 * to request from the user a `save as'.
		 */
		if(STRISEMPTY(item->path))
		{
		    /* Query the user to save the manual page as and
		     * saves the manual page as
		     */
		    if(EditorSaveAsQuery(
			editor,
			node,			/* Manual page */
			TRUE			/* Verbose */
		    ) != 0)
		    {
			/* User aborted or save failed */
			editor->freeze_count--;
			EditorSetBusy(editor, FALSE);
			return;
		    }
		}
		else
		{           
		    /* Save */
		    if(EditorFileSave(
			editor,
			node			/* Manual page */
		    ) != 0)
		    {
			/* Save failed, do not close */
			editor->freeze_count--;
			EditorSetBusy(editor, FALSE);
			return;
		    }
		}
		break;

	      case CDIALOG_RESPONSE_NO:
		break;

	      case CDIALOG_RESPONSE_CANCEL:
	      case CDIALOG_RESPONSE_NOT_AVAILABLE:
		editor->freeze_count--;
		EditorSetBusy(editor, FALSE);
		return;
		break;
	    }
	}

	/* Remove this manual page from the layout GtkCTree */
	gtk_clist_freeze(clist);
	gtk_ctree_remove_node(ctree, node);
	gtk_clist_thaw(clist);

	/* Clear the undo/redo list */
	EditorClearUndosList(editor);

	/* Clear the values */
	EditorClearValues(editor);

	EditorSetStatusMessage(editor, "Manual page closed");
	EditorUpdateMenus(editor);

	editor->freeze_count--;
	EditorSetBusy(editor, FALSE);
}

/*
 *	Add header callback.
 */
void EditorManualAddHeaderCB(GtkWidget *widget, gpointer data)
{
	gint		i,
			ncolumns;
	gchar **strv;
	GdkPixmap *closed_pm, *opened_pm;
	GdkBitmap *closed_mask, *opened_mask;
	GtkWidget *toplevel;
	GtkCList *clist;
	GtkCTree *ctree;
	GtkCTreeNode	*node,
			*toplevel_node,
			*new_node = NULL;
	editor_item_struct *item;
	me_core_struct *core;
	me_pixmaps_list_struct *pixmaps_list;
	editor_struct *editor = EDITOR(data);
	if(editor == NULL)
	    return;

	if(editor->freeze_count > 0)
	    return;

	EditorSetBusy(editor, TRUE);
	editor->freeze_count++;

	toplevel = editor->toplevel;
	ctree = GTK_CTREE(editor->layout_ctree);
	clist = GTK_CLIST(ctree);
	core = editor->core;
	pixmaps_list = &core->pixmaps_list;

	/* Get the selected node */
	node = EditorCurrentSelectedNode(editor);
	if(node == NULL)
	{
	    CDialogSetTransientFor(toplevel);
	    CDialogGetResponse(
"Add Header Failed",
"No manual page selected.\n\
\n\
Please select the manual page that you want\n\
to add a header to",
		NULL,
		CDIALOG_ICON_WARNING,
		CDIALOG_BTNFLAG_OK,
		CDIALOG_BTNFLAG_OK
	    );
	    CDialogSetTransientFor(NULL);
	    editor->freeze_count--;
	    EditorSetBusy(editor, FALSE);
	    return;
	}

	/* Get the toplevel node of the selected node */
	toplevel_node = EditorGetToplevelNode(node);
	if(toplevel_node == NULL)
	{
	    CDialogSetTransientFor(toplevel);
	    CDialogGetResponse(
"Add Header Failed",
"No manual page selected.\n\
\n\
Please select the manual page that you want\n\
to add a header to",
		NULL,
		CDIALOG_ICON_WARNING,
		CDIALOG_BTNFLAG_OK,
		CDIALOG_BTNFLAG_OK
	    );
	    CDialogSetTransientFor(NULL);
	    editor->freeze_count--;
	    EditorSetBusy(editor, FALSE);
	    return;
	}
	else
	{
	    /* Check if this manual page already has a header */
	    GtkCTreeNode *node = GTK_CTREE_NODE_CHILD(toplevel_node);
	    while(node != NULL)
	    {
		item = EDITOR_ITEM(gtk_ctree_node_get_row_data(
		    ctree,
		    node
		));
		if(item != NULL)
		{
		    if(item->type == EDITOR_ITEM_HEADER)
		    {
			CDialogSetTransientFor(toplevel);
			CDialogGetResponse(
"Add Header Failed",
"The selected manual page already has a header.",
			    NULL,
			    CDIALOG_ICON_WARNING,
			    CDIALOG_BTNFLAG_OK,
			    CDIALOG_BTNFLAG_OK
			);
			CDialogSetTransientFor(NULL);
			editor->freeze_count--;
			EditorSetBusy(editor, FALSE);
			return;
		    }
		}

		node = GTK_CTREE_NODE_SIBLING(node);
	    }
	}

	gtk_clist_freeze(clist);

	ncolumns = MAX(clist->columns, 1);
	strv = (gchar **)g_malloc(ncolumns * sizeof(gchar *));
	for(i = 0; i < ncolumns; i++)
	    strv[i] = "";

	if(ncolumns > 0)
	    strv[0] = "Header";
	closed_pm = pixmaps_list->manpage_heading_20x20;
	closed_mask = pixmaps_list->manpage_heading_20x20_mask;
	opened_pm = closed_pm;
	opened_mask = closed_mask;

	/* Get the toplevel node's EditorItem and add the header
	 * to it
	 */
	item = EditorBranchGetData(
	    ctree,
	    toplevel_node
	);
	if(item != NULL)
	{
	    /* Make sure the EditorItem is of type EDITOR_ITEM_FILE */
	    if(item->type == EDITOR_ITEM_FILE)
	    {
		/* Add the header node as the first child */
		new_node = gtk_ctree_insert_node(
		    ctree,
		    toplevel_node,		/* Parent */
		    GTK_CTREE_NODE_CHILD(toplevel_node),
		    strv,
		    ME_LIST_ICON_TEXT_SPACING,
		    closed_pm, closed_mask,
		    opened_pm, opened_mask,
		    TRUE,			/* Is leaf */
		    FALSE			/* Not expanded */
		);
	    }
	}
	if(new_node != NULL)
	{
	    /* Create a new EditorItem for the header */
	    editor_item_struct *item = EditorItemNewWithValues(
		EDITOR_ITEM_HEADER,
		editor,
		new_node
	    );
	    EditorBranchSetData(
		ctree,
		new_node,
		item, EditorItemDestroyCB
	    );

	    /* Set the has changes flag on the new EditorItem */
	    EditorSetNodeHasChanges(
		editor,
		new_node,
		TRUE
	    );
	}

	/* Update Layout GtkCTree's column width */
	gtk_clist_set_column_width(
	    clist,
	    0,
	    gtk_clist_optimal_column_width(clist, 0)
	);

	gtk_clist_thaw(clist);

	g_free(strv);

	EditorSetStatusMessage(editor, "New header added");
	EditorUpdateMenus(editor);

	editor->freeze_count--;
	EditorSetBusy(editor, FALSE);

	/* Select the new node */
	EditorSelect(
	    editor,
	    new_node
	);
}

/*
 *	Add section callback.
 */
void EditorManualAddSectionCB(GtkWidget *widget, gpointer data)
{
	gint		i,
			ncolumns;
	gchar **strv;
	GdkPixmap *closed_pm, *opened_pm;
	GdkBitmap *closed_mask, *opened_mask;
	GtkWidget *toplevel;
	GtkCList *clist;
	GtkCTree *ctree;
	GtkCTreeNode		*node,
				*toplevel_node,
				*new_node = NULL;
	me_core_struct *core;
	me_pixmaps_list_struct *pixmaps_list;
	editor_item_struct *item;
	editor_struct *editor = EDITOR(data);
	if(editor == NULL)
	    return;

	if(editor->freeze_count > 0)
	    return;

	EditorSetBusy(editor, TRUE);
	editor->freeze_count++;

	toplevel = EDITOR_TOPLEVEL(editor->toplevel);
	ctree = GTK_CTREE(editor->layout_ctree);
	clist = GTK_CLIST(ctree);
	core = editor->core;
	pixmaps_list = &core->pixmaps_list;

	/* Get the selected node */
	node = EditorCurrentSelectedNode(editor);
	if(node == NULL)
	{
	    CDialogSetTransientFor(toplevel);
	    CDialogGetResponse(
"Add Section Failed",
"No manual page selected.\n\
\n\
Please select the manual page that you want\n\
to add the section to",
		NULL,
		CDIALOG_ICON_WARNING,
		CDIALOG_BTNFLAG_OK,
		CDIALOG_BTNFLAG_OK
	    );
	    CDialogSetTransientFor(NULL);
	    editor->freeze_count--;
	    EditorSetBusy(editor, FALSE);
	    return;
	}

	/* Get the toplevel node of the selected node */
	toplevel_node = EditorGetToplevelNode(node);
	if(toplevel_node == NULL)
	{
	    CDialogSetTransientFor(toplevel);
	    CDialogGetResponse(
"Add Section Failed",
"No manual page selected.\n\
\n\
Please select the manual page that you want\n\
to add the section to",
		NULL,
		CDIALOG_ICON_WARNING,
		CDIALOG_BTNFLAG_OK,
		CDIALOG_BTNFLAG_OK
	    );
	    CDialogSetTransientFor(NULL);
	    editor->freeze_count--;
	    EditorSetBusy(editor, FALSE);
	    return;
	}

	gtk_clist_freeze(clist);

	ncolumns = MAX(clist->columns, 1);
	strv = (gchar **)g_malloc(ncolumns * sizeof(gchar *));
	for(i = 0; i < ncolumns; i++)
	    strv[i] = "";

	if(ncolumns > 0)
	    strv[0] = "UNTITLED";
	closed_pm = pixmaps_list->manpage_section_20x20;
	closed_mask = pixmaps_list->manpage_section_20x20_mask;
	opened_pm = closed_pm;
	opened_mask = closed_mask;

	/* Get the selected node's EditorItem */
	item = EditorBranchGetData(
	    ctree,
	    node
	);
	if(item != NULL) 
	{
	    /* Insert by the EditorItem's type */
	    switch(item->type)
	    {
	      case EDITOR_ITEM_FILE:
		new_node = gtk_ctree_insert_node(
		    ctree,
		    node,			/* Parent */
		    NULL,			/* Append */
		    strv,
		    ME_LIST_ICON_TEXT_SPACING,
		    closed_pm, closed_mask,
		    opened_pm, opened_mask,
		    TRUE,			/* Is leaf */
		    FALSE			/* Not expanded */
		);
		break;

	      case EDITOR_ITEM_HEADER:
	      case EDITOR_ITEM_SECTION:
		new_node = gtk_ctree_insert_node(
		    ctree,
		    GTK_CTREE_NODE_PARENT(node),	/* Parent */
		    node,			/* Sibling */
		    strv,
		    ME_LIST_ICON_TEXT_SPACING,
		    closed_pm, closed_mask,
		    opened_pm, opened_mask,
		    TRUE,			/* Is leaf */
		    FALSE			/* Not expanded */
		);
		break;
	    } 
	}
	if(new_node != NULL)
	{
	    /* Create a new EditorItem for the new section */
	    item = EditorItemNewWithValues(
		EDITOR_ITEM_SECTION,
		editor,
		new_node
	    );
	    if(item != NULL)
	    {
		g_free(item->section_name);
		item->section_name = STRDUP(strv[0]);
	    }
	    EditorBranchSetData(
		ctree,
		new_node,
		item, EditorItemDestroyCB
	    );

	    /* Set the has changes flag on the new EditorItem */
	    EditorSetNodeHasChanges(
		editor,
		new_node,
		TRUE
	    );
	}

	/* Update the Layout GtkCTree's column width */
	gtk_clist_set_column_width(
	    clist,
	    0,					/* Column */
	    gtk_clist_optimal_column_width(clist, 0)
	);

	gtk_clist_thaw(clist);

	g_free(strv);

	EditorSetStatusMessage(editor, "New section added");

	EditorUpdateMenus(editor);

	editor->freeze_count--;
	EditorSetBusy(editor, FALSE);

	/* Select the new node */
	EditorSelect(
	    editor,
	    new_node
	);
}

/*
 *	Removes a header or section.
 */
void EditorManualRemoveSectionCB(GtkWidget *widget, gpointer data)
{
	gboolean yes_to_all = FALSE;
	gboolean need_break, need_continue;
	GtkCTree *ctree;
	GtkCTreeNode *branch;
	GtkWidget *toplevel;
	me_core_struct *core;
	editor_item_struct *item;
	editor_struct *editor = EDITOR(data);
	if(editor == NULL)
	    return;

	if(editor->freeze_count > 0)
	    return;

	toplevel = editor->toplevel;
	ctree = GTK_CTREE(editor->layout_ctree);
	core = editor->core;

	EditorSetBusy(editor, TRUE);
	EditorSetStatusProgress(editor, 0.0f);

	/* Get the selected node */
	branch = EditorCurrentSelectedNode(editor);
	while(branch != NULL)
	{
	    item = EditorBranchGetData(ctree, branch);
	    if(item == NULL)
	    {
		branch = NULL;
		continue;
	    }

	    /* Handle by type */
	    need_continue = FALSE;
	    need_break = FALSE;
	    switch(item->type)
	    {
	      case EDITOR_ITEM_HEADER:
		EditorSetStatusMessage(
		    editor,
		    "Removing header..."
		);
		if(!yes_to_all)
		{
		    gint status;

		    CDialogSetTransientFor(toplevel);
		    status = CDialogGetResponse(
"Confirm Remove Header",
"Are you sure you want to remove the header\n\
from the manual page?",
			NULL,
			CDIALOG_ICON_WARNING,
			CDIALOG_BTNFLAG_YES | CDIALOG_BTNFLAG_YES_TO_ALL |
			CDIALOG_BTNFLAG_NO | CDIALOG_BTNFLAG_CANCEL,
			CDIALOG_BTNFLAG_NO
		    );
		    CDialogSetTransientFor(NULL);
		    switch(status)
		    {
		      case CDIALOG_RESPONSE_YES_TO_ALL:
			yes_to_all = TRUE;
		      case CDIALOG_RESPONSE_YES:
			break;

		      case CDIALOG_RESPONSE_CANCEL:
		      case CDIALOG_RESPONSE_NOT_AVAILABLE:
			need_break = TRUE;
			break;

		      default:
			need_continue = TRUE;
			break;
		    }
		}
		break;

	      case EDITOR_ITEM_SECTION:
		EditorSetStatusMessage(
		    editor,
		    "Removing section..."
		);
		if(!yes_to_all)
		{
		    gint status;
		    gchar *msg = g_strdup_printf(
"Are you sure you want to remove section\n`%s' from the manual page?",
			item->section_name
		    );
		    CDialogSetTransientFor(toplevel);
		    status = CDialogGetResponse(
"Confirm Remove Section",
			msg,
			NULL,
			CDIALOG_ICON_WARNING,
			CDIALOG_BTNFLAG_YES | CDIALOG_BTNFLAG_YES_TO_ALL |
			CDIALOG_BTNFLAG_NO | CDIALOG_BTNFLAG_CANCEL,
			CDIALOG_BTNFLAG_NO
		    );
		    CDialogSetTransientFor(NULL);
		    g_free(msg);
		    switch(status)
		    {
		      case CDIALOG_RESPONSE_YES_TO_ALL:
			yes_to_all = TRUE;
		      case CDIALOG_RESPONSE_YES:
			break;

		      case CDIALOG_RESPONSE_CANCEL:
		      case CDIALOG_RESPONSE_NOT_AVAILABLE:
			need_break = TRUE;
			break;

		      default:
			need_continue = TRUE;
			break;
		    }
		}
		break;

	      default:
		/* Some other type, skip it */
		need_continue = TRUE;
		break;
	    }
	    if(need_break)
		break;
	    if(need_continue)
	    {
		branch = NULL;
		continue;
	    }

	    /* Remove this branch, the destroy notify callback will
	     * deallocate the branch item data and mark this branch
	     * as unselected.
	     */
	    gtk_ctree_remove_node(ctree, branch);

	    EditorSetStatusProgress(editor, 1.0);

	    branch = NULL;
	}


	/* Update layout clist column width */
	gtk_clist_set_column_width(
	    (GtkCList *)ctree,
	    0,			/* Column */
	    gtk_clist_optimal_column_width((GtkCList *)ctree, 0)
	);

	EditorSetStatusProgress(editor, 0.0);
	EditorSetStatusMessage(editor, "Remove done");

	/* Update menus on editor */
	EditorUpdateMenus(editor);

	EditorSetBusy(editor, FALSE);
}

/*
 *	Manual page properties callback.
 */
void EditorManualPropertiesCB(GtkWidget *widget, gpointer data)
{
	editor_struct *editor = EDITOR(data);
	if(editor == NULL)
	    return;

	if(editor->freeze_count > 0)
	    return;


}

/*
 *      Manual page preview callback.
 */
void EditorManualPreviewCB(GtkWidget *widget, gpointer data)
{
	gint i, viewer_num = -1, editor_num = -1;
	viewer_struct *viewer = NULL;
	me_core_struct *core;
	editor_struct *editor = EDITOR(data);
	if(editor == NULL)
	    return;

	if(editor->freeze_count > 0)
	    return;

	EditorSetBusy(editor, TRUE);

	core = editor->core;

	/* Get number of this editor on core structure */
	for(i = 0; i < core->total_editors; i++)
	{
	    if(editor == core->editor[i])
	    {
		editor_num = i;
		break;
	    }
	}

	/* Create a new viewer as needed */
	i = editor->viewer_num;
	if((i >= 0) && (i < core->total_viewers))
	{
	    viewer = core->viewer[i];
	    if(viewer != NULL)
		viewer_num = i;
	}

	/* If viewer is NULL then we need to create a new one */
	if(viewer == NULL)
	{
	    /* Create new viewer and add it to the core structure */
	    if(core->total_viewers < 0)
		core->total_viewers = 0;
	    i = core->total_viewers;
	    core->total_viewers = i + 1;
	    core->viewer = (viewer_struct **)g_realloc(
		core->viewer,
		core->total_viewers * sizeof(viewer_struct *)
	    );
	    if(core->viewer == NULL)
	    {
		core->total_viewers = 0;
		EditorSetBusy(editor, FALSE);
		return;
	    }

	    core->viewer[i] = viewer = ViewerNew(
		core,
		editor_num
	    );

	    if(viewer == NULL)
		viewer_num = -1;
	    else
		viewer_num = i;

	    /* Update viewer number on editor */
	    editor->viewer_num = viewer_num;
	}

	/* Viewer valid? */
	if(viewer != NULL)
	{
	    /* Map as needed */
	    if(!GTK_WIDGET_MAPPED(viewer->toplevel))
		ViewerMap(viewer);

	    EditorPreview(
		editor,
		EditorCurrentSelectedNode(editor),
		viewer_num
	    );
	}

	EditorSetBusy(editor, FALSE);
}

/*
 *	Syntax highlight toggle callback.
 */
void EditorManualSyntaxHighlightToggleCB(GtkWidget *widget, gpointer data)
{
	gboolean syntax_highlighting_state;
	GtkWidget *toplevel;
	CfgList *cfg_list;
	me_core_struct *core;
	editor_struct *editor = EDITOR(data);
	if((widget == NULL) || (editor == NULL))
	    return;

	if(editor->freeze_count > 0)
	    return;

	editor->freeze_count++;

	toplevel = editor->toplevel;
	core = editor->core;
	cfg_list = core->cfg_list;

	syntax_highlighting_state = ME_GET_B(ME_CFG_PARM_EDITOR_SYNTAX_HIGHLIGHTING);

	/* Toggle the syntax highlighting */
	syntax_highlighting_state = !syntax_highlighting_state;
	ME_SET_B(
	    ME_CFG_PARM_EDITOR_SYNTAX_HIGHLIGHTING,
	    syntax_highlighting_state
	);

	/* If turning on syntax highlighting then warn the user about
	 * problems with the GtkText widget
	 */
	if(syntax_highlighting_state)
	{
	    gchar	*prog_tmp_path = MECreateTemporaryDirectory(core),
			*msg = g_strdup_printf(
"This program makes use of the GTK+ GtkText widget which\n\
has known bugs that may cause applications to crash\n\
under certain circumstances (such as syntax highlighting).\n\
\n\
You are advised not to use syntax highlighting, however\n\
if you do and a crash does occure then this program will\n\
try its best to save any opened documents into the\n\
program's tempory directory:\n\
\n\
    %s",
		prog_tmp_path
	    );
	    g_free(prog_tmp_path);

	    CDialogSetTransientFor(toplevel);
	    CDialogGetResponse(
		"Syntax Highlighting Warning",
"Enabling syntax highlighting may cause crashes due to\n\
bugs in the GTK+ GtkText widget.\n\
\n\
Click on `help' for explaination.",
		msg,
		CDIALOG_ICON_WARNING,
		CDIALOG_BTNFLAG_OK | CDIALOG_BTNFLAG_HELP,
		CDIALOG_BTNFLAG_OK
	    );
	    g_free(msg);
	    CDialogSetTransientFor(NULL);
	}

	/* Notify about the change in the syntax highlighting */
	MEQueueEmitReconfigured(core);

	editor->freeze_count--;
}


/*
 *	Undo callback.
 */
void EditorUndoCB(GtkWidget *widget, gpointer data)
{
	gint status;
	GtkWidget *toplevel;
	editor_struct *editor = EDITOR(data);
	if(editor == NULL)
	    return;

	if(editor->freeze_count > 0)
	    return;

	/* Nothing to undo? */
	if(editor->total_undos <= 0)
	    return;

	toplevel = editor->toplevel;

	/* Mark that we're redoing */
	editor_processing_undo = editor;                                      
	EditorSetBusy(editor, TRUE);

	/* Perform undo procedure, passing the input list as the undo
	 * list and the redo list as the output list. The first item
	 * on the in list will be applied and an inverse operation
	 * will be placed into the output list.
	 */
	status = EditorUndoDoApply(
	    editor,
	    &editor->undo, &editor->total_undos,
	    &editor->redo, &editor->total_redos,
	    editor->max_undos
	);

	/* Flush and signals generated during undo apply */
	while(gtk_events_pending() > 0)
	    gtk_main_iteration();

	/* Mark as no longer undoing */
	editor_processing_undo = NULL;
	EditorSetBusy(editor, FALSE);

	/* Undo failed? */
	if(status)
	{
	    CDialogSetTransientFor(toplevel);
	    CDialogGetResponse(
		"Undo Failed",
"The undo process could not be performed.",
"Circumstances or an error prevented the undo\n\
operation from being performed.",
		CDIALOG_ICON_ERROR,
		CDIALOG_BTNFLAG_OK | CDIALOG_BTNFLAG_HELP,
		CDIALOG_BTNFLAG_OK
	    );
	    CDialogSetTransientFor(NULL);
	}

	EditorUpdateMenus(editor);
}

/*
 *	Redo callback.
 */
void EditorRedoCB(GtkWidget *widget, gpointer data)
{
	gint status;
	GtkWidget *toplevel;
	editor_struct *editor = EDITOR(data);
	if(editor == NULL)
	    return;

	if(editor->freeze_count > 0)
	    return;

	/* Nothing to redo? */
	if(editor->total_redos <= 0)
	    return;

	toplevel = editor->toplevel;

	/* Mark that we're redoing */
	editor_processing_undo = editor;
	EditorSetBusy(editor, TRUE);

	/* Perform redo procedure, passing the input list as the redo
	 * list and the undo list as the output list. The first item
	 * on the in list will be applied and an inverse operation
	 * will be placed into the output list.
	 */
	status = EditorUndoDoApply(
	    editor,
	    &editor->redo, &editor->total_redos,
	    &editor->undo, &editor->total_undos,
	    editor->max_undos
	);

	/* Flush and signals generated during undo apply */
	while(gtk_events_pending() > 0)
	    gtk_main_iteration();

	/* Mark as no longer undoing */
	editor_processing_undo = NULL;
	EditorSetBusy(editor, FALSE);

	/* Undo failed? */
	if(status)
	{
	    CDialogSetTransientFor(toplevel);
	    CDialogGetResponse(
		"Redo Failed",
"The redo process could not be performed.",
"Circumstances or an error prevented the redo\n\
operation from being performed.",
		CDIALOG_ICON_ERROR,
		CDIALOG_BTNFLAG_OK | CDIALOG_BTNFLAG_HELP,
		CDIALOG_BTNFLAG_OK
	    );
	    CDialogSetTransientFor(NULL);
	}

	EditorUpdateMenus(editor);
}


/*
 *	Cut.
 */
void EditorCutCB(GtkWidget *widget, gpointer data)
{
	gint start_pos, end_pos;
	GtkCTreeNode *sel_node;
	GtkEditable *editable;
	GtkText *text;
	editor_item_struct *item;
	me_core_struct *core;
	editor_struct *editor = EDITOR(data);
	if((widget == NULL) || (editor == NULL))
	    return;

	if(editor->freeze_count > 0)
	    return;

	core = editor->core;

	/* Get the selected EditorItem */
	sel_node = EditorCurrentSelectedNode(editor);
	item = EditorBranchGetData(
	    GTK_CTREE(editor->layout_ctree),
	    sel_node
	);
	if(item == NULL)
	    return;

	/* Get the GtkText based on the selected item type */
	text = NULL;
	switch(item->type)
	{
	  case EDITOR_ITEM_HEADER:
	    text = GTK_TEXT(editor->header_comments_text);
	    break;
	  case EDITOR_ITEM_SECTION:
	    text = GTK_TEXT(editor->section_text);
	    break;
	  case EDITOR_ITEM_FILE:
	    break;
	}           
	if(text == NULL)
	    return;

	editable = GTK_EDITABLE(text);
	start_pos = editable->selection_start_pos,
	end_pos = editable->selection_end_pos;
	if(end_pos < start_pos)
	{
	    const gint i = end_pos;
	    end_pos = start_pos;
	    start_pos = i;
	}

	/* Has selection and valid selection bounds? */
	if(editable->has_selection &&
	   (start_pos > -1) && (end_pos > -1) &&
	   ((end_pos - start_pos) > 0)
	)
	{
	    gtk_text_freeze(text);
	    gtk_editable_cut_clipboard(editable);
	    gtk_text_thaw(text);
	    EditorSetNodeHasChanges(
		editor, sel_node, TRUE
	    );
	}

	EditorUpdateMenus(editor);
	EditorSetStatusMessage(editor, "Selection cut");
}

/*
 *	Copy.
 */
void EditorCopyCB(GtkWidget *widget, gpointer data)
{
	gint start_pos, end_pos;
	GtkCTreeNode *sel_node;
	GtkEditable *editable;
	GtkText *text;
	editor_item_struct *item;
	me_core_struct *core;
	editor_struct *editor = EDITOR(data);
	if((widget == NULL) || (editor == NULL))
	    return;

	if(editor->freeze_count > 0)
	    return;

	core = editor->core;

	/* Get the selected EditorItem */
	sel_node = EditorCurrentSelectedNode(editor);
	item = EditorBranchGetData(
	    GTK_CTREE(editor->layout_ctree),
	    sel_node
	);
	if(item == NULL)
	    return;

	/* Get the GtkText based on the selected item type */
	text = NULL;
	switch(item->type)
	{
	  case EDITOR_ITEM_HEADER:
	    text = GTK_TEXT(editor->header_comments_text);
	    break;
	  case EDITOR_ITEM_SECTION:
	    text = GTK_TEXT(editor->section_text);
	    break;
	  case EDITOR_ITEM_FILE:
	    break;
	}           
	if(text == NULL)
	    return;

	editable = GTK_EDITABLE(text);
	start_pos = editable->selection_start_pos,
	end_pos = editable->selection_end_pos;
	if(end_pos < start_pos)
	{
	    const gint i = end_pos;
	    end_pos = start_pos;
	    start_pos = i;
	}

	/* Has selection and valid selection bounds? */
	if(editable->has_selection &&
	   (start_pos > -1) && (end_pos > -1) &&
	   ((end_pos - start_pos) > 0)
	)
	{
	    gtk_text_freeze(text);
	    gtk_editable_copy_clipboard(editable);
	    gtk_text_thaw(text);
	}

	EditorUpdateMenus(editor);
	EditorSetStatusMessage(editor, "Selection coppied");
}

/*
 *	Paste.
 */
void EditorPasteCB(GtkWidget *widget, gpointer data)
{
	GtkCTreeNode *sel_node;
	GtkEditable *editable;
	GtkText *text;
	editor_item_struct *item;
	me_core_struct *core;
	editor_struct *editor = EDITOR(data);
	if((widget == NULL) || (editor == NULL))
	    return;

	if(editor->freeze_count > 0)
	    return;

	core = editor->core;
	sel_node = EditorCurrentSelectedNode(editor);
	item = EditorBranchGetData(
	    GTK_CTREE(editor->layout_ctree),
	    sel_node
	);
	if(item == NULL)
	    return;

	/* Get the GtkText based on the selected item type */
	text = NULL;
	switch(item->type)
	{
	  case EDITOR_ITEM_HEADER:
	    text = GTK_TEXT(editor->header_comments_text);
	    break;
	  case EDITOR_ITEM_SECTION:
	    text = GTK_TEXT(editor->section_text);
	    break;
	  case EDITOR_ITEM_FILE:
	    break;
	}           
	if(text == NULL)
	    return;

	editable = GTK_EDITABLE(text);

	gtk_text_freeze(text);
	gtk_editable_paste_clipboard(editable);
	gtk_text_thaw(text);

	EditorSetNodeHasChanges(
	    editor, sel_node, TRUE
	);

	EditorUpdateMenus(editor);
	EditorSetStatusMessage(editor, "Paste done");
}

/*
 *	Delete Text.
 */
void EditorDeleteCB(GtkWidget *widget, gpointer data)
{

}

/*
 *	Select All.
 */
void EditorSelectAllCB(GtkWidget *widget, gpointer data)
{
	GtkEditable *editable = NULL;
	GtkText *text;
	editor_item_struct *item;
	editor_struct *editor = EDITOR(data);
	if((widget == NULL) || (editor == NULL))
	    return;

	if(editor->freeze_count > 0)
	    return;

	item = EditorCurrentSelectedItem(editor);
	if(item == NULL)
	    return;

	switch(item->type)
	{
	  case EDITOR_ITEM_HEADER:
	    editable = GTK_EDITABLE(editor->header_comments_text);
	    break;
	  case EDITOR_ITEM_SECTION:
	    editable = GTK_EDITABLE(editor->section_text);
	    break;
	  case EDITOR_ITEM_FILE:
	    break;
	}
	text = GTK_TEXT(editable);
	if(editable == NULL)
	    return;

	gtk_text_freeze(text);
	gtk_editable_select_region(editable, 0, -1);
	gtk_text_thaw(text);
}

/*
 *	Unselect All.
 */
void EditorUnselectAllCB(GtkWidget *widget, gpointer data)
{
	GtkEditable *editable = NULL;
	GtkText *text;
	editor_item_struct *item;
	editor_struct *editor = EDITOR(data);
	if((widget == NULL) || (editor == NULL))
	    return;
	
	if(editor->freeze_count > 0)
	    return;

	item = EditorCurrentSelectedItem(editor);
	if(item == NULL)
	    return;
	
	switch(item->type)
	{
	  case EDITOR_ITEM_HEADER: 
	    editable = GTK_EDITABLE(editor->header_comments_text);
	    break;
 	  case EDITOR_ITEM_SECTION:
	    editable = GTK_EDITABLE(editor->section_text);
	    break;
	  case EDITOR_ITEM_FILE:
	    break;
	}
	text = GTK_TEXT(editable);
	if(editable == NULL)
	    return;

	gtk_text_freeze(text);
	gtk_editable_select_region(editable, -1, -1);
	gtk_text_thaw(text);
}

/*
 *	Options callback.
 */
void EditorOptionsCB(GtkWidget *widget, gpointer data)
{
	GtkWidget	*toplevel,
			*cfg_win;
	me_core_struct *core;
	editor_struct *editor = EDITOR(data);
	if((widget == NULL) || (editor == NULL))
	    return;

	if(editor->freeze_count > 0)
	    return;

	toplevel = editor->toplevel;
	core = editor->core;

	cfg_win = core->options_window;
	if(cfg_win == NULL)
	    core->options_window = cfg_win = MEOptionsWinNew(core);

	if(GTK_WIDGET_MAPPED(cfg_win))
	{
	    gtk_widget_hide(cfg_win);
	    gtk_widget_show(cfg_win);
	    return;
	}

	CfgWinGetValues(cfg_win, FALSE);
	CfgWinSetHasChanges(cfg_win, FALSE);
	MEOptionsWinSelectPage(
	    cfg_win,
	    "Appearance"
	);
	CfgWinUpdateMenus(cfg_win);

	if(cfg_win->window != NULL)
	{
	    const gint	x_offset = 0,
			y_offset = 0;
	    gint	x, y,
			width1, height1,
			width2, height2;
	    GdkWindow	*window1 = toplevel->window,
			*window2 = cfg_win->window;

	    gdk_window_get_root_origin(window1, &x, &y);
	    gdk_window_get_size(window1, &width1, &height1);
	    gdk_window_get_size(window2, &width2, &height2);
	    gtk_widget_set_uposition(
		cfg_win,
		((width2 > 0) ?
		    (x + (width1 / 2) - (width2 / 2)) : (x + 20)
		) + x_offset,
		((height2 > 0) ?
		    (y + (height1 / 2) - (height2 / 2)) : (y + 20)
		) + y_offset
	    );
	}

	gtk_widget_show(cfg_win);
}


/* Macro to set the GtkText *text variable to the appropriate
 * value. Uses variables editor, branch, item, and text.
 */
#define EDITOR_GET_TEXT_WIDGET	{		\
 text = NULL;					\
 switch(item->type) {				\
  case EDITOR_ITEM_HEADER:			\
   break;					\
  case EDITOR_ITEM_SECTION:			\
   text = GTK_TEXT(editor->section_text);	\
   break;					\
  case EDITOR_ITEM_FILE:			\
   break;					\
 }						\
}

/* Pre code macro for editor fmt callbacks, sets up variables;
 * font, text, editable, ctree, branch, item, insert_pos,
 * end_pos, core, and editor.
 */
#define EDITOR_FMT_CB_PRE				\
	int insert_pos, end_pos;			\
	GdkFont *font;					\
	GtkWidget *toplevel;				\
	GtkEditable *editable;				\
	GtkText *text;					\
	GtkCTree *ctree;				\
	GtkCTreeNode *branch;				\
	editor_item_struct *item;			\
	me_core_struct *core;				\
	editor_struct *editor = EDITOR(data);		\
	if(editor == NULL)				\
	    return;					\
							\
	if(editor->freeze_count > 0)			\
	    return;					\
							\
	core = editor->core;				\
	toplevel = editor->toplevel;			\
	ctree = GTK_CTREE(editor->layout_ctree);	\
	branch = EditorCurrentSelectedNode(editor);	\
	if((core == NULL) || (ctree == NULL) || (branch == NULL)) \
	    return;					\
							\
	font = core->editable_font;			\
							\
	item = EditorBranchGetData(ctree, branch);	\
	if(item == NULL)				\
	    return;					\
							\
	EDITOR_GET_TEXT_WIDGET				\
	if(text == NULL)				\
	    return;					\
							\
	editable = GTK_EDITABLE(text);			\
							\
	if(editable->has_selection)			\
	{						\
	    insert_pos = (gint)editable->selection_start_pos; \
	    end_pos = (gint)editable->selection_end_pos; \
	    if(insert_pos > end_pos)			\
	    {						\
		int tmp_pos = end_pos;			\
		end_pos = insert_pos;			\
		insert_pos = tmp_pos;			\
	    }						\
	}						\
	else						\
	{						\
	    insert_pos = (gint)editable->current_pos;	\
	    end_pos = -1;				\
	}						\
							\

/* Post code macro for editor fmt callbacks, uses variables;
 * insert_pos, end_pos, new_str, new_str_len, editable, and editor.
 */
#define EDITOR_FMT_CB_POST	\
{ \
 gint position; \
 \
 position = insert_pos; \
 /* Need to call insert callback, notifying of changes, this will \
  * record an undo and mark has changes on the selected branch. \
  */ \
 EditorTextInsertCB( \
  editable, new_str, new_str_len, \
  &position, editor \
 ); \
 gtk_editable_set_position(editable, insert_pos + new_str_len); \
 gtk_text_set_point(GTK_TEXT(editable), insert_pos + new_str_len); \
 gtk_widget_grab_focus(GTK_WIDGET(editable)); \
}

/*
 *	Paragraph left callback.
 */
void EditorFmtParagraphLeftCB(GtkWidget *widget, gpointer data)
{
	const char *new_str = "<LP>";
	int new_str_len;
	EDITOR_FMT_CB_PRE

	new_str_len = strlen(new_str);

	gtk_text_set_point(text, insert_pos);
	gtk_text_freeze(text);
	gtk_text_insert(
	    text,
	    font,
	    NULL,
	    NULL,
	    new_str, new_str_len
	);
	gtk_text_thaw(text);

	EDITOR_FMT_CB_POST
}

/*
 *	Paragraph right callback.
 */
void EditorFmtParagraphRightCB(GtkWidget *widget, gpointer data)
{
	const char *new_str = "<RP>";
	int new_str_len;
	EDITOR_FMT_CB_PRE

	new_str_len = strlen(new_str);

	gtk_text_set_point(text, insert_pos);
	gtk_text_freeze(text);
	gtk_text_insert(
	    text,
	    font,
	    NULL,
	    NULL,
	    new_str, new_str_len
	);
	gtk_text_thaw(text);

	EDITOR_FMT_CB_POST
}

/*
 *	Paragraph indented callback.
 */
void EditorFmtParagraphIndentedCB(GtkWidget *widget, gpointer data)
{
	const char *new_str = "<IP>";
	int new_str_len;
	EDITOR_FMT_CB_PRE

	new_str_len = strlen(new_str);

	gtk_text_set_point(text, insert_pos);
	gtk_text_freeze(text);
	gtk_text_insert(
	    text, font, NULL, NULL,
	    new_str, new_str_len
	);
	gtk_text_thaw(text);

	EDITOR_FMT_CB_POST
}

/*
 *      List item paragraph callback.
 */
void EditorFmtListItemCB(GtkWidget *widget, gpointer data)
{
	const gchar *new_str = "<TP>";
	gint new_str_len;
	EDITOR_FMT_CB_PRE

	new_str_len = strlen(new_str);

	gtk_text_set_point(text, insert_pos);
	gtk_text_freeze(text);
	gtk_text_insert(
	    text, font, NULL, NULL,
	    new_str, new_str_len
	);
	gtk_text_thaw(text);

	EDITOR_FMT_CB_POST
}

/*
 *	Bold callback.
 */
void EditorFmtBoldCB(GtkWidget *widget, gpointer data)
{
	const gchar *new_str = "<fB>";
	gint new_str_len;
	EDITOR_FMT_CB_PRE
 
	new_str_len = strlen(new_str);
  
	gtk_text_set_point(text, insert_pos);
	gtk_text_freeze(text);
	gtk_text_insert(
	    text, font, NULL, NULL,
	    new_str, new_str_len
	);
	gtk_text_thaw(text);

	EDITOR_FMT_CB_POST

	/* Ending tag */
	new_str = "<fR>";
	if(end_pos < 0)
	{
	    insert_pos = insert_pos + new_str_len;
	    new_str_len = strlen(new_str);

	    gtk_editable_set_position(editable, insert_pos);
	    gtk_text_freeze(text);
	    gtk_text_insert(
		text, font, NULL, NULL,
		new_str, new_str_len
	    );
	    gtk_text_thaw(text);
	}
	else
	{
	    int new_end_pos = end_pos + new_str_len;

	    insert_pos = end_pos + new_str_len;
	    end_pos = -1;
	    new_str_len = strlen(new_str);
	    new_end_pos += new_str_len;
 
	    gtk_text_set_point(text, insert_pos);
	    gtk_text_freeze(text);
	    gtk_text_insert(
		text, font, NULL, NULL,
		new_str, new_str_len
	    );
	    gtk_text_thaw(text);
	}

	EDITOR_FMT_CB_POST
}

/*
 *      Underline callback.
 */
void EditorFmtUnderlineCB(GtkWidget *widget, gpointer data)
{
	const char *new_str = "<fI>";
	int new_str_len;
	EDITOR_FMT_CB_PRE

	new_str_len = strlen(new_str);

	gtk_text_set_point(text, insert_pos);
	gtk_text_freeze(text);
	gtk_text_insert(
	    text, font, NULL, NULL,
	    new_str, new_str_len
	);
	gtk_text_thaw(text);
 
	EDITOR_FMT_CB_POST

	/* Ending tag */
	new_str = "<fR>";
	if(end_pos < 0)
	{
	    insert_pos = insert_pos + new_str_len;
	    new_str_len = strlen(new_str);

	    gtk_editable_set_position(editable, insert_pos);              
	    gtk_text_freeze(text);
	    gtk_text_insert(
		text, font, NULL, NULL,
		new_str, new_str_len
	    ); 
	    gtk_text_thaw(text);
	}
	else
	{
	    int new_end_pos = end_pos + new_str_len;

	    insert_pos = end_pos + new_str_len;
	    end_pos = -1;
	    new_str_len = strlen(new_str); 
	    new_end_pos += new_str_len;

	    gtk_text_set_point(text, insert_pos);
	    gtk_text_freeze(text);
	    gtk_text_insert(
		text, font, NULL, NULL,
		new_str, new_str_len  
	    );
	    gtk_text_thaw(text);
	}

	EDITOR_FMT_CB_POST
}

/*
 *	Line break callback.
 */
void EditorFmtLineBreakCB(GtkWidget *widget, gpointer data)
{
	const char *new_str = "<br>";
	int new_str_len;
	EDITOR_FMT_CB_PRE

	new_str_len = strlen(new_str);

	gtk_text_set_point(text, insert_pos);
	gtk_text_freeze(text);
	gtk_text_insert(
	    text, font, NULL, NULL,
	    new_str, new_str_len
	);
	gtk_text_thaw(text);

	EDITOR_FMT_CB_POST
}

/*
 *	Ampersand callback.
 */
void EditorFmtAmpersandCB(GtkWidget *widget, gpointer data)
{
	const char *new_str = "&amp;";
	int new_str_len;
	EDITOR_FMT_CB_PRE

	new_str_len = strlen(new_str);

	gtk_text_set_point(text, insert_pos);
	gtk_text_freeze(text);
	gtk_text_insert(
	    text, font, NULL, NULL,
	    new_str, new_str_len
	);
	gtk_text_thaw(text);

	EDITOR_FMT_CB_POST
}

/*
 *      Less than callback.
 */
void EditorFmtLessThanCB(GtkWidget *widget, gpointer data)
{
	const char *new_str = "&lt;";
	int new_str_len;
	EDITOR_FMT_CB_PRE

	new_str_len = strlen(new_str);

	gtk_text_set_point(text, insert_pos);
	gtk_text_freeze(text);
	gtk_text_insert(
	    text, font, NULL, NULL,   
	    new_str, new_str_len
	);
	gtk_text_thaw(text);

	EDITOR_FMT_CB_POST
}

/*
 *	Greater than callback.
 */
void EditorFmtGreaterThanCB(GtkWidget *widget, gpointer data)
{
	const char *new_str = "&gt;";
	int new_str_len;
	EDITOR_FMT_CB_PRE

	new_str_len = strlen(new_str);

	gtk_text_set_point(text, insert_pos);
	gtk_text_freeze(text);
	gtk_text_insert(
	    text, font, NULL, NULL,
	    new_str, new_str_len
	);
	gtk_text_thaw(text);

	EDITOR_FMT_CB_POST
}

/*
 *	Strip tags from selected text callback.
 */
void EditorStripTagsCB(GtkWidget *widget, gpointer data)
{
	gint i, j, c, len, made_strip, strip_loss = 0;
	EDITOR_FMT_CB_PRE

	/* No selection? */
	if(end_pos < 0)
	    return;

	len = end_pos - insert_pos;
	if(len <= 0)
	{
	    CDialogSetTransientFor(toplevel);
	    CDialogGetResponse(
"No Text Selected",
"No text selected to strip tags from.",
"You have not selected any text in which to strip\n\
tags from. If you wish to strip tags from the\n\
entire section then you must first go to\n\
edit->select all and then strip tags otherwise\n\
use the pointer to select the section of text you\n\
want to strip tags from and then strip tags.",
		CDIALOG_ICON_WARNING,
		CDIALOG_BTNFLAG_OK | CDIALOG_BTNFLAG_HELP,
		CDIALOG_BTNFLAG_OK
	    );
	    CDialogSetTransientFor(NULL);
	    return;
	}

	do
	{
	    made_strip = 0;
	    end_pos = MIN(end_pos, gtk_text_get_length(text));

	    /* Iterate from insert position to ending position */
	    for(i = insert_pos, len = 0; i < end_pos; i++)
	    {
		c = GTK_TEXT_INDEX(text, i);

		/* Is character a tag deliminator start? */
		if(c == '<')
		{
		    len++;	/* Count deliminator start char */

		    /* Seek to ending tag deliminator (if any) */
		    for(j = i + 1; j < end_pos; j++)
		    {
			len++;	/* Count this char regardless */

			/* Begin calculating length */
			c = GTK_TEXT_INDEX(text, j);
			if(c == '>')
			    break;
		    }
		    /* No ending tag found? */
		    if(j >= end_pos)
			len = 0;

		    break;
		}
	    }

	    /* Got a tag? */
	    if((i < end_pos) && (len > 0))
	    {
		/* Call text delete callback before deleting */
		EditorTextDeleteCB(
		    editable, i, i + len, editor
		);

		/* Delete characters specified by len */
		gtk_text_freeze(text);
		gtk_text_set_point(text, i);
		gtk_text_forward_delete(text, len);
		gtk_text_thaw(text);

		made_strip = 1;
		strip_loss += len;
	    }

	    /* Need to reduce end position */
	    end_pos -= len;

	} while(made_strip);


	EditorUpdateMenus(editor);
	EditorSetStatusMessage(editor, "Strip tags done");
}

#undef EDITOR_FMT_CB_POST
#undef EDITOR_FMT_CB_PRE
#undef EDITOR_GET_TEXT_WIDGET


/*
 *	Maps the Find In Pages dialog.
 */
void EditorMapFIPCB(GtkWidget *widget, gpointer data)
{
	GtkWidget *toplevel;
	editor_fip_struct *d;
	me_core_struct *core;
	editor_struct *editor = EDITOR(data);
	if(editor == NULL)
	    return;

	if(editor->freeze_count > 0)
	    return;

	EditorSetBusy(editor, TRUE);
	editor->freeze_count++;

	toplevel = editor->toplevel;
	core = editor->core;

	/* Get/create the Find In Pages dialog */
	d = editor->fip_dlg;
	if(d == NULL)
	    editor->fip_dlg = d = EditorFIPNew(
		core,
		editor
	    );
	if(d == NULL)
	{
	    editor->freeze_count--;
	    EditorSetBusy(editor, FALSE);
	    return;
	}

	/* If the dialog is already mapped then just remap it */
	if(EditorFIPIsMapped(d))
	{
	    EditorFIPUnmap(d);
	    EditorFIPMap(d);
	    editor->freeze_count--;
	    EditorSetBusy(editor, FALSE);
	    return;
	}

	/* Center the dialog to the Editor's toplevel */
	if(d->toplevel->window != NULL)
	{
	    const gint  x_offset = 0,
			y_offset = 0;
	    gint        x, y,
			width1, height1,
			width2, height2;
	    GdkWindow   *window1 = toplevel->window,
			*window2 = d->toplevel->window;

	    gdk_window_get_root_origin(window1, &x, &y);
	    gdk_window_get_size(window1, &width1, &height1);
	    gdk_window_get_size(window2, &width2, &height2);
	    gtk_widget_set_uposition(
		d->toplevel,
		((width2 > 0) ?
		    (x + (width1 / 2) - (width2 / 2)) : (x + 20)
		) + x_offset,
		((height2 > 0) ?
		    (y + (height1 / 2) - (height2 / 2)) : (y + 20)
		) + y_offset
	    );
	}

	/* Map the Find In Pages dialog */
	EditorFIPMap(d);

	editor->freeze_count--;
	EditorSetBusy(editor, FALSE);
}

/*
 *	Find GtkCombo "activate" signal callback.
 */
void EditorFindComboActivateCB(GtkWidget *widget, gpointer data)
{
	editor_struct *editor = EDITOR(data);
	if((widget == NULL) || (editor == NULL))                    
	    return;

	if(editor->freeze_count > 0)
	    return;

	EditorFindCBNexus(
	    editor,
	    FALSE				/* Not case sensitive */
	);

}

/*
 *	Replace all callback.
 *
 *	Note this function is a bit different than the other find and
 *	replace callback functions, it does all the setting up of values
 *	and calling functions by itself.
 */
void EditorReplaceAllCB(GtkWidget *widget, gpointer data)
{
	gboolean case_sensitive = FALSE;
	gint noccurances_replaced = 0;
	gchar		*needle,
			*replacement;
	GtkWidget *toplevel;
	GtkText *text = NULL;
	GtkEditable *editable = NULL;
	GtkCombo	*find_combo,
			*replace_combo;
	GtkCTree *ctree;
	GtkCTreeNode *sel_node;
	editor_item_struct *item;
	editor_struct *editor = EDITOR(data);
	if((widget == NULL) || (editor == NULL))
	    return;

	if(editor->freeze_count > 0)
	    return;

	toplevel = editor->toplevel;
	ctree = GTK_CTREE(editor->layout_ctree);
	sel_node = EditorCurrentSelectedNode(editor);
	find_combo = GTK_COMBO(editor->find_combo);
	replace_combo = GTK_COMBO(editor->replace_combo);

	EditorSetBusy(editor, TRUE);

	/* Get the selected EditorItem */
	item = EditorBranchGetData(
	    ctree,
	    sel_node
	);
	if(item == NULL)
	{
	    CDialogSetTransientFor(toplevel);
	    CDialogGetResponse(
"No Item Selected",
"No item selected to search and replace through.",
"You have not selected an item with text\n\
to search and replace through.",
		CDIALOG_ICON_WARNING,
		CDIALOG_BTNFLAG_OK | CDIALOG_BTNFLAG_HELP,
		CDIALOG_BTNFLAG_OK
	    );
	    CDialogSetTransientFor(NULL);
	    EditorSetBusy(editor, FALSE);
	    return;
	}

	/* Get the needle */
	needle = STRDUP(gtk_entry_get_text(GTK_ENTRY(find_combo->entry)));

	/* Get the replacement */
	replacement = STRDUP(gtk_entry_get_text(GTK_ENTRY(replace_combo->entry)));

	/* Update values to combo lists */
	GUIComboAddItem(GTK_WIDGET(find_combo), needle);
	GUIComboAddItem(GTK_WIDGET(replace_combo), replacement);

	/* Handle by the EditorItem's type */
	switch(item->type)
	{
	  case EDITOR_ITEM_HEADER:
	    text = GTK_TEXT(editor->header_comments_text);
	    editable = GTK_EDITABLE(text);

	    gtk_widget_set_sensitive(GTK_WIDGET(text), FALSE);

	    /* Replace all */
	    noccurances_replaced += EditorReplaceAll(
		editor,
		text,
		needle,
		replacement,
		case_sensitive
	    );

	    gtk_widget_set_sensitive(GTK_WIDGET(text), TRUE);

	    /* Did we replace any occurances? */
	    if(noccurances_replaced > 0)
	    {
		gchar *msg = g_strdup_printf(
		    "Replaced %i %s",
		    noccurances_replaced,
		    (noccurances_replaced == 1) ? "occurance" : "occurances"
		);
		EditorSetStatusMessage(editor, msg);
		g_free(msg);
	    }
	    else
	    {
		CDialogSetTransientFor(toplevel);
		CDialogGetResponse(
"String Not Found",
"The string was not found.",
		    NULL,
		    CDIALOG_ICON_SEARCH,
		    CDIALOG_BTNFLAG_OK,
		    CDIALOG_BTNFLAG_OK
		);
		CDialogSetTransientFor(NULL);
		EditorSetStatusMessage(editor, "Replace done");
	    }
	    break;

	  case EDITOR_ITEM_SECTION:
	    text = GTK_TEXT(editor->section_text);
	    editable = GTK_EDITABLE(text);

	    gtk_widget_set_sensitive(GTK_WIDGET(text), FALSE);

	    /* Replace all */
	    noccurances_replaced += EditorReplaceAll(
		editor,
		text,
		needle,
		replacement,
		case_sensitive
	    );

	    gtk_widget_set_sensitive(GTK_WIDGET(text), TRUE);

	    /* Did we replace any occurances? */
	    if(noccurances_replaced > 0)
	    {
		gchar *msg = g_strdup_printf(
		    "Replaced %i %s",
		    noccurances_replaced,
		    (noccurances_replaced == 1) ? "occurance" : "occurances"
		);
		EditorSetStatusMessage(editor, msg);
		g_free(msg);
	    }
	    else
	    {
		CDialogSetTransientFor(toplevel);
		CDialogGetResponse(
"String Not Found",
"The string was not found.",
		    NULL,
		    CDIALOG_ICON_SEARCH,
		    CDIALOG_BTNFLAG_OK,
		    CDIALOG_BTNFLAG_OK
		);
		CDialogSetTransientFor(NULL);
		EditorSetStatusMessage(editor, "Replace done");
	    }
	    break;

	  case EDITOR_ITEM_FILE:
	    break;
	}

	/* Set the selected EditorItem's has changes flag */
	EditorSetNodeHasChanges(
	    editor,
	    sel_node,
	    TRUE
	);

	g_free(replacement);
	g_free(needle);

	EditorUpdateMenus(editor);
	EditorSetBusy(editor, FALSE);
}

/*
 *	Replace Entire Page callback.
 */
void EditorReplaceEntirePageCB(GtkWidget *widget, gpointer data)
{
	gboolean case_sensitive = FALSE;
	gint		response,
			noccurances_replaced = 0;
	gchar		*needle,
			*replacement;
	GtkWidget *toplevel;
	GtkCombo	*find_combo,
			*replace_combo;
	GtkCTree *ctree;
	GtkCTreeNode *sel_node;
	editor_struct *editor = EDITOR(data);
	if((widget == NULL) || (editor == NULL))
	    return;

	if(editor->freeze_count > 0)
	    return;

	EditorSetBusy(editor, TRUE);

	toplevel = editor->toplevel;
	ctree = GTK_CTREE(editor->layout_ctree);
	sel_node = EditorCurrentSelectedNode(editor);
	find_combo = GTK_COMBO(editor->find_combo);
	replace_combo = GTK_COMBO(editor->replace_combo);

	/* Confirm before replace entire page */
	CDialogSetTransientFor(toplevel);
	response = CDialogGetResponse(
"Confirm Replace Entire Page",
"You will not be able to undo this operation,\n\
are you sure you want to find and replace the\n\
entire selected Manual Page?",
	    NULL,
	    CDIALOG_ICON_WARNING,
	    CDIALOG_BTNFLAG_YES | CDIALOG_BTNFLAG_NO,
	    CDIALOG_BTNFLAG_NO
	);
	CDialogSetTransientFor(NULL);
	switch(response)
	{
	  case CDIALOG_RESPONSE_YES_TO_ALL:
	  case CDIALOG_RESPONSE_YES:
	    break;

	  default:
	    EditorSetBusy(editor, FALSE);
	    return;
	    break;
	}

	/* Get the selected node's toplevel node */
	sel_node = EditorGetToplevelNode(sel_node);
	if(sel_node == NULL)
	{
	    CDialogSetTransientFor(toplevel);
	    CDialogGetResponse(
"No Manual Page Selected",
"No manual page selected to search and replace through.",
		NULL,
		CDIALOG_ICON_WARNING,
		CDIALOG_BTNFLAG_OK,
		CDIALOG_BTNFLAG_OK
	    );
	    CDialogSetTransientFor(NULL);
	    EditorSetBusy(editor, FALSE);
	    return;
	}

	/* Get the needle */
	needle = STRDUP(gtk_entry_get_text(GTK_ENTRY(find_combo->entry)));

	/* Get the replacement */
	replacement = STRDUP(gtk_entry_get_text(GTK_ENTRY(replace_combo->entry)));

	/* Update the values to the GtkCombo lists */
	GUIComboAddItem(GTK_WIDGET(find_combo), needle);
	GUIComboAddItem(GTK_WIDGET(replace_combo), replacement);                

	/* Replace entire page */
	noccurances_replaced += EditorReplaceEntirePageCBIterate(
	    editor,
	    sel_node,
	    needle,
	    replacement,
	    case_sensitive
	);
	if(noccurances_replaced > 0)
	{
	    gchar *msg = g_strdup_printf(
		"Replaced %i %s",
		noccurances_replaced,
		(noccurances_replaced == 1) ? "occurance" : "occurances"
	    );
	    EditorSetStatusMessage(editor, msg);
	    g_free(msg);
	}
	else
	{
	    EditorSetStatusMessage(
		editor,
		"No occurances found"
	    );
	}

	/* Set the selected EditorItem's has changes flag */
	EditorSetNodeHasChanges(
	    editor,
	    sel_node,
	    TRUE
	);

	g_free(replacement);
	g_free(needle);

	EditorUpdateMenus(editor);
	EditorSetBusy(editor, FALSE);
}

/*
 *	Called by EditorReplaceEntirePageCB() to replace the entire page.
 *
 *	The editor's selected branch will be updated then force selected
 *	to the specified branch.
 */
static gint EditorReplaceEntirePageCBIterate(
	editor_struct *editor,
	GtkCTreeNode *sel_node,
	gchar *needle,
	const gchar *replacement,
	const gboolean case_sensitive
)
{
	gint noccurances_replaced = 0;
	GtkEditable *editable = NULL;
	GtkText *text = NULL;
	GtkCTree *ctree = GTK_CTREE(editor->layout_ctree);
	GtkCTreeNode *node;
	editor_item_struct *item;

	if(sel_node == NULL)
	    return(noccurances_replaced);

	/* Do not set the Editor as processing or frozen, otherwise
	 * we will not be able to apply and reget the values
	 */

	/* Force select the specified EditorItem */
	EditorSelect(
	    editor,
	    sel_node
	);

	/* Get the selected EditorItem */
	item = EditorBranchGetData(
	    ctree,
	    sel_node
	);
	if(item == NULL)
	    return(noccurances_replaced);

	switch(item->type)
	{
	  case EDITOR_ITEM_HEADER:
	    text = GTK_TEXT(editor->header_comments_text);
	    editable = GTK_EDITABLE(text);

	    gtk_widget_set_sensitive(GTK_WIDGET(text), FALSE);

	    /* Replace all */   
	    noccurances_replaced += EditorReplaceAll(
		editor,
		text,
		needle,
		replacement,  
		case_sensitive
	    );

	    gtk_widget_set_sensitive(GTK_WIDGET(text), TRUE);
	    break;

	  case EDITOR_ITEM_SECTION:
	    text = GTK_TEXT(editor->section_text);
	    editable = GTK_EDITABLE(text);

	    gtk_widget_set_sensitive(GTK_WIDGET(text), FALSE);

	    /* Replace all */
	    noccurances_replaced += EditorReplaceAll(
		editor,
		text,
		needle,
		replacement,
		case_sensitive
	    );

	    gtk_widget_set_sensitive(GTK_WIDGET(text), TRUE);
	    break;

	  case EDITOR_ITEM_FILE:
	    break;
	}

	/* Iterate through each child GtkCTreeNode */
	node = GTK_CTREE_NODE_CHILD(sel_node);
	while(node != NULL)
	{
	    noccurances_replaced += EditorReplaceEntirePageCBIterate(
		editor,
		node,
		needle,
		replacement,
		case_sensitive
	    );

	    node = GTK_CTREE_NODE_SIBLING(node);
	}

	return(noccurances_replaced);
}


/*
 *	Replace GtkCombo "activate" signal callback.
 */
void EditorReplaceComboActivateCB(GtkWidget *widget, gpointer data)
{
	gboolean case_sensitive = FALSE;
	gchar *s;
	GtkCombo *combo;
	editor_struct *editor = EDITOR(data);
	if((widget == NULL) || (editor == NULL))
	    return;

	if(editor->freeze_count > 0)
	    return;

	combo = GTK_COMBO(widget);

	/* Get the current replace value */
	s = STRDUP(gtk_entry_get_text(GTK_ENTRY(combo->entry)));

	/* Record the replace value on the GtkCombo's history */
	GUIComboAddItem(GTK_WIDGET(combo), s);

	g_free(s);

	EditorReplaceCBNexus(
	    editor,
	    case_sensitive
	);
}


/*
 *	Find nexus.
 */
static void EditorFindCBNexus(
	editor_struct *editor,
	const gboolean case_sensitive
)
{
	gboolean	got_match,
			search_wrapped;
	gint start_pos;
	gchar		*haystack = NULL,
			*needle = NULL;
	GtkWidget *toplevel = editor->toplevel;
	GtkCombo *find_combo = GTK_COMBO(editor->find_combo);
	GtkText *text;
	GtkEditable *editable;
	editor_item_struct *item;

	EditorSetBusy(editor, TRUE);

	/* Get the selected EditorItem */
	item = EditorCurrentSelectedItem(editor);
	if(item == NULL)
	{
	    CDialogSetTransientFor(toplevel);
	    CDialogGetResponse(
"No Item Selected",
"No item selected to search through.",
"You have not selected an item with text\n\
to search through.",
		CDIALOG_ICON_WARNING,
		CDIALOG_BTNFLAG_OK | CDIALOG_BTNFLAG_HELP,
		CDIALOG_BTNFLAG_OK
	    );
	    CDialogSetTransientFor(NULL);
	    EditorSetBusy(editor, FALSE);
	    return;
	}

	/* Get/copy the needle */
	needle = STRDUP(gtk_entry_get_text(GTK_ENTRY(find_combo->entry)));

	/* Handle by the EditorItem's type */
	switch(item->type)
	{
	  case EDITOR_ITEM_HEADER:
	    text = GTK_TEXT(editor->header_comments_text);
	    editable = GTK_EDITABLE(text);

	    /* Get the haystack from the header's text */
	    haystack = gtk_editable_get_chars(editable, 0, -1);

	    /* Set the starting position plus one */
	    start_pos = gtk_editable_get_position(editable) + 1;

	    /* Search */
	    got_match = EditorFind(
		editor,
		text,
		haystack,
		needle,
		(gint)gtk_text_get_length(text),
		start_pos,
		case_sensitive,
		TRUE,				/* Scroll to matched position */
		&search_wrapped
	    );
	    /* Did we get a match? */
	    if(got_match)
	    {
		/* Search wrapped? */
		if(search_wrapped)
		{
		    CDialogSetTransientFor(toplevel);
		    CDialogGetResponse(
"Search Wrapped",
"Search wrapped.",
			NULL,
			CDIALOG_ICON_INFO,
			CDIALOG_BTNFLAG_OK,
			CDIALOG_BTNFLAG_OK
		    );
		    CDialogSetTransientFor(NULL);
		}
	    }
	    else
	    {
		CDialogSetTransientFor(toplevel);
		CDialogGetResponse(
"String Not Found",
"The string was not found.",
		    NULL,
		    CDIALOG_ICON_SEARCH,
		    CDIALOG_BTNFLAG_OK,
		    CDIALOG_BTNFLAG_OK
		);
		CDialogSetTransientFor(NULL);
	    }
	    break;

	  case EDITOR_ITEM_SECTION:
	    text = GTK_TEXT(editor->section_text);
	    editable = GTK_EDITABLE(text);

	    /* Get the haystack from the section text */
	    haystack = gtk_editable_get_chars(editable, 0, -1);

	    /* Set the starting position plus one */
	    start_pos = gtk_editable_get_position(editable) + 1;

	    /* Search */
	    got_match = EditorFind(
		editor,
		text,
		haystack,
		needle,
		(gint)gtk_text_get_length(text),
		start_pos,
		case_sensitive,
		TRUE,				/* Scroll to matched position */
		&search_wrapped
	    );
	    /* Did we get a match? */
	    if(got_match)
	    {
		/* Search wrapped? */
		if(search_wrapped)
		{
		    CDialogSetTransientFor(toplevel);
		    CDialogGetResponse(
"Search Wrapped",
"Search wrapped.",
			NULL,
			CDIALOG_ICON_INFO,
			CDIALOG_BTNFLAG_OK,
			CDIALOG_BTNFLAG_OK
		    );
		    CDialogSetTransientFor(NULL);
		}
	    }
	    else
	    {
		CDialogSetTransientFor(toplevel);
		CDialogGetResponse(  
"String Not Found",
"The string was not found.",
		    NULL,
		    CDIALOG_ICON_SEARCH,
		    CDIALOG_BTNFLAG_OK,
		    CDIALOG_BTNFLAG_OK
		);
		CDialogSetTransientFor(NULL);
	    }
	    break;

	  case EDITOR_ITEM_FILE:
	    break;
	}

	g_free(haystack);
	g_free(needle);

	EditorUpdateMenus(editor);
	EditorSetBusy(editor, FALSE);
}

/*
 *	Replace nexus.
 */
static void EditorReplaceCBNexus(
	editor_struct *editor,
	const gboolean case_sensitive
)
{
	gboolean made_replace;
	GtkWidget *toplevel = editor->toplevel;
	GtkCombo	*find_combo = GTK_COMBO(editor->find_combo),
			*replace_combo = GTK_COMBO(editor->replace_combo);
	GtkEditable *editable;
	GtkText *text;
	GtkCTree *ctree = GTK_CTREE(editor->layout_ctree);
	GtkCTreeNode *sel_node = EditorCurrentSelectedNode(editor);
	gint start_pos;
	gchar		*haystack = NULL,
			*needle = NULL,
			*replacement = NULL;
	editor_item_struct *item;

	EditorSetBusy(editor, TRUE);

	/* Get the selected EditorItem */
	item = EditorBranchGetData(
	    ctree,
	    sel_node
	);
	if(item == NULL)
	{
	    CDialogSetTransientFor(toplevel);
	    CDialogGetResponse(
"No Item Selected",
"No item selected to search and replace through.",
		NULL,
		CDIALOG_ICON_WARNING,
		CDIALOG_BTNFLAG_OK,
		CDIALOG_BTNFLAG_OK
	    );
	    CDialogSetTransientFor(NULL);
	    EditorSetBusy(editor, FALSE);
	    return;
	}

	/* Get the needle */
	needle = STRDUP(gtk_entry_get_text(GTK_ENTRY(find_combo->entry)));

	/* Get the replacement */
	replacement = STRDUP(gtk_entry_get_text(GTK_ENTRY(replace_combo->entry)));

	/* Handle by the EditorItem's type */
	switch(item->type)
	{
	  case EDITOR_ITEM_HEADER:
	    text = GTK_TEXT(editor->header_comments_text);
	    editable = GTK_EDITABLE(text);

	    /* Get the haystack from the header text */
	    haystack = gtk_editable_get_chars(editable, 0, -1);

	    /* Set the starting position plus one */
	    start_pos = gtk_editable_get_position(editable) + 1;

	    /* Replace */
	    made_replace = EditorReplace(
		editor,
		text,
		haystack,
		needle,
		replacement,
		(gint)gtk_text_get_length(text),
		start_pos,
		case_sensitive
	    );
	    /* Did we make a replacement? */
	    if(made_replace)
	    {

	    }
	    else
	    {
		/* No replacement, implies search string not found */
		CDialogSetTransientFor(toplevel);
		CDialogGetResponse(
"String Not Found",
"The string was not found.",
		    NULL,
		    CDIALOG_ICON_SEARCH,
		    CDIALOG_BTNFLAG_OK,
		    CDIALOG_BTNFLAG_OK
		);
		CDialogSetTransientFor(NULL);
	    }
	    break;

	  case EDITOR_ITEM_SECTION:
	    text = GTK_TEXT(editor->section_text);
	    editable = GTK_EDITABLE(text);

	    /* Get the haystack from the section text */
	    haystack = gtk_editable_get_chars(editable, 0, -1);

	    /* Set the starting position plus one */
	    start_pos = gtk_editable_get_position(editable) + 1;

	    /* Replace */
	    made_replace = EditorReplace(
		editor,
		text,
		haystack,
		needle,
		replacement,
		(gint)gtk_text_get_length(text),
		start_pos,
		case_sensitive
	    );
	    /* Did we make a replacement? */
	    if(made_replace)
	    {

	    }
	    else
	    {
		/* No replacement, implies search string not found */
		CDialogSetTransientFor(toplevel);
		CDialogGetResponse(
"String Not Found",
"The string was not found.",
		    NULL,
		    CDIALOG_ICON_SEARCH,
		    CDIALOG_BTNFLAG_OK,
		    CDIALOG_BTNFLAG_OK
		);
		CDialogSetTransientFor(NULL);
	    }
	    break;

	  case EDITOR_ITEM_FILE:
	    break;
	}

	/* Set the selected node's has changes flag */
	EditorSetNodeHasChanges(
	    editor,
	    sel_node,
	    TRUE
	);

	g_free(replacement);
	g_free(haystack);
	g_free(needle);

	EditorUpdateMenus(editor);
	EditorSetBusy(editor, FALSE);
}

/*
 *	Right Click GtkMenu map callback.
 */
gint EditorMenuMapCB(GtkWidget *widget, GdkEvent *event, gpointer data)
{
	GtkWidget *w;
	GdkEventButton *button;
	editor_struct *editor = EDITOR(data);
	if((editor == NULL) || (event == NULL) || (widget == NULL))
	    return(TRUE);

	if(editor->freeze_count > 0)
	    return(TRUE);

	/* Is event type button press? If so then get pointer to button
	 * event structure.
	 */
	if((event->type == GDK_BUTTON_PRESS) ||
	   (event->type == GDK_BUTTON_RELEASE)
	)
	    button = (GdkEventButton *)event;
	else
	    button = NULL;

	/* Layout ctree? */
	if(widget == editor->layout_ctree)
	{
	    w = editor->layout_menu;
	    if((button != NULL) && (w != NULL))
	    {
		if(button->button == 3)
		{
		    gint row, column;
		    GtkCTreeNode *branch;


		    /* Try to select branch button event occured
		     * over first.
		     */
		    if(gtk_clist_get_selection_info(
			(GtkCList *)widget,
			button->x, button->y,
			&row, &column
		    ))
		    {
			branch = gtk_ctree_node_nth(
			    (GtkCTree *)widget, row
			);
			/* Branch valid and not selected? */
			if((branch != NULL) &&
			   (branch != EditorCurrentSelectedNode(editor))
			)
			{
			    EditorSelect(editor, branch);
			}
		    }

		    /* Map menu */
		    gtk_menu_popup(
			GTK_MENU(w),
			NULL, NULL, NULL, editor,
			button->button, button->time
		    );
		}
	    }
	}
	/* Edit panel text widgets */
	else if((widget == editor->header_comments_text) ||
		(widget == editor->section_text)
	)
	{
	    w = editor->edit_panel_menu;
	    if((button != NULL) && (w != NULL))
	    {
		/* Call key press event for text widget since position
		 * of cursor may have changed.
		 */
		EditorTextKeyEventCB(
		    (GtkEditable *)widget, NULL, data
		);

		if(button->button == 3)
		{
		    /* Map menu */
		    gtk_menu_popup(
			GTK_MENU(w),
			NULL, NULL, NULL, editor,
			button->button, button->time
		    );

		    /* Need to mark the text widget's button as 0 or
		     * else it will keep marking.
		     */
		    GTK_TEXT(widget)->button = 0;
		    gtk_grab_remove(widget);
		}
	    }
	}

	return(TRUE);
}


/*
 *	Expand/colapse callback.
 */
void EditorExpandCB(GtkWidget *widget, gpointer data)
{
	GtkCTree *ctree;
	GtkCTreeNode *node; 
	editor_struct *editor = EDITOR(data);
	if(editor == NULL)
	    return;

	if(editor->freeze_count > 0)
	    return;

	ctree = GTK_CTREE(editor->layout_ctree);
	node = EditorCurrentSelectedNode(editor);
	if(node == NULL)
	    return;

	if(GTK_CTREE_NODE_EXPANDED(node))
	    gtk_ctree_collapse(ctree, node);
	else
	    gtk_ctree_expand(ctree, node);
}


/*
 *	Layout GtkCTree "tree_select_row" signal callback.
 */
void EditorLayoutCTreeSelectCB(
	GtkCTree *ctree, GtkCTreeNode *node, gint column,
	gpointer data
)
{
	gint row;
	GtkCList *clist;
	editor_item_struct *item;
	editor_struct *editor = EDITOR(data);
	if((ctree == NULL) || (editor == NULL))
	    return;

	if(editor->freeze_count > 0)
	    return;

	clist = GTK_CLIST(ctree);

	EditorSetBusy(editor, TRUE);       

	/* Set the DND drag icon */
	GUIDNDSetDragIconFromCTreeSelection(ctree);

	/* Scroll if row not visibile */
	row = gtk_clist_find_row_from_data(
	    clist,
	    EditorBranchGetData(ctree, node)
	);
	if(row > -1)
	{
	    if(gtk_clist_row_is_visible(clist, row) !=
		GTK_VISIBILITY_FULL
	    )
		gtk_clist_moveto(
		    clist,
		    row, 0,			/* Row, column */
		    0.5f, 0.0f			/* Row, column */
		);
	}

	/* Get the values for the selected EditorItem */
	if(node != NULL)
	{
	    gchar *msg;

	    /* Get the values from the EditorItem to the widgets */
	    EditorGetValues(
		editor,
		node
	    );

	    /* Set the Status Bar message */
	    item = EditorBranchGetData(
		ctree, 
		node
	    );
	    if(item != NULL)
	    {
		switch(item->type)
		{
		  case EDITOR_ITEM_FILE:
		    msg = g_strdup_printf(
			"Selected manual page \"%s\"",
			item->name
		    );
		    EditorSetStatusMessage(editor, msg);
		    g_free(msg);
		    break;

		  case EDITOR_ITEM_HEADER:
		    EditorSetStatusMessage(editor, "Selected header");
		    break;

		  case EDITOR_ITEM_SECTION:
		    msg = g_strdup_printf(
			"Selected section \"%s\"",
			item->section_name
		    );
		    EditorSetStatusMessage(editor, msg);
		    g_free(msg);
		    break;

		  default:
		    EditorSetStatusMessage(editor, "Selected unsupported type");
		    break;
		}
	    }
	    else
	    {
		EditorSetStatusMessage(editor, "");
	    }
	}
	else
	{
	    EditorSetStatusMessage(editor, "");
	}

	EditorUpdateMenus(editor);

	EditorSetBusy(editor, FALSE);
}

/*
 *	Layout GtkCTree "tree_unselect_row" signal callback.
 */
void EditorLayoutCTreeUnselectCB(
	GtkCTree *ctree, GtkCTreeNode *node, gint column, gpointer data
)
{
	editor_struct *editor = EDITOR(data);
	if((ctree == NULL) || (editor == NULL))
	    return;

	if(editor->freeze_count > 0)
	    return;

	EditorSetBusy(editor, TRUE);

	/* Set the values from the panel widgets to the selected
	 * EditorItem
	 */
	EditorSetValues(editor, node);

	/* Clear values on editor's editing widgets, this is so
	 * user does not type in something into those widgets
	 * while no branch is selected (thus any changes would be
	 * lost when a branch gets selected again)
	 */
	EditorClearValues(editor);

	/* Clear undo/redo list on editor, this is so that they can't
	 * get applied by the user on the newly selected branch's
	 * data
	 */
	EditorClearUndosList(editor);

	EditorUpdateMenus(editor);

	EditorSetBusy(editor, FALSE);
}

/*
 *	Editor layout ctree branch expand (and collapse) callback, called
 *	whenever a branch is expanded or collapsed.
 */
void EditorLayoutCTreeExpandCB(
	GtkCTree *ctree, GtkCTreeNode *node, gpointer data
)
{
	GtkCList *clist;
	editor_struct *editor = EDITOR(data);
	if((ctree == NULL) || (editor == NULL))
	    return;

	if(editor->freeze_count > 0)
	    return;

	clist = GTK_CLIST(ctree);
 
	/* Update layout clist column width */
	gtk_clist_set_column_width(
	    clist,
	    0,					/* Column */
	    gtk_clist_optimal_column_width(clist, 0)
	);
}


/*
 *	Syntax highlight timeout callback.
 *
 *	For performing syntax highlighting a delayed time after text has
 *	changed or been deleted on the section_text widget.
 *
 *	If editor is currently processing then this operation will not be
 *	performed.
 *
 *	This operation will only be performed once, the value of the
 *	editor's synhl_timeout_cb_id will be set to (guint)(-1) after
 *	this function has been called and this function always returns
 *	FALSE to indicate this function is not to be called again
 *	unless we add a new timeout callback.
 */
gint EditorSyntaxHighlightTimeoutCB(gpointer data)
{
	editor_struct *editor = EDITOR(data);
	if(editor == NULL)
	    return(FALSE);

	if(editor->freeze_count > 0)
	    return(FALSE);

	editor->freeze_count++;

	/* Reset synhl_timeout_cb_id, since this function will not be
	 * called again untill we add a new timeout callback.
	 */
	editor->synhl_timeout_cb_id = 0;

	/* Syntax highlighting */
	EditorSyntaxHighlight(
	    editor, editor->section_text,
	    editor->synhl_timeout_cb_cursor_pos,
	    editor->synhl_timeout_cb_start_pos,
	    editor->synhl_timeout_cb_end_pos
	);

	editor->synhl_timeout_cb_cursor_pos = -1;
	editor->synhl_timeout_cb_start_pos = -1;
	editor->synhl_timeout_cb_end_pos = -1;

	editor->freeze_count--;

	return(FALSE);
}


/*
 *	Editor GtkText widget key event callback, this is to check
 *	when the cursor has moved so that the position value on the
 *	status bar is properly set.
 *
 *	This function always returns FALSE.
 */
gint EditorTextKeyEventCB(
	GtkEditable *editable, GdkEventKey *key, gpointer data
)
{
	int c, i, row, column;
	int cur_pos, length;
	editor_struct *editor = EDITOR(data);
	if((editor == NULL) || (editable == NULL))
	    return(TRUE);

	/* Get cursor position */
	cur_pos = (gint)gtk_editable_get_position(editable);

	/* Get length of buffer */
	length = (gint)gtk_text_get_length(GTK_TEXT(editable));

	if(cur_pos >= length)
	    cur_pos = length - 1;
	if(cur_pos < 0)
	    cur_pos = 0;

	for(i = 0, row = 0, column = 0; i < cur_pos; i++)
	{
	    c = GTK_TEXT_INDEX(GTK_TEXT(editable), i);
	    if((gchar)c == '\n')
	    {
		column = 0;
		row++;
	    }
	    else
	    {
		column++;
	    }
	}

	EditorSetStatusPosition(editor, row, column);

	return(TRUE);
}


/*
 *	Text change callback.
 */
void EditorTextChangeCB(GtkEditable *editable, gpointer data)
{
	editor_struct *editor = EDITOR(data);
	if((editor == NULL) || (editable == NULL))
	    return;


	return;
}

/*
 *	Editor text widget insert text callback.
 */
void EditorTextInsertCB(
	GtkEditable *editable, const gchar *text, gint length,
	gint *position, gpointer data
)
{
	int i;
	gint start_pos, end_pos, text_len;
	gpointer u;
	editor_undo_remove_chars_struct *u_remove;
	GtkWidget *w;
	editor_struct *editor = EDITOR(data);
	if((editor == NULL) || (editable == NULL) || (position == NULL))
	    return;

	/* Do not process this signal if editor is processing an undo or
	 * redo since this signal may be called during the undo process.
	 */
	if(editor_processing_undo == editor)
	    return;

	w = GTK_WIDGET(editable);

	/* Get positions */
	start_pos = (*position);
	end_pos = start_pos + length;

	/* Get text length */
	text_len = end_pos - start_pos;
	if(text_len < 0)
	    text_len *= -1;  

	/* Sanitize positions */
	if(end_pos < start_pos)
	{
	    gint tmp_pos = end_pos;
	    end_pos = start_pos;
	    start_pos = tmp_pos;
	}

	/* No text length? */
	if(text_len == 0)
	    return;

	/* Sanitize total */
	if(editor->total_undos < 0)
	    editor->total_undos = 0;

/* Macro to allocate a new undo structure of type EDITOR_UNDO_REMOVE_CHARS
 * on the editor->undo list. Uses variables i, u, editor, w,
 * start_pos, and u_remove. Will return this function if there was an
 * error and u_remove will be set properly to the new structure.
 */
#define DO_NEW_UNDO_REMOVE_STRUCT	{	\
 i = editor->total_undos;			\
 editor->total_undos = i + 1;			\
 editor->undo = (gpointer *)g_realloc(		\
  editor->undo,					\
  editor->total_undos * sizeof(gpointer)	\
 );						\
 if(editor->undo == NULL) {			\
  editor->total_undos = 0;			\
  return;					\
 }						\
 /* Shift */					\
 for(i = editor->total_undos - 1; i > 0; i--)	\
  editor->undo[i] = editor->undo[i - 1];	\
 /* Add new struct to beginning of list */	\
 u = EditorUndoNew(				\
  editor,					\
  EDITOR_UNDO_REMOVE_CHARS,			\
  "Insert Characters"				\
 );						\
 editor->undo[0] = u;   /* Record new struct to undo list */ \
						\
 /* Set initial values for new structure if possible */ \
 u_remove = (editor_undo_remove_chars_struct *)u; \
 if(u_remove != NULL) {				\
  u_remove->w = w;				\
  u_remove->position = start_pos;		\
 }						\
}

	/* Get first undo structure, allocate as needed if it is NULL
	 * or is not of type EDITOR_UNDO_REMOVE_CHARS.
	 */
	if(editor->total_undos > 0)
	    u = editor->undo[0];
	else
	    u = NULL;
	u_remove = (editor_undo_remove_chars_struct *)u;
	if((u == NULL) ? 1 : ((*(int *)u) != EDITOR_UNDO_REMOVE_CHARS))
	{
	    DO_NEW_UNDO_REMOVE_STRUCT
	    if(u_remove == NULL)
		return;

	    EditorSetNodeHasChanges(
		editor,
		EditorCurrentSelectedNode(editor),
		TRUE
	    );
	    EditorUpdateMenus(editor);
	}
	u = NULL;       /* Don't use u past this point */

	/* Sanitize length on undo structure */
	if(u_remove->length < 0)
	    u_remove->length = 0;

	/* Are the new positions contiguous? */
	if((u_remove->position + u_remove->length) == start_pos)
	{
	    u_remove->length += text_len;
	}
	else
	{
	    /* New segment being inserted, need to allocate new undo
	     * structure.
	     */
	    DO_NEW_UNDO_REMOVE_STRUCT
	    if(u_remove == NULL)
		return;

	    u_remove->position = start_pos;
	    u_remove->length = text_len;

	    EditorSetNodeHasChanges(
		editor,
		EditorCurrentSelectedNode(editor),
		TRUE
	    );
	    EditorUpdateMenus(editor);
	}


	/* Remove excess undo structures */
	if(editor->total_undos > editor->max_undos)
	{
	    int max = MAX(editor->max_undos, 0);

	    /* Delete all undo structures greater than the max */
	    for(i = editor->total_undos - 1; i >= max; i--)
		EditorUndoDelete(editor, editor->undo[i]);

	    /* Update total and reallocate pointer array */
	    editor->total_undos = max;
	    if(editor->total_undos > 0)
	    {
		editor->undo = (gpointer *)g_realloc(
		    editor->undo,
		    editor->total_undos * sizeof(gpointer)
		);
		if(editor->undo == NULL)
		    editor->total_undos = 0;
	    }   
	    else
	    {
		g_free(editor->undo);
		editor->undo = NULL;
		editor->total_undos = 0;
	    }
	}


	/* Mark currently selected branch item data as having changes */
	EditorSetNodeHasChanges(
	    editor,
	    EditorCurrentSelectedNode(editor),
	    TRUE
	);

#undef DO_NEW_UNDO_REMOVE_STRUCT

	/* If this is the editor's section_text widget then add a timeout
	 * callback for syntax highlighting.
	 */
	if(GTK_WIDGET(editable) == editor->section_text)
	{
	    /* Cancel previous syntax highlighting as needed */
	    if(editor->synhl_timeout_cb_id != 0)
	    {
		gtk_timeout_remove(editor->synhl_timeout_cb_id);
		editor->synhl_timeout_cb_id = 0;
	    }

	    /* Set cursor position as needed and if less than */
	    if(editor->synhl_timeout_cb_cursor_pos < 0)
		editor->synhl_timeout_cb_cursor_pos = start_pos;
	    else
		editor->synhl_timeout_cb_cursor_pos = MIN(
		    editor->synhl_timeout_cb_cursor_pos,
		    start_pos
		);

	    /* Set start position as needed and if less than */
	    if(editor->synhl_timeout_cb_start_pos < 0)
		editor->synhl_timeout_cb_start_pos = start_pos;
	    else
		editor->synhl_timeout_cb_start_pos = MIN(
		    editor->synhl_timeout_cb_start_pos,
		    start_pos
		);

	    /* Set end position as needed and if greater than */
	    if(editor->synhl_timeout_cb_end_pos < 0)
		editor->synhl_timeout_cb_end_pos = end_pos;
	    else
		editor->synhl_timeout_cb_end_pos = MAX(
		    editor->synhl_timeout_cb_end_pos,
		    end_pos
		);

	    /* Add new timeout callback for syntax highlighting */
	    editor->synhl_timeout_cb_id = gtk_timeout_add(
		500,
		EditorSyntaxHighlightTimeoutCB,
		editor
	    );
	}
}

/*
 *	Editor text widget delete text callback.
 */
void EditorTextDeleteCB(
	GtkEditable *editable, gint start_pos, gint end_pos, gpointer data
)
{
	int i;
	gint text_len;
	gchar *text_buf;
	GtkWidget *w;
	gpointer u;
	editor_undo_insert_chars_struct *u_insert;
	editor_struct *editor = EDITOR(data);
	if((editor == NULL) || (editable == NULL))
	    return;

	/* Do not process this signal if editor is processing an undo or
	 * redo since this signal may be called during the undo process.
	 */
	if(editor_processing_undo == editor)
	    return;

	w = GTK_WIDGET(editable);


	/* Get text length */
	text_len = end_pos - start_pos;
	if(text_len < 0)
	    text_len *= -1;

	/* Sanitize positions */
	if(end_pos < start_pos)
	{
	    gint tmp_pos = end_pos;
	    end_pos = start_pos;
	    start_pos = tmp_pos;
	}

	/* No text length? */
	if(text_len == 0)
	    return;

	/* Sanitize total */
	if(editor->total_undos < 0)
	    editor->total_undos = 0;

/* Macro to allocate a new undo structure of type EDITOR_UNDO_INSERT_CHARS
 * on the editor->undo list. Uses variables i, u, editor, w,
 * start_pos, and u_insert. Will return this function if there was an
 * error and u_insert will be set properly to the new structure.
 */
#define DO_NEW_UNDO_INSERT_STRUCT	\
{ \
 i = editor->total_undos; \
 editor->total_undos = i + 1; \
 editor->undo = (gpointer *)g_realloc( \
  editor->undo, \
  editor->total_undos * sizeof(gpointer) \
 ); \
 if(editor->undo == NULL) \
 { \
  editor->total_undos = 0; \
  return; \
 } \
 /* Shift */ \
 for(i = editor->total_undos - 1; i > 0; i--) \
  editor->undo[i] = editor->undo[i - 1]; \
 /* Add new struct to beginning of list */ \
 u = EditorUndoNew( \
  editor, EDITOR_UNDO_INSERT_CHARS, \
  "Remove Characters" \
 ); \
 editor->undo[0] = u;	/* Record new struct to undo list */ \
 \
 /* Set initial values for new structure if possible */ \
 u_insert = (editor_undo_insert_chars_struct *)u; \
 if(u_insert != NULL) \
 { \
  u_insert->w = w; \
  u_insert->position = start_pos; \
 } \
}

	/* Get first undo structure, allocate as needed if it is NULL
	 * or is not of type EDITOR_UNDO_INSERT_CHARS.
	 */
	if(editor->total_undos > 0)
	    u = editor->undo[0];
	else
	    u = NULL;
	u_insert = (editor_undo_insert_chars_struct *)u;
	if((u == NULL) ? 1 : ((*(int *)u) != EDITOR_UNDO_INSERT_CHARS))
	{
	    DO_NEW_UNDO_INSERT_STRUCT
	    if(u_insert == NULL)
		return;

	    EditorSetNodeHasChanges(
		editor,
		EditorCurrentSelectedNode(editor),
		TRUE
	    );
	    EditorUpdateMenus(editor);
	}
	u = NULL;	/* Don't use u past this point */

	/* Sanitize length on undo structure */
	if(u_insert->length < 0)
	    u_insert->length = 0;

	/* Are the new positions contiguous? */
	if(u_insert->position == start_pos)
	{
	    /* Contiguous forward delete */
	    int old_len = u_insert->length;


	    /* Get characters that are about to be deleted */
	    text_buf = gtk_editable_get_chars(
		GTK_EDITABLE(w), start_pos, end_pos
	    );

	    /* Increase allocation for text on undo structure */
	    u_insert->length += text_len;
	    u_insert->characters = (gchar *)g_realloc(
		u_insert->characters,
		(u_insert->length + 1) * sizeof(gchar)
	    );
	    if(u_insert->characters == NULL)
	    {
		u_insert->length = 0;
	    }
	    else
	    {
		if(text_buf != NULL)
		    memcpy(
			&u_insert->characters[old_len],
			text_buf,
			text_len * sizeof(gchar)
		    );
		u_insert->characters[u_insert->length] = '\0';
	    } 

	    g_free(text_buf);
	    text_buf = NULL;
	}
	else if(u_insert->position == end_pos)
	{
	    /* Contiguous backward delete */
	    int old_len = u_insert->length;


	    /* Get characters that are about to be deleted */
	    text_buf = gtk_editable_get_chars(
		GTK_EDITABLE(w), start_pos, end_pos
	    );

	    /* Increase allocation for text on undo structure */
	    u_insert->length += text_len;
	    u_insert->characters = (gchar *)g_realloc(
		u_insert->characters,
		(u_insert->length + 1) * sizeof(gchar)
	    );
	    if(u_insert->characters == NULL)
	    {
		u_insert->length = 0;
	    }
	    else
	    {
		if(old_len > 0)
		    memmove(
			&u_insert->characters[text_len],
			u_insert->characters,
			old_len * sizeof(gchar)
		    );

		if(text_buf != NULL)
		    memcpy(
			u_insert->characters,
			text_buf,
			text_len * sizeof(gchar)
		    );
		u_insert->characters[u_insert->length] = '\0';
	    }

	    g_free(text_buf);
	    text_buf = NULL;

	    /* Need to update position */
	    u_insert->position = start_pos;
	}
	else
	{
	    /* New segment being deleted, need to allocate new undo
	     * structure.
	     */
	    int old_len;

	    DO_NEW_UNDO_INSERT_STRUCT
	    if(u_insert == NULL)
		return;

	    old_len = u_insert->length;

	    /* Get characters that are about to be deleted */
	    text_buf = gtk_editable_get_chars(
		GTK_EDITABLE(w), start_pos, end_pos
	    );

	    /* Increase allocation for text on undo structure */
	    u_insert->length += text_len;
	    u_insert->characters = (gchar *)g_realloc(
		u_insert->characters,
		(u_insert->length + 1) * sizeof(gchar)
	    );
	    if(u_insert->characters == NULL)
	    {
		u_insert->length = 0;
	    }
	    else
	    {
		if(text_buf != NULL)
		    memcpy(
			&u_insert->characters[old_len],
			text_buf,
			text_len * sizeof(gchar)
		    );
		u_insert->characters[u_insert->length] = '\0';
	    }

	    g_free(text_buf);
	    text_buf = NULL;

	    EditorSetNodeHasChanges(
		editor,
		EditorCurrentSelectedNode(editor),
		TRUE
	    );
	    EditorUpdateMenus(editor);
	}


	/* Remove excess undo structures */
	if(editor->total_undos > editor->max_undos)
	{
	    int max = MAX(editor->max_undos, 0);

	    /* Delete all undo structures greater than the max */
	    for(i = editor->total_undos - 1; i >= max; i--)
		EditorUndoDelete(editor, editor->undo[i]);

	    /* Update total and reallocate pointer array */
	    editor->total_undos = max;
	    if(editor->total_undos > 0)
	    {
		editor->undo = (gpointer *)g_realloc(
		    editor->undo,
		    editor->total_undos * sizeof(gpointer)
		);
		if(editor->undo == NULL)
		    editor->total_undos = 0;
	    }
	    else
	    {
		g_free(editor->undo);
		editor->undo = NULL;
		editor->total_undos = 0;
	    }
	}


	/* Mark currently selected branch item data as having changes */
	EditorSetNodeHasChanges(
	    editor,
	    EditorCurrentSelectedNode(editor),
	    TRUE
	);

#undef DO_NEW_UNDO_INSERT_STRUCT

	/* If this is the editor's section_text widget then add a timeout
	 * callback for syntax highlighting.
	 */
	if(GTK_WIDGET(editable) == editor->section_text)
	{
	    /* Cancel previous syntax highlighting as needed */
	    if(editor->synhl_timeout_cb_id != 0)
	    {
		gtk_timeout_remove(editor->synhl_timeout_cb_id);
		editor->synhl_timeout_cb_id = 0;
	    }

	    /* Set cursor position as needed and if less than */
	    if(editor->synhl_timeout_cb_cursor_pos < 0)
		editor->synhl_timeout_cb_cursor_pos = start_pos;
	    else
		editor->synhl_timeout_cb_cursor_pos = MIN(
		    editor->synhl_timeout_cb_cursor_pos,
		    start_pos
		);

	    /* Set start position as needed and if less than */
	    if(editor->synhl_timeout_cb_start_pos < 0)
		editor->synhl_timeout_cb_start_pos = start_pos;
	    else
		editor->synhl_timeout_cb_start_pos = MIN(
		    editor->synhl_timeout_cb_start_pos,
		    start_pos
		);

	    /* Set end position as needed and if greater than */
	    if(editor->synhl_timeout_cb_end_pos < 0)
		editor->synhl_timeout_cb_end_pos = end_pos;
	    else
		editor->synhl_timeout_cb_end_pos = MAX(
		    editor->synhl_timeout_cb_end_pos,
		    end_pos
		);

	    /* Add new timeout callback for syntax highlighting */
	    editor->synhl_timeout_cb_id = gtk_timeout_add(
		500,
		EditorSyntaxHighlightTimeoutCB,
		editor
	    );
	}
}


/*
 *	Reconfigured callback.
 */
void EditorReconfiguredCB(editor_struct *editor)
{
	GtkRcStyle *rcstyle;
	CfgList *cfg_list;
	editor_item_struct *item;
	me_core_struct *core;

	if(editor == NULL)
	    return;

	core = editor->core;
	cfg_list = core->cfg_list;

	/* Update the syntax highlighting flag */
	editor->syntax_highlighting = ME_GET_B(
	    ME_CFG_PARM_EDITOR_SYNTAX_HIGHLIGHTING
	);

	/* Update the Tool Bar */
	ToolBarSetButtonDecals(
	    editor->tool_bar,
	    (ToolBarButtonDecals)ME_GET_I(ME_CFG_PARM_EDITOR_TOOL_BAR_BUTTON_DECALS)
	);
	ToolBarSetRelief(
	    editor->tool_bar,
	    (GtkReliefStyle)ME_GET_I(ME_CFG_PARM_EDITOR_TOOL_BAR_RELIEF)
	);

	/* Update the colors */
	if(editor->colormap != NULL)
	{
	    GdkColormap *colormap = editor->colormap;
	    GdkColor *ct;
	    CfgColor *cs;

	    ct = &editor->normal_text_color;
	    cs = ME_GET_COLOR(ME_CFG_PARM_COLOR_EDITOR_FG);
	    if(cs != NULL)
	    {
		GDK_COLOR_SET_COEFF(ct, cs->r, cs->g, cs->b);
		GDK_COLORMAP_FREE_COLOR(colormap, ct);
		GDK_COLORMAP_ALLOC_COLOR(colormap, ct);
	    }

	    ct = &editor->deliminator_text_color;
	    cs = ME_GET_COLOR(ME_CFG_PARM_COLOR_EDITOR_DELIMINATOR);
	    if(cs != NULL)
	    {
		GDK_COLOR_SET_COEFF(ct, cs->r, cs->g, cs->b);
		GDK_COLORMAP_FREE_COLOR(colormap, ct);
		GDK_COLORMAP_ALLOC_COLOR(colormap, ct);
	    }

	    ct = &editor->tag_text_color;
	    cs = ME_GET_COLOR(ME_CFG_PARM_COLOR_EDITOR_TAG);
	    if(cs != NULL)
	    {
		GDK_COLOR_SET_COEFF(ct, cs->r, cs->g, cs->b);
		GDK_COLORMAP_FREE_COLOR(colormap, ct);
		GDK_COLORMAP_ALLOC_COLOR(colormap, ct);
	    }

	    ct = &editor->symbol_text_color;
	    cs = ME_GET_COLOR(ME_CFG_PARM_COLOR_EDITOR_SYMBOL);
	    if(cs != NULL)
	    {
		GDK_COLOR_SET_COEFF(ct, cs->r, cs->g, cs->b);
		GDK_COLORMAP_FREE_COLOR(colormap, ct);
		GDK_COLORMAP_ALLOC_COLOR(colormap, ct);
	    }
	}

	/* Update the GtkEntry and GtkText styless */
	rcstyle = core->editable_rcstyle;
	gtk_widget_modify_style(
	    editor->header_name_entry,
	    rcstyle
	);
	gtk_widget_modify_style(
	    editor->header_section_number_entry,
	    rcstyle
	);
	gtk_widget_modify_style(
	    editor->header_version_entry,
	    rcstyle
	);
	gtk_widget_modify_style(
	    editor->header_author_entry,
	    rcstyle
	);
	gtk_widget_modify_style(
	    editor->header_catagory_entry,
	    rcstyle
	);
	gtk_widget_modify_style(
	    editor->header_comments_text,
	    rcstyle
	);

	gtk_widget_modify_style(
	    editor->section_name_entry,
	    rcstyle
	);
	gtk_widget_modify_style(
	    editor->section_text,
	    rcstyle
	);

	/* Update the displayed data */
	item = EditorCurrentSelectedItem(editor);
	if(item != NULL)
	{
	    /* Check if the selected item is a section */
	    if(item->type == EDITOR_ITEM_SECTION)
	    {
		/* Need to apply and fetch values to ensure update
		 * on section_text widget's text. The function
		 * EditorGetValues() will call syntax
		 * highlighting on the newly loaded text.
		 */
		GtkCTreeNode *node = EditorCurrentSelectedNode(editor);
		EditorSetValues(editor, node);
		EditorGetValues(editor, node);
	    }
	}

	EditorUpdateMenus(editor);
}
