/*
                         Editor Find In Pages Dialog
 */

#ifndef EDITOR_FIND_IN_PAGES_H
#define EDITOR_FIND_IN_PAGES_H

#include <gtk/gtk.h>
#include "me_types.h"


/* typedef struct _editor_fip_struct	editor_fip_struct; */

typedef struct _editor_fip_item_struct	editor_fip_item_struct;
#define EDITOR_FIP_ITEM(p)		((editor_fip_item_struct *)(p))


/*
 *	Find In Pages Dialog:
 */
struct _editor_fip_struct {

	GtkWidget	*toplevel;
	GtkAccelGroup	*accelgrp;
	gint		freeze_count,
			busy_count,
			stop_count;
	gboolean	processing;

	me_core_struct	*core;
	editor_struct	*editor;

	GtkWidget	*find_combo,
			*match_all_radio,
			*match_any_radio,
			*match_phrase_radio,
			*case_sensitive_toggle,
			*results_clist,
			*goto_btn,
			*find_btn,
			*clear_btn,
			*stop_btn,
			*close_btn;

	GtkWidget	*results_menu,
			*goto_mi;

	GtkWidget	*status_bar_dock,
			*status_bar_toplevel,
			*status_bar_progress,
			*status_bar_label;

	gfloat		status_bar_progress_pos_last;
};


/*
 *	Find In Pages Dialog Item:
 */
struct _editor_fip_item_struct {

	editor_fip_struct	*fip;

	/* Matched GtkCTreeNode on the Editor's Layout GtkCTree */
	GtkCTreeNode	*matched_node;

	/* This will be used for matching the EditorItem on the
	 * Editor when the user wants to "Go To" the matched result,
	 * this is better since the found_branch may get deleted and
	 * matching by the branch's data seems to work out better
	 */
	editor_item_struct	*editor_item;

	/* Matched text position, can be -1 */
	gint		text_sel_start,
			text_sel_end;

	/* Matched line number, can be -1 */
	gint		line_num;

};


extern void EditorFIPItemDelete(editor_fip_item_struct *item);

/* Item Add/Remove */
extern gint EditorFIPAddItem(
        editor_fip_struct *fip,
        const gchar *manual_page,
        const gchar *section,
        const gint line_num,
        const gchar *excerpt,
        const gint text_sel_start, const gint text_sel_end,
        GtkCTreeNode *matched_node,
        editor_item_struct *editor_item
);

/* Find In Pages Dialog */
extern editor_fip_struct *EditorFIPNew(
        me_core_struct *core,
        editor_struct *editor
);
extern void EditorFIPUpdateMenus(editor_fip_struct *fip);
extern void EditorFIPSetStatusMessage(
	editor_fip_struct *fip,
	const gchar *msg
);
extern void EditorFIPSetStatusProgress(
	editor_fip_struct *fip,
	const gfloat percent
);
extern void EditorFIPSetBusy(
        editor_fip_struct *fip,
        const gboolean busy
);
extern gboolean EditorFIPIsMapped(editor_fip_struct *fip);
extern void EditorFIPMap(editor_fip_struct *fip);
extern void EditorFIPUnmap(editor_fip_struct *fip);
extern void EditorFIPDelete(editor_fip_struct *fip);


#endif	/* EDITOR_FIND_IN_PAGES_H */
