#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "guiutils.h"
#include "cdialog.h"

#include "editor.h"
#include "editor_find_in_pages.h"
#include "editor_find_in_pages_cb.h"
#include "editor_find_in_pages_op.h"
#include "manedit.h"


void EditorFIPItemDestroyCB(gpointer data);

gint EditorFIPDeleteEventCB(GtkWidget *widget, GdkEvent *event, gpointer data);
void EditorFIPCloseMCB(GtkWidget *widget, gpointer data);

void EditorFIPFindCB(GtkWidget *widget, gpointer data);
void EditorFIPFindActivateCB(GtkWidget *widget, gpointer data);
void EditorFIPClearCB(GtkWidget *widget, gpointer data);
void EditorFIPStopCB(GtkWidget *widget, gpointer data);
void EditorFIPGotoCB(GtkWidget *widget, gpointer data);

gint EditorFIPMenuMapCB(GtkWidget *widget, GdkEvent *event, gpointer data);

void EditorFIPResultsListSelectCB(
	GtkCList *clist, gint row, gint column,
	GdkEventButton *event, gpointer data
);
void EditorFIPResultsListUnselectCB(
	GtkCList *clist, gint row, gint column,
	GdkEventButton *event, gpointer data
);
void EditorFIPResultsListColumnClickCB(
	GtkCList *clist, gint column, gpointer data
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)


/*
 *	Item destroy callback.
 */
void EditorFIPItemDestroyCB(gpointer data)
{
	editor_fip_item_struct *item = EDITOR_FIP_ITEM(data);
	if(item == NULL)
	    return;

	EditorFIPItemDelete(item);
}

/*
 *	Toplevel GtkWindow "delete_event" signal callback.
 */
gint EditorFIPDeleteEventCB(GtkWidget *widget, GdkEvent *event, gpointer data)
{
	editor_fip_struct *fip = EDITOR_FIP(data);
	if(fip == NULL)
	    return(TRUE);

	if((fip->freeze_count > 0) || fip->processing)
	    return(TRUE);

	EditorFIPUnmap(fip);

	return(TRUE);
}

/*
 *	Close callback.
 */
void EditorFIPCloseMCB(GtkWidget *widget, gpointer data)
{
	editor_fip_struct *fip = EDITOR_FIP(data);
	if(fip == NULL)
	    return;

	if((fip->freeze_count > 0) || fip->processing)
	    return;

	EditorFIPUnmap(fip);
}


/*
 *	Find button callback.
 */
void EditorFIPFindCB(GtkWidget *widget, gpointer data)
{
	gchar *value;
	GtkCombo *find_combo;
	editor_fip_struct *fip = EDITOR_FIP(data);
	if(fip == NULL)
	    return;

	if((fip->freeze_count > 0) || fip->processing)
	    return;

	find_combo = GTK_COMBO(fip->find_combo);
	value = STRDUP(gtk_entry_get_text(GTK_ENTRY(find_combo->entry)));

	/* Call combo widget to set new value and activate, this will
	 * then call EditorFIPFindActivateCB()
	 */
	GUIComboActivateValue(GTK_WIDGET(find_combo), value);

	g_free(value);
}

/*
 *	Find GtkCombo "activate" signal callback.
 */
void EditorFIPFindActivateCB(GtkWidget *widget, gpointer data)
{
	gboolean case_sensitive = FALSE;
	gint		matches,
			throughness = 0;
	gchar		*s,
			*search_string;
	GtkCombo *find_combo;
	GtkCList *results_clist;
	editor_fip_struct *fip = EDITOR_FIP(data);
	if(fip == NULL)
	    return;

	if((fip->freeze_count > 0) || fip->processing)
	    return;

	find_combo = GTK_COMBO(fip->find_combo);
	results_clist = GTK_CLIST(fip->results_clist);

	/* Get search string and make a copy of it */
	search_string = STRDUP(gtk_entry_get_text(GTK_ENTRY(find_combo->entry)));
	if(search_string == NULL)
	    return;

	/* Throughness */
	if(GTK_TOGGLE_BUTTON_GET_ACTIVE(fip->match_all_radio))
	    throughness = 0;
	if(GTK_TOGGLE_BUTTON_GET_ACTIVE(fip->match_any_radio))
	    throughness = 1;
	if(GTK_TOGGLE_BUTTON_GET_ACTIVE(fip->match_phrase_radio))
	    throughness = 2;

	/* Case sensitive */
	case_sensitive = GTK_TOGGLE_BUTTON_GET_ACTIVE(fip->case_sensitive_toggle);

	/* Begin finding */
	EditorFIPSetBusy(fip, TRUE);

	EditorFIPSetStatusMessage(fip, "Searching...");
	EditorFIPSetStatusProgress(fip, 0.0f);

	gtk_clist_freeze(results_clist);

	/* Clear the results GtkCList */
	gtk_clist_clear(results_clist);

	/* Search */
	matches = EditorFIPDoFind(
	    fip,
	    fip->editor,
	    throughness,
	    search_string,
	    case_sensitive
	);

	g_free(search_string);

	gtk_clist_thaw(results_clist);

	/* Print match results */
	if(matches > 0)
	    s = g_strdup_printf(
		"%i %s",
		matches,
		(matches == 1) ? "match" : "matches"
	    );
	else
	    s = g_strdup("No matches");
	EditorFIPSetStatusMessage(fip, s);
	g_free(s);
	EditorFIPSetStatusProgress(fip, 0.0f);

	EditorFIPUpdateMenus(fip);

	EditorFIPSetBusy(fip, FALSE);
}

/*
 *	Clear callback.
 */
void EditorFIPClearCB(GtkWidget *widget, gpointer data)
{
	GtkCList *clist;
	GtkCombo *combo;
	editor_fip_struct *fip = EDITOR_FIP(data);
	if(fip == NULL)
	    return;

	if(fip->freeze_count > 0)
	    return;

	combo = GTK_COMBO(fip->find_combo);
	clist = GTK_CLIST(fip->results_clist);

	EditorFIPSetBusy(fip, TRUE);

	gtk_entry_set_text(GTK_ENTRY(combo->entry), "");

#if 0
/* Just clear the find combo's entry widget */
	gtk_clist_freeze(clist);
	gtk_clist_clear(clist);
	gtk_clist_thaw(clist);
#endif

	EditorFIPUpdateMenus(fip);
	EditorFIPSetBusy(fip, FALSE);
}

/*
 *	Stop button callback.
 */
void EditorFIPStopCB(GtkWidget *widget, gpointer data)
{
	editor_fip_struct *fip = EDITOR_FIP(data);
	if(fip == NULL)
	    return;

	if(fip->freeze_count > 0)
	    return;

	fip->stop_count++;
	if(fip->stop_count < 1)
	    fip->stop_count = 1;
}

/*
 *	Goto callback.
 */
void EditorFIPGotoCB(GtkWidget *widget, gpointer data)
{
	gint row;			/* Row on dialog's result clist */
	GtkWidget *toplevel;
	GtkCList *results_clist;	/* Dialog's results clist */
	GtkCTree *layout_ctree;
	GtkCTreeNode *branch;
	editor_item_struct *item;
	editor_struct *editor;
	editor_fip_item_struct *result_item;
	editor_fip_struct *fip = EDITOR_FIP(data);
	if(fip == NULL)
	    return;

	if(fip->freeze_count > 0)
	    return;

	editor = fip->editor;
	toplevel = editor->toplevel;
	layout_ctree = GTK_CTREE(editor->layout_ctree);
	results_clist = GTK_CLIST(fip->results_clist);

	/* Get the selected row on the results list */
	if(results_clist->selection_end == NULL)
	    return;
	row = (gint)results_clist->selection_end->data;
	if((row < 0) || (row >= results_clist->rows))
	    return;

	result_item = EDITOR_FIP_ITEM(gtk_clist_get_row_data(
	    results_clist,
	    row
	));
	if(result_item == NULL)
	    return;

	/* Get the EditorItem */
	item = result_item->editor_item;
	if(item == NULL)
	{
	    CDialogSetTransientFor(toplevel);
	    CDialogGetResponse(
"GoTo Failed",
"No data is avialable on the selected item.",
		NULL,
		CDIALOG_ICON_WARNING,
		CDIALOG_BTNFLAG_OK,
		CDIALOG_BTNFLAG_OK
	    );
	    CDialogSetTransientFor(NULL);
	    return;
	}


	/* Match the branch on the editor's layout ctree with the
	 * branch item pointer which is suppose to be a unique
	 * pointer on one of the branches.
	 */
	branch = gtk_ctree_find_by_row_data(
	    layout_ctree, NULL, (gpointer)item
	);
	if(branch == NULL)
	{
	    CDialogSetTransientFor(toplevel);
	    CDialogGetResponse(
"Non-Critical Internal error!",
"Tree branch on the editor's layout tree could not be\n\
found with the result's branch item pointer.",
"The goto procedure could not be completed due to\n\
an unfindable reference to the layout tree's branch\n\
by branch item pointer. It is possible that the\n\
branch and corresponding item data pointer no longer\n\
exist on the layout tree.",
		CDIALOG_ICON_ERROR,
		CDIALOG_BTNFLAG_OK | CDIALOG_BTNFLAG_HELP,
		CDIALOG_BTNFLAG_OK
	    );
	    CDialogSetTransientFor(NULL);
	    return;
	}

	/* Select the matched branch on the layout ctree */
	EditorSelect(editor, branch);

	/* Flush events to make sure selecting is done properly */
	while(gtk_events_pending() > 0)
	    gtk_main_iteration();

	/* Now handle by item type, note that item is valid since
	 * branch is not NULL according to gtk_ctree_find_by_row_data().
	 */
	switch(item->type)
	{
	  case EDITOR_ITEM_HEADER:
	    if((result_item->text_sel_start >= 0) &&
	       (result_item->text_sel_end >= 0)
	    )
	    {
		GtkText *text = GTK_TEXT(editor->header_comments_text);
		if(text != NULL)
		{
		    gtk_text_set_point(
			text,
			(guint)result_item->text_sel_start
		    );
		    gtk_editable_set_position(
			GTK_EDITABLE(text),
			result_item->text_sel_start
		    );
		    gtk_editable_select_region(
			GTK_EDITABLE(text),
			result_item->text_sel_start,
			result_item->text_sel_end
		    );
		}
	    }
	    break;

	  case EDITOR_ITEM_SECTION:
	    if((result_item->text_sel_start >= 0) &&
	       (result_item->text_sel_end >= 0)
	    )
	    {
		GtkText *text = (GtkText *)editor->section_text;
		if(text != NULL)
		{
		    gtk_text_set_point(
			text,
			(guint)result_item->text_sel_start
		    );
		    gtk_editable_set_position(
			GTK_EDITABLE(text),
			result_item->text_sel_start
		    );
		    gtk_editable_select_region(
			GTK_EDITABLE(text),
			result_item->text_sel_start,
			result_item->text_sel_end
		    );
		}
	    }
	    break;

	  case EDITOR_ITEM_FILE:
	    break;
	}

	/* Unmap the Find In Pages dialog */
	EditorFIPUnmap(fip);
}


/*
 *	Menu map callback.
 */
gint EditorFIPMenuMapCB(GtkWidget *widget, GdkEvent *event, gpointer data)
{
	GdkEventButton *button;
	GtkWidget *w;
	editor_fip_struct *fip = EDITOR_FIP(data);
	if((widget == NULL) || (fip == NULL))
	    return(TRUE);

	if(fip->freeze_count > 0)
	    return(TRUE);

	/* Is event type button press? If so then get pointer to button
	 * event structure.
	 */
	if((event->type == GDK_BUTTON_PRESS) ||
	   (event->type == GDK_BUTTON_RELEASE)
	)
	    button = (GdkEventButton *)event;
	else
	    button = NULL;

	/* Results clist? */
	if(widget == fip->results_clist)
	{
	    w = fip->results_menu;
	    if((button != NULL) && (w != NULL))
	    {
		if(button->button == 3)
		{
		    gint row, column;

		    /* Try to select branch first */
		    if(gtk_clist_get_selection_info(
			GTK_CLIST(widget),
			button->x, button->y,
			&row, &column
		    ))
		    {
			gtk_clist_select_row(
			    GTK_CLIST(widget), row, column
			);
		    }

		    /* Map menu */
		    gtk_menu_popup(
			GTK_MENU(w),
			NULL, NULL,
			NULL, NULL,
			button->button, button->time
		    );
		}
	    }
	}

	return(TRUE);
}

/*
 *	Results GtkCList "select_row" signal callback.
 */
void EditorFIPResultsListSelectCB(
	GtkCList *clist, gint row, gint column,
	GdkEventButton *event, gpointer data
)
{
	editor_fip_struct *fip = EDITOR_FIP(data);
	if((clist == NULL) || (fip == NULL))
	    return;

	if(fip->freeze_count > 0)
	    return;

	EditorFIPSetBusy(fip, TRUE);

	/* Scroll if row not completely visable */
	if(gtk_clist_row_is_visible(clist, row) !=
	    GTK_VISIBILITY_FULL
	)
	    gtk_clist_moveto(
		clist,
		row, 0,				/* Row, column */
		0.5f, 0.0f			/* Row, column */
	    );

	/* Double click? */
	if(event != NULL)
	{
	    if(event->type == GDK_2BUTTON_PRESS)
	    {
		/* Goto */
		EditorFIPGotoCB(GTK_WIDGET(clist), data);
	    }
	}

	EditorFIPUpdateMenus(fip);
	EditorFIPSetBusy(fip, FALSE);
}

/*
 *	Results GtkCList "unselect_row" signal callback.
 */
void EditorFIPResultsListUnselectCB(
	GtkCList *clist, gint row, gint column,
	GdkEventButton *event, gpointer data
)
{
	editor_fip_struct *fip = EDITOR_FIP(data);
	if((clist == NULL) || (fip == NULL))
	    return;
}


/*
 *	Results GtkCList "column_click" signal callback.
 */
void EditorFIPResultsListColumnClickCB(
	GtkCList *clist, gint column, gpointer data
)
{
	editor_fip_struct *fip = EDITOR_FIP(data);
	if((clist == NULL) || (fip == NULL))
	    return;

	if(fip->freeze_count > 0)
	    return;

	EditorFIPSetBusy(fip, TRUE);

	gtk_clist_freeze(clist);
	gtk_clist_set_sort_column(clist, column);
	gtk_clist_sort(clist);
	gtk_clist_thaw(clist);

	EditorFIPSetBusy(fip, FALSE);
}
