#include <stdlib.h>
#include <string.h>
#include <gtk/gtk.h>

#include "guiutils.h"
#include "cdialog.h"

#include "editor.h"
#include "editor_values.h"
#include "editor_syntax_highlight.h"

#include "manedit.h"
#include "me_op.h"
#include "config.h"


void EditorGetValues(
	editor_struct *editor,
	GtkCTreeNode *node
);
void EditorSetValues(
	editor_struct *editor,
	GtkCTreeNode *node
);
void EditorClearValues(editor_struct *editor);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)


/*
 *	Gets the values to display from the GtKCTreeNode.
 */
void EditorGetValues(
	editor_struct *editor,
	GtkCTreeNode *node
)
{
	gint i;
	GList *glist;
	GtkWidget *w, *contents_toplevel = NULL;
	GtkCTree *ctree;
	editor_item_type type = -1;
	editor_item_struct *item;
	me_core_struct *core;
	me_pixmaps_list_struct *pixmaps_list;

	if((editor == NULL) || (node == NULL))
	    return;

	editor->freeze_count++;

	ctree = GTK_CTREE(editor->layout_ctree);
	item = EditorBranchGetData(
	    ctree,
	    node
	);
	core = editor->core;
	pixmaps_list = &core->pixmaps_list;

	/* Get the EditorItem's type */
	if(item != NULL)
	    type = item->type;

	/* Map/unmap the panels */
	for(glist = editor->contents_toplevels_list, i = 0;
	    glist != NULL;
	    glist = g_list_next(glist), i++
	)
	{
	    w = (GtkWidget *)glist->data;
	    if(w == NULL)
		continue;

	    if(i == (gint)type)
	    {
		contents_toplevel = w;
		gtk_widget_show(w);
	    }
	    else
	    {
		gtk_widget_hide(w);
	    }
	}

	/* Load the data into the widgets from the EditorItem */
	switch(type)
	{
	  case EDITOR_ITEM_FILE:
	    /* Recreate the file panel widgets to display the current
	     * values of the selected manual page
	     */
	    if((item != NULL) && (contents_toplevel != NULL))
	    {
		const gint	border_major = 5;
/*				border_minor = 2; */
		gint font_size;
#if defined(PROG_LANGUAGE_POLISH)
		const gchar *font_encoding = "iso8859-2";
#else
		const gchar *font_encoding = "iso8859-1";
#endif
		gchar *font_name_h1_bold;
		GdkFont *font;
		GtkRcStyle *rcstyle;
		GtkStyle *style;
		GList *children_list = gtk_container_children(GTK_CONTAINER(contents_toplevel));
		GtkWidget	*parent,
				*parent2,
				*parent3;

		/* Delete all the child widgets */
		children_list = g_list_copy(children_list);
		if(children_list != NULL)
		{
		    g_list_foreach(
			children_list,
			(GFunc)gtk_widget_destroy,
			NULL
		    );
		    g_list_free(children_list);
		}


		/* Set up the contents toplevel GtkBox */
		parent = contents_toplevel;
		style = gtk_widget_get_style(parent);
		font = style->font;
		gtk_container_border_width(GTK_CONTAINER(parent), border_major);
		gtk_box_set_spacing(GTK_BOX(parent), border_major);

		font_size = GDK_FONT_GET_FONT_NAME_SIZE(font);
	        if(font_size < 3)
	            font_size = 3;

		font_name_h1_bold = g_strdup_printf(
"-adobe-helvetica-bold-r-normal-*-%i-*-*-*-*-*-%s",
		    font_size + 2,
	            font_encoding
		);


		/* Identification GtkFrame */
		w = gtk_frame_new("Identification");
		gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
		gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
		gtk_widget_show(w);
		parent2 = w;

		w = gtk_vbox_new(FALSE, border_major);
		gtk_container_border_width(GTK_CONTAINER(w), border_major);
		gtk_container_add(GTK_CONTAINER(parent2), w);
		gtk_widget_show(w);
		parent2 = w;

		/* Name */
		w = gtk_hbox_new(FALSE, border_major);
		gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
		gtk_widget_show(w);
		parent3 = w;

		w = gtk_pixmap_new(
		    pixmaps_list->manual_closed_32x32,
		    pixmaps_list->manual_closed_32x32_mask
		);
		gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
		gtk_widget_show(w);

		rcstyle = gtk_rc_style_new();
		rcstyle->font_name = STRDUP(font_name_h1_bold);
		w = gtk_label_new((item->name != NULL) ?
		    item->name : "*No Name*"
		);
		gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_LEFT);
		gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
		gtk_widget_modify_style(w, rcstyle);
		GTK_RC_STYLE_UNREF(rcstyle);
		gtk_widget_show(w);

		/* Location */
		w = gtk_hbox_new(FALSE, border_major);
		gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
		gtk_widget_show(w);
		parent3 = w;

		w = gtk_label_new("Location:");
		gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_RIGHT);
		gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
		gtk_widget_show(w);

		if(STRISEMPTY(item->path))
		{
		    w = gtk_label_new("*Not Saved*");
		}
		else
		{
		    gchar *s = g_dirname(item->path);
		    w = gtk_label_new(s);
		    g_free(s);
		}
		gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_LEFT);
		gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
		gtk_widget_show(w);

		/* Section */
		w = gtk_hbox_new(FALSE, border_major);
		gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
		gtk_widget_show(w);
		parent3 = w;

		w = gtk_label_new("Section:");
		gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_RIGHT);
		gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
		gtk_widget_show(w);

		if(STRISEMPTY(item->path))
		{
		    w = gtk_label_new("*Unknown*");
		}
		else
		{
		    gint section_num;
		    const gchar	*s,
				*ext = strchr(
			item->name,
			'.'
		    );
		    if(ext != NULL)
			ext++;
		    else
			ext = item->path;
		    section_num = ATOI(ext);
		    switch(section_num)
		    {
		      case 1:
			s = ME_MANPAGE_SECTION_NAME_1;
			break;
		      case 2:
			s = ME_MANPAGE_SECTION_NAME_2;
			break;
		      case 3:
			s = ME_MANPAGE_SECTION_NAME_3;
			break;
		      case 4:
			s = ME_MANPAGE_SECTION_NAME_4;
			break;
		      case 5:
			s = ME_MANPAGE_SECTION_NAME_5;
			break;
		      case 6:
			s = ME_MANPAGE_SECTION_NAME_6;
			break;
		      case 7:
			s = ME_MANPAGE_SECTION_NAME_7;
			break;
		      case 8:
			s = ME_MANPAGE_SECTION_NAME_8;
			break;
		      case 9:
			s = ME_MANPAGE_SECTION_NAME_9;
			break;
		      default:
			s = "*Unknown*";
			break;
		    }
		    w = gtk_label_new(s);
		}
		gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_LEFT);
		gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
		gtk_widget_show(w);


		/* Data GtkFrame */
		w = gtk_frame_new("Data");
		gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
		gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
		gtk_widget_show(w);
		parent2 = w;

		w = gtk_vbox_new(FALSE, border_major);
		gtk_container_border_width(GTK_CONTAINER(w), border_major);
		gtk_container_add(GTK_CONTAINER(parent2), w);
		gtk_widget_show(w);
		parent2 = w;

		/* Has Changes */
		w = gtk_hbox_new(FALSE, border_major);
		gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
		gtk_widget_show(w);
		parent3 = w;

		w = gtk_label_new("Has Changes:");
		gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_RIGHT);
		gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
		gtk_widget_show(w);

		w = gtk_label_new(
		    item->has_changes ? "Yes" : "No"
		);
		gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_LEFT);
		gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
		gtk_widget_show(w);

		/* Compression */
		w = gtk_hbox_new(FALSE, border_major);
		gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
		gtk_widget_show(w);
		parent3 = w;

		w = gtk_label_new("Compression:");
		gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_RIGHT);
		gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
		gtk_widget_show(w);

		if(STRISEMPTY(item->path))
		{
		    w = gtk_label_new("None");
		}
		else
		{
		    const gchar	*s,
				*ext = strrchr(
			item->path,
			'.'
		    );
		    if(ext == NULL)
			ext = item->path;
		    if(!g_strcasecmp(ext, ".gz"))
			s = "GZip";
		    else if(!g_strcasecmp(ext, ".bz2"))
			s = "BZip2";
		    else if(!g_strcasecmp(ext, ".Z"))
			s = "Z Compress";
		    else
			s = "None";
		    w = gtk_label_new(s);
		}
		gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_LEFT);
		gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
		gtk_widget_show(w);




		g_free(font_name_h1_bold);
	    }
	    break;

	  case EDITOR_ITEM_HEADER:
	    if(item != NULL)
	    {
		gint		i,
				len;
		gchar *line;
		GdkFont *font;
		GtkWidget *w;
		GtkText *text;

		if(item->header_name != NULL)
		    gtk_entry_set_text(
			GTK_ENTRY(editor->header_name_entry),
			item->header_name
		    );
		if(item->header_section_number != NULL)
		    gtk_entry_set_text(
			GTK_ENTRY(editor->header_section_number_entry),
			item->header_section_number
		    );
		if(item->header_version != NULL)
		    gtk_entry_set_text(
			GTK_ENTRY(editor->header_version_entry),
			item->header_version
		    );
		if(item->header_author != NULL)
		    gtk_entry_set_text(
			GTK_ENTRY(editor->header_author_entry),
			item->header_author
		    );
		if(item->header_catagory != NULL)
		    gtk_entry_set_text(
			GTK_ENTRY(editor->header_catagory_entry),
			item->header_catagory
		    );

		/* Comments */
		w = editor->header_comments_text;
		text = GTK_TEXT(w);
		len = gtk_text_get_length(text);
		gtk_text_set_point(text, 0);
		gtk_text_freeze(text);
		gtk_text_forward_delete(text, len);
		gtk_text_thaw(text);

		while(gtk_events_pending() > 0)
		    gtk_main_iteration();

		font = core->editable_font;
		gtk_text_freeze(GTK_TEXT(w));
		for(i = 0; i < item->nlines; i++)
		{
		    line = item->lines_list[i];
		    if(line == NULL)
			continue;

		    /* Insert this line */
		    gtk_text_insert(
			text,
			font,
			NULL,			/* Foreground color */
			NULL,			/* Background color */
			line,
			-1
		    );

		    /* Append a newline */
		    gtk_text_insert(
			text,
			font,
			NULL,			/* Foreground color */
			NULL,			/* Background color */
			"\n",
			-1
		    );
		}
		gtk_text_thaw(GTK_TEXT(w));
	    }
	    break;

	  case EDITOR_ITEM_SECTION:
	    if(item != NULL)
	    {
		gint		i,
				len;
		gchar *line;
		GdkFont *font;
		GtkWidget *w;
		GtkText *text;

		/* Section name */
		gtk_entry_set_text(
		    GTK_ENTRY(editor->section_name_entry),
		    (item != NULL) ? item->section_name : ""
		);

		/* Section text */
		w = editor->section_text;
		text = GTK_TEXT(w);
		len = gtk_text_get_length(text);
		gtk_text_set_point(text, 0);
		gtk_text_freeze(text);
		gtk_text_forward_delete(text, len);
		gtk_text_thaw(text);

		while(gtk_events_pending() > 0)
		    gtk_main_iteration();

		font = core->editable_font;
		gtk_text_freeze(text);
		for(i = 0; i < item->nlines; i++)
		{
		    line = item->lines_list[i];
		    if(line == NULL)
			continue;

		    /* Insert this line */
		    gtk_text_insert(
			text,
			font,
			NULL,			/* Foreground color */
			NULL,			/* Background color */
			line,
			-1
		    );

		    /* Append a newline */
		    gtk_text_insert( 
			text,
			font,
			NULL,			/* Foreground color */
			NULL,			/* Background color */
			"\n",
			-1
		    );
		}
		gtk_text_thaw(text);

		while(gtk_events_pending() > 0)
		    gtk_main_iteration();

		/* Syntax highlighting for entire buffer */
		EditorSyntaxHighlight(
		    editor,
		    w,
		    -1, -1, -1
		);
	    }
	    break;
	}

	editor->freeze_count--;
}

/*
 *	Sets the displayed values to the GtkCTreeNode.
 */
void EditorSetValues(
	editor_struct *editor,
	GtkCTreeNode *node
)
{
	GtkCTree *ctree;
	editor_item_struct *item;
	me_pixmaps_list_struct *pixmaps_list;
	me_core_struct *core;

	if((editor == NULL) || (node == NULL))
	    return;

	editor->freeze_count++;

	ctree = GTK_CTREE(editor->layout_ctree);
	core = editor->core;
	pixmaps_list = &core->pixmaps_list;

	item = EditorBranchGetData(
	    ctree,
	    node
	);
	if(item == NULL)
	{
	    /* No data to apply */
	    editor->freeze_count--;
	    return;
	}

	/* Check ctree node node data type */
	switch(item->type)
	{
	  case EDITOR_ITEM_FILE:

	    break;
			    
	  case EDITOR_ITEM_HEADER:
	    if(item != NULL)
	    {
		gchar *s, **strv;
		GtkWidget *w;

		g_free(item->header_name);
		item->header_name = STRDUP(gtk_entry_get_text(GTK_ENTRY(
		    editor->header_name_entry
		)));

		g_free(item->header_section_number);
		item->header_section_number = STRDUP(gtk_entry_get_text(GTK_ENTRY(
		    editor->header_section_number_entry
		)));

		g_free(item->header_version);
		item->header_version = STRDUP(gtk_entry_get_text(GTK_ENTRY(
		    editor->header_version_entry
		)));

		g_free(item->header_author);
		item->header_author = STRDUP(gtk_entry_get_text(GTK_ENTRY(
		    editor->header_author_entry
		)));

		g_free(item->header_catagory); 
		item->header_catagory = STRDUP(gtk_entry_get_text(GTK_ENTRY(
		    editor->header_catagory_entry
		)));

		/* Comments */
		w = editor->header_comments_text;

		/* Delete the existing lines list */
		g_strfreev(item->lines_list);
		item->lines_list = NULL;
		item->nlines = 0;

		/* Get the new lines list */
		s = gtk_editable_get_chars(
		    GTK_EDITABLE(w),
		    0, -1
		);
		strv = g_strsplit(s, "\n", -1);
		g_free(s);
		if(strv != NULL)
		{
		    gint i;
		    item->lines_list = strv;
		    for(i = 0; strv[i] != NULL; i++);
		    item->nlines = i;
		}
	    }
	    break;

	  case EDITOR_ITEM_SECTION:
	    if(item != NULL)
	    {
		gchar *s, **strv;
		GtkWidget *w;

		g_free(item->section_name);
		item->section_name = STRDUP(gtk_entry_get_text(GTK_ENTRY(
		    editor->section_name_entry
		)));

		if(item->section_name != NULL)
		    gtk_ctree_node_set_pixtext(
			ctree,
			node,
			0,
			item->section_name,
			ME_LIST_ICON_TEXT_SPACING,
			pixmaps_list->manpage_section_20x20,
			pixmaps_list->manpage_section_20x20_mask
		    );

		w = editor->section_text;

		/* Delete the existing lines list */
		g_strfreev(item->lines_list);
		item->lines_list = NULL;
		item->nlines = 0;

		/* Get the new lines list */
		s = gtk_editable_get_chars(GTK_EDITABLE(w), 0, -1);
		strv = g_strsplit(s, "\n", -1);
		g_free(s);
		if(strv != NULL)
		{
		    gint i;
		    item->lines_list = strv;
		    for(i = 0; strv[i] != NULL; i++);
		    item->nlines = i;
		}
	    }
	    break;
	}

	editor->freeze_count--;
}

/*
 *	Clears all the displayed values.
 */
void EditorClearValues(editor_struct *editor)
{
	gint len;
	GtkWidget *w;
	GtkText *text;

	if(editor == NULL)
	    return;

	editor->freeze_count++;

	/* Unmap all the panels */
	g_list_foreach(
	    editor->contents_toplevels_list,
	    (GFunc)gtk_widget_hide,
	    NULL
	);


	w = (GtkWidget *)g_list_nth_data(
	    editor->contents_toplevels_list,
	    (guint)EDITOR_ITEM_FILE
	);
	if(w != NULL)
	{
	    /* Delete all the child widgets */
	    GList *children_list = gtk_container_children(GTK_CONTAINER(w));
	    children_list = g_list_copy(children_list);
	    if(children_list != NULL)
	    {
		g_list_foreach(
		    children_list,
		    (GFunc)gtk_widget_destroy,
		    NULL
		);
		g_list_free(children_list);
	    }
	}

	/* Header */
	gtk_entry_set_text(
	    GTK_ENTRY(editor->header_name_entry),
	    ""
	);
	gtk_entry_set_text(
	    GTK_ENTRY(editor->header_section_number_entry),
	    ""
	);
	gtk_entry_set_text(
	    GTK_ENTRY(editor->header_version_entry),
	    ""
	);
	gtk_entry_set_text(
	    GTK_ENTRY(editor->header_author_entry),
	    ""
	);
	gtk_entry_set_text(
	    GTK_ENTRY(editor->header_catagory_entry),
	    ""
	);

	w = editor->header_comments_text;
	text = GTK_TEXT(w);
	len = gtk_text_get_length(GTK_TEXT(w));
	gtk_text_set_point(text, 0);
	gtk_text_freeze(text);
	gtk_text_forward_delete(text, len);
	gtk_text_thaw(text);


	/* Section */
	gtk_entry_set_text(
	    GTK_ENTRY(editor->section_name_entry),
	    ""
	);
	w = editor->section_text;
	text = GTK_TEXT(w);
	len = gtk_text_get_length(GTK_TEXT(w));
	gtk_text_set_point(text, 0);
	gtk_text_freeze(text);
	gtk_text_forward_delete(text, len);
	gtk_text_thaw(text);

	editor->freeze_count--;
}
