/*
                              ManEdit Core
 */

#ifndef MANEDIT_H
#define MANEDIT_H

#include <gtk/gtk.h>

#include "cfg.h"

#include "about_dlg.h"
#include "editor.h"
#include "viewer.h"
#include "me_types.h"


/*
 *	Global switch to use gtk_text_*() function for GtkText widget
 *	text deleting instead of gtk_editable_*() functions.
 */
extern gboolean		use_text_delete;


/*
 *	Pixmaps list structure:
 */
typedef struct {

	GdkPixmap       *manedit_20x20;
	GdkBitmap       *manedit_20x20_mask;

	GdkPixmap       *editor_20x20;
	GdkBitmap       *editor_20x20_mask;

	GdkPixmap       *viewer_20x20;
	GdkBitmap       *viewer_20x20_mask;

	GdkPixmap       *folder_closed_20x20;
	GdkBitmap       *folder_closed_20x20_mask;

        GdkPixmap       *folder_opened_20x20;
        GdkBitmap       *folder_opened_20x20_mask;

	GdkPixmap	*manual_closed_20x20;
	GdkBitmap	*manual_closed_20x20_mask;

	GdkPixmap	*manual_closed_32x32;
	GdkBitmap	*manual_closed_32x32_mask;

	GdkPixmap	*manual_opened_20x20;
	GdkBitmap	*manual_opened_20x20_mask;

	GdkPixmap	*manpage_heading_20x20;
	GdkBitmap	*manpage_heading_20x20_mask;

	GdkPixmap	*manpage_section_20x20;
	GdkBitmap	*manpage_section_20x20_mask;

	GdkPixmap	*options_20x20;
	GdkBitmap	*options_20x20_mask;

} me_pixmaps_list_struct;


/*
 *	ManEdit Core:
 */
struct _me_core_struct {

	CfgList		*cfg_list;
	gchar		*cfg_file;

	gint		untitled_count;

	guint		main_toid;
	guint		reconfigured_idle_id;

	gboolean	need_close_all_windows;

	/* GdkCursors */
	GdkCursor	*busy_cursor,
			*text_cursor;

	/* GtkRcStyles */
	GtkRcStyle	*editable_rcstyle,
			*manpage_rcstyle;

	/* GdkFonts */
	GdkFont		*editable_font,
			*manpage_normal_font,
			*manpage_bold_font,
			*manpage_underline_font;

	me_pixmaps_list_struct	pixmaps_list;

	/* About Dialog  */
	about_dialog_struct	*about_dialog;

	/* Editors */
	editor_struct	**editor;
	gint		total_editors;

	/* Viewers */
	viewer_struct	**viewer;
	gint		total_viewers;

	/* Options Window */
	GtkWidget	*options_window;
};


extern me_core_struct *MENew(
	gint argc, gchar **argv,
	gint *status_rtn
);
extern void MEDelete(
	me_core_struct *core,
	const gboolean save_cfg
);


#endif	/* MANEDIT_H */
