#ifndef MANPAGE_VIEW_H
#define MANPAGE_VIEW_H

#include <gtk/gtk.h>
#include "ca_string.h"


typedef struct _MPGeometry		MPGeometry;
#define MP_GEOMETRY(p)			((MPGeometry *)(p))

typedef struct _MPVString		MPVString;
#define MPV_STRING(p)			((MPVString *)(p))

typedef struct _MPVObject		MPVObject;
#define MPV_OBJECT(p)			((MPVObject *)(p))
#define MPV_OBJECT_TARGET_TYPE		"MPView/Object"

typedef struct _MPView			MPView;
#define MP_VIEW(p)			((MPView *)(p))
#define MP_VIEW_KEY			"MPView"


/*
 *	MPGeometry:
 */
struct _MPGeometry {
	gint		x, y,
			width, height;
};


/* MPVString */
extern MPVString *MPVStringNew(void);
extern void MPVStringDelete(MPVString *ms);

extern gboolean MPVStringIsEmpty(MPVString *ms);

extern void MPVStringClearLineEnds(MPVString *ms);

extern gchar *MPVStringToRegularString(
	MPVString *ms,
	const gint len
);

extern void MPVStringSegmentAddAttribute(
	MPVString *ms,
	const gint start_i, const gint end_i,
	const CAStringAttributes attr
);
extern void MPVStringSegmentRemoveAttribute(
	MPVString *ms,
	const gint start_i, const gint end_i,
	const CAStringAttributes attr
);

extern void MPVStringSegmentRemoveAttribute(
	MPVString *ms,
	const gint start_i, const gint end_i,
	const CAStringAttributes attr
);

extern void MPVStringSelect(
	MPVString *ms,
	const gint selection_start_pos,
	const gint selection_end_pos,
	const gboolean clear_existing_selection
);
extern void MPVStringUnselectAll(MPVString *ms);

extern void MPVStringInsertCharacter(
	MPVString *ms,
	const gint i,
	const gchar c,
	const CAStringAttributes attr
);
extern void MPVStringInsertText(
	MPVString *ms,
	const gint i,
	const CAString *ca_string
);
extern void MPVStringRemoveCharacter(
	MPVString *ms,
	const gint i
);
extern void MPVStringRemoveText(
	MPVString *ms,
	const gint i,
	const gint len
);


/* Translate Coordinates */
extern gboolean MPViewTranslateObjectCoordinatesToIndex(
	GtkWidget *w,
	const gint obj_num,
	const gint string_num,
	const gint x, const gint y,
	gint *cursor_pos_rtn
);
extern gboolean MPViewTranslateCoordinatesToIndex(
	GtkWidget *w,
	const gint x, const gint y,
	gint *obj_num_rtn,
	gint *string_num_rtn,
	gint *cursor_pos_rtn
);
extern gboolean MPViewTranslateCoordinatesToLine(
	GtkWidget *w,
	const gint y,
        const gint start_obj_num,
        const gint start_string_num,
	gint *obj_num_rtn,
	gint *string_num_rtn,
	gint *line_num_rtn
);

/* XML I/O */
extern gint MPViewInsertObjectsFromXMLString(
	GtkWidget *w,
	const gint i,
	const gchar *s,
        gint progress_cb(const gulong, const gulong, gpointer),
        gpointer progress_data
);
extern gchar *MPViewGetObjectXMLString(
	GtkWidget *w,
	MPVObject *o
);

/* Set Cursor Position */
extern void MPViewSetCursorPosition(
	GtkWidget *w,
	const gint obj_num,
	const gint string_num,
	const gint cursor_pos
);
extern void MPViewGetCursorPosition(
	GtkWidget *w,
	gint *obj_num_rtn,
	gint *string_num_rtn,
	gint *cursor_pos_rtn,
	MPGeometry **cursor_geo_rtn
);

/* Select */
extern void MPViewSelect(
	GtkWidget *w,
	const gint obj_num,
	const gint string_num,
	const gint selection_start_pos,
	const gint selection_end_pos
);
extern void MPViewSelectObjectStringAll(
	GtkWidget *w,
	const gint obj_num,
	const gint string_num
);
extern void MPViewSelectObjectsRange(
	GtkWidget *w,
	const gint start_obj_num,
	const gint start_string_num,
	const gint start_cursor_pos,
	const gint end_obj_num,
	const gint end_string_num,
	const gint end_cursor_pos
);
extern void MPViewSelectObjectAll(
	GtkWidget *w,
	const gint obj_num
);

/* Unselect */
extern void MPViewUnselectAll(GtkWidget *w);
extern void MPViewUnselectObjectStringAll(
	GtkWidget *w,
	const gint obj_num,
	const gint string_num
);
extern void MPViewUnselectObjectAll(
	GtkWidget *w,
	const gint obj_num
);

/* Remove Selected Characters */
extern void MPViewRemoveSelectedCharacters(GtkWidget *w);

/* Visibility */
extern GdkVisibilityState MPViewGetObjectVisibility(
	GtkWidget *w,
	const gint obj_num
);
extern GdkVisibilityState MPViewGetCursorVisibilityX(GtkWidget *w);
extern GdkVisibilityState MPViewGetCursorVisibilityY(GtkWidget *w);
extern GdkVisibilityState MPViewGetCursorVisibility(GtkWidget *w);

/* Move To */
extern void MPViewMoveToObject(
	GtkWidget *w,
	const gint obj_num,
	const gfloat x_coeff, const gfloat y_coeff
);
extern void MPViewMoveToCursorX(
	GtkWidget *w,
	const gfloat x_coeff
);
extern void MPViewMoveToCursorY(
	GtkWidget *w,
	const gfloat y_coeff
);
extern void MPViewMoveToCursor(
	GtkWidget *w,
	const gfloat x_coeff, const gfloat y_coeff
);

/* MPView Editing */
extern void MPViewCutString(GtkWidget *w);
extern void MPViewCopyString(GtkWidget *w);
extern void MPViewCutObjects(GtkWidget *w);
extern void MPViewCopyObjects(GtkWidget *w);
extern void MPViewPaste(GtkWidget *w);
extern void MPViewDelete(GtkWidget *w);

/* MPView */
extern GtkWidget *MPViewNew(void);
extern void MPViewResize(
	GtkWidget *w,
	const gint width, const gint height
);
extern void MPViewUpdateCursorCoordinates(GtkWidget *w);
extern void MPViewDrawLineBar(GtkWidget *w);
extern void MPViewQueueResize(GtkWidget *w);
extern void MPViewDraw(GtkWidget *w);
extern void MPViewQueueDraw(GtkWidget *w);


#endif	/* MANPAGE_VIEW_H */
