#include <string.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "guiutils.h"

#include "ca_string.h"
#include "manpage_view.h"
#include "manpage_view_private.h"


void MPViewDestroyCB(gpointer data);
void MPViewLineBarRealizeCB(GtkWidget *widget, gpointer data);
gint MPViewLineBarConfigureEventCB(
	GtkWidget *widget, GdkEventConfigure *configure, gpointer data
);
gint MPViewLineBarExposeEventCB(
	GtkWidget *widget, GdkEventExpose *expose, gpointer data
);
gint MPViewLineBarFocusEventCB(
	GtkWidget *widget, GdkEventFocus *focus, gpointer data
);
gint MPViewLineBarKeyEventCB(
	GtkWidget *widget, GdkEventKey *key, gpointer data
);
gint MPViewLineBarButtonEventCB(
	GtkWidget *widget, GdkEventButton *button, gpointer data
);
void MPViewLineBarMotionEventCBUpdateSelectionIterate(
	MPView *v,
	const gint x, const gint y,
	const GdkModifierType state
);
gint MPViewLineBarMotionEventCB(
	GtkWidget *widget, GdkEventMotion *motion, gpointer data
);
void MPViewRealizeCB(GtkWidget *widget, gpointer data);
gint MPViewConfigureEventCB(
	GtkWidget *widget, GdkEventConfigure *configure, gpointer data
);
gint MPViewExposeEventCB(
	GtkWidget *widget, GdkEventExpose *expose, gpointer data
);
gint MPViewFocusEventCB(
	GtkWidget *widget, GdkEventFocus *focus, gpointer data
);
gint MPViewKeyEventCB(
	GtkWidget *widget, GdkEventKey *key, gpointer data
);
gint MPViewButtonEventCB(
	GtkWidget *widget, GdkEventButton *button, gpointer data
);
void MPViewMotionEventCBUpdateSelectionIterate(
	MPView *v,
	const gint x, const gint y,
	const GdkModifierType state
);
gint MPViewMotionEventCB(
	GtkWidget *widget, GdkEventMotion *motion, gpointer data
);
void MPViewAdjustmentValueChangedCB(
	GtkAdjustment *adj, gpointer data
);
gint MPViewResizeIdleCB(gpointer data);
gint MPViewLineBarScrollTOCB(gpointer data);
gint MPViewScrollTOCB(gpointer data);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)


/*
 *	MPView "destroy" signal callback.
 */
void MPViewDestroyCB(gpointer data)
{
	GdkColormap *colormap;
	MPView *v = MP_VIEW(data);
	if(v == NULL)
	    return;

/*
g_print("MPViewDestroyCB(): %p\n", v);
 */

	v->freeze_count++;

	if(v->resize_idle_id != 0)
	{
	    gtk_idle_remove(v->resize_idle_id);
	    v->resize_idle_id = 0;
	}
	if(v->scroll_toid != 0)
	{
	    gtk_timeout_remove(v->scroll_toid);
	    v->scroll_toid = 0;
	}

	if(v->objects_list != NULL)
	{
	    g_list_foreach(
		v->objects_list,
		(GFunc)MPVObjectDelete,
		NULL
	    );
	    g_list_free(v->objects_list);
	}

	GTK_OBJECT_UNREF(v->vadjustment);
	GTK_OBJECT_UNREF(v->hadjustment);

	GDK_GC_UNREF(v->gc);

	colormap = v->colormap;
#define FREE_COLOR(_c_)	{			\
 if((_c_) != NULL) {				\
  if(colormap != NULL) {			\
   GDK_COLORMAP_FREE_COLOR(colormap, (_c_));	\
  }						\
  g_free(_c_);					\
 }						\
}
	FREE_COLOR(v->normal_color);
	FREE_COLOR(v->bold_color);
	FREE_COLOR(v->underline_color);
	FREE_COLOR(v->heading_color);
	FREE_COLOR(v->bg_color);
#undef FREE_COLOR
	GDK_COLORMAP_UNREF(colormap);

	GDK_FONT_UNREF(v->normal_font);
	GDK_FONT_UNREF(v->bold_font);
	GDK_FONT_UNREF(v->underline_font);
	GDK_FONT_UNREF(v->heading_font);

	GDK_PIXMAP_UNREF(v->pixmap);
	GDK_PIXMAP_UNREF(v->line_bar_pixmap);

	GDK_CURSOR_DESTROY(v->text_cursor);
	GDK_CURSOR_DESTROY(v->line_bar_cursor);

	GDK_BITMAP_UNREF(v->line_break_bitmap);
	GDK_BITMAP_UNREF(v->string_end_bitmap);

	v->freeze_count--;

	g_free(v);
}

/*
 *	Line Bar GtkDrawingArea "realize" signal callback.
 */
void MPViewLineBarRealizeCB(GtkWidget *widget, gpointer data)
{
	GdkWindow *window;
	MPView *v = MP_VIEW(data);
	if((widget == NULL) || (v == NULL))
	    return;

	window = widget->window;

	gdk_window_set_cursor(window, v->line_bar_cursor);

	/* Wait for MPViewRealizeCB to be called to set the
	 * MP_VIEW_REALIZED flag
	 */
}

/*
 *	Line Bar GtkDrawingArea "configure_event" signal callback.
 */
gint MPViewLineBarConfigureEventCB(
	GtkWidget *widget, GdkEventConfigure *configure, gpointer data
)
{
	MPView *v = MP_VIEW(data);
	if((widget == NULL) || (configure == NULL) || (v == NULL))
	    return(FALSE);

	if((configure->width > 0) && (configure->height > 0))
	{
	    GDK_PIXMAP_UNREF(v->line_bar_pixmap);
	    v->line_bar_pixmap = gdk_pixmap_new(
		configure->window,
		configure->width, configure->height,
		-1
	    );
	}
	MPViewQueueDraw(v->toplevel);

	return(TRUE);
}

/*
 *	Line Bar GtkDrawingArea "expose_event" signal callback.
 */
gint MPViewLineBarExposeEventCB(
	GtkWidget *widget, GdkEventExpose *expose, gpointer data
)
{
	MPView *v = MP_VIEW(data);
	if((widget == NULL) || (expose == NULL) || (v == NULL))
	    return(FALSE);

	MPViewDrawLineBar(v->toplevel);

	return(TRUE);
}

/*
 *	Line Bar GtkDrawingArea "focus_in_event" or "focus_out_event"
 *	signal callback.
 */
gint MPViewLineBarFocusEventCB(
	GtkWidget *widget, GdkEventFocus *focus, gpointer data
)
{
	MPView *v = MP_VIEW(data);
	if((widget == NULL) || (focus == NULL) || (v == NULL))
	    return(FALSE);

	switch((gint)focus->type)
	{
	  case GDK_FOCUS_CHANGE:
	    if(focus->in)
		GTK_WIDGET_SET_FLAGS(widget, GTK_HAS_FOCUS);
	    else
		GTK_WIDGET_UNSET_FLAGS(widget, GTK_HAS_FOCUS);
	    MPViewQueueDraw(v->toplevel);
	    break;
	}

	return(TRUE);
}

/*
 *	Line Bar GtkDrawingArea "key_press_event" or
 *	"key_release_event" signal callback.
 */
gint MPViewLineBarKeyEventCB(
	GtkWidget *widget, GdkEventKey *key, gpointer data
)
{
	GdkEventKey e;
	GtkWidget *w;
	MPView *v = MP_VIEW(data);
	if((widget == NULL) || (key == NULL) || (v == NULL))
	    return(FALSE);

	if(v->freeze_count > 0)
	    return(FALSE);

	w = v->da;

	/* Line bar key events should be forwarded to the view
	 * GtkDrawingArea's MPViewKeyEventCB()
	 */
	e.type = key->type;
	e.window = w->window;
	e.send_event = TRUE;			/* Is send event */
	e.time = key->time;
	e.state = key->state;
	e.keyval = key->keyval;
	e.length = key->length;
	e.string = key->string;

	gtk_signal_emit_stop_by_name(
	    GTK_OBJECT(widget),
	    (key->type == GDK_KEY_PRESS) ?
		"key_press_event" : "key_release_event"
	);

	return(MPViewKeyEventCB(w, &e, v));
}

/*
 *	Line Bar GtkDrawingArea "button_press_event" or
 *	"button_release_event" signal callback.
 */
gint MPViewLineBarButtonEventCB(
	GtkWidget *widget, GdkEventButton *button, gpointer data
)
{
	MPView *v = MP_VIEW(data);
	if((widget == NULL) || (button == NULL) || (v == NULL))
	    return(FALSE);

	if(v->freeze_count > 0)
	    return(FALSE);

	switch((gint)button->type)
	{
	  case GDK_BUTTON_PRESS:
	    if(!GTK_WIDGET_HAS_FOCUS(widget))
		gtk_widget_grab_focus(widget);
	    switch(button->button)
	    {
	      case GDK_BUTTON1:
		if(button->state & GDK_CONTROL_MASK)
		{

		}
		else if(button->state & GDK_SHIFT_MASK)
		{
		    /* Get the previous cursor position */
		    MPViewGetCursorPosition(
			v->toplevel,
			&v->last_drag_obj_num,
			&v->last_drag_string_num,
			&v->last_drag_cursor_pos,
			NULL			/* No cursor geometry return */
		    );
		}
		else
		{
		    MPViewUnselectAll(v->toplevel);
		}
		if(TRUE)
		{
		    gint	obj_num,
				string_num,
				line_num;

		    /* Get the new MPVObject, MPVString and line number
		     * that was clicked on
		     */
		    if(MPViewTranslateCoordinatesToLine(
			v->toplevel,
			(gint)button->y,
			0,			/* Start object */
			0,			/* Start string */
			&obj_num,
			&string_num,
			&line_num
		    ))
		    {
			MPVObject *o = MPViewGetObject(v, obj_num);
			MPVString *ms = MPVObjectGetString(o, string_num);
			if((o != NULL) && (ms != NULL))
			{
			    gint	i,
					cursor_pos;
			    CAString *as_ptr = ms->ca_string;

			    /* Seek to the pointer to the line number
			     * excluding the count of the first line
			     */
			    for(i = 0; i < line_num; i++)
				as_ptr = CAStringNextLine(as_ptr);

			    cursor_pos = CAStringLengthPointers(
				ms->ca_string,
				as_ptr
			    );

			    /* Select between the previous cursor
			     * position and the new cursor position
			     */
			    if(button->state & GDK_SHIFT_MASK)
			    {
				MPViewSelectObjectsRange(
				    v->toplevel,
				    v->last_drag_obj_num,
				    v->last_drag_string_num,
				    v->last_drag_cursor_pos,
				    obj_num,
				    string_num,
				    cursor_pos
				);
				/* Leave the last drag positions as they
				 * were for this case
				 */
			    }
			    else
			    {
				v->last_drag_obj_num = obj_num;
				v->last_drag_string_num = string_num;
				v->last_drag_cursor_pos = cursor_pos;
			    }
			}
		    }

		    MPViewLineBarMotionEventCBUpdateSelectionIterate(
			v,
			(gint)button->x, (gint)button->y,
			button->state
		    );

		    /* Focus cell, set selection, and mark dragging */
		    if(!(v->flags & MP_VIEW_DRAG_SELECT))
		    {
			v->flags |= MP_VIEW_DRAG_SELECT;
			v->last_motion_x = (gint)button->x;
			v->last_motion_y = (gint)button->y;
			gdk_pointer_grab(
			    button->window,
			    FALSE,		/* Not owner events */
			    GDK_BUTTON_PRESS_MASK |
				GDK_BUTTON_RELEASE_MASK |
				GDK_POINTER_MOTION_MASK |
				GDK_POINTER_MOTION_HINT_MASK,
			    NULL,		/* Do not confine to a
						 * GdkWindow */
			    NULL,		/* No GdkCursor */
			    button->time
			);
		    }
		}
		break;

	      case GDK_BUTTON2:
		if(v->flags & MP_VIEW_READ_ONLY)
		    break;
		MPViewUnselectAll(v->toplevel);
		if(TRUE)
		{
		    gint	obj_num,
				string_num,
				line_num;

		    if(MPViewTranslateCoordinatesToLine(
			v->toplevel,
			(gint)button->y,
			0,			/* Start object */
			0,			/* Start string */
			&obj_num,
			&string_num,
			&line_num
		    ))
		    {
			MPVObject *o = MPViewGetObject(v, obj_num);
			MPVString *ms = MPVObjectGetString(o, string_num);
			if((o != NULL) && (ms != NULL))
			{
			    gint	i,
					cursor_pos;
			    CAString *as_ptr = ms->ca_string;

			    /* Seek to the pointer to the line number
			     * excluding the count of the first line
			     */
			    for(i = 0; i < line_num; i++)
				as_ptr = CAStringNextLine(as_ptr);

			    cursor_pos = CAStringLengthPointers(
				ms->ca_string,
				as_ptr
			    );

			    /* Update the cursor position */
			    MPViewSetCursorPosition(
				v->toplevel,
				obj_num,
				string_num,
				cursor_pos
			    );

			    /* Paste */
			    MPViewPaste(v->toplevel);
			}
		    }
		}
		break;

	      case GDK_BUTTON4:
		if(v->vadjustment != NULL)
		{
		    GtkAdjustment *adj = v->vadjustment;
		    const gfloat inc = MAX(
			(0.25f * adj->page_size),
			adj->step_increment
		    );
		    GTK_ADJUSTMENT_SET_VALUE(
			adj,
			adj->value - inc
		    );
		}
		break;

	      case GDK_BUTTON5:
		if(v->vadjustment != NULL)
		{
		    GtkAdjustment *adj = v->vadjustment;
		    const gfloat inc = MAX(
			(0.25f * adj->page_size),
			adj->step_increment
		    );
		    GTK_ADJUSTMENT_SET_VALUE(
			adj,
			adj->value + inc
		    );
		}
		break;
	    }
	    break;

	  case GDK_BUTTON_RELEASE:
	    switch(button->button)
	    {
	      case GDK_BUTTON1:
		if(v->flags & MP_VIEW_DRAG_SELECT)
		{
		    v->flags &= ~MP_VIEW_DRAG_SELECT;
		    v->last_motion_x = 0;
		    v->last_motion_y = 0;
		    v->last_drag_obj_num = -1;
		    v->last_drag_string_num = -1;
		    v->last_drag_cursor_pos = -1;
		    gdk_pointer_ungrab(button->time);
		}
		/* Remove the scroll timeout callback if it was active */
		if(v->scroll_toid != 0)
		{
		    gtk_timeout_remove(v->scroll_toid);
		    v->scroll_toid = 0;
		}
		/* Copy the selected objects to the selection */
		MPViewCopyObjects(v->toplevel);
		break;
	    }
	    break;
	}

	return(TRUE);
}

/*
 *	Called by MPViewLineBarButtonEventCB(),
 *	MPViewLineBarMotionEventCB(), and others to update the selection.
 *
 *	The MPView's selection, cursor position, and last drag values
 *	will be updated.
 *
 *	The x and y specifies the current pointer coordinates
 *	relative to the Line Bar GtkDrawingArea.
 */
void MPViewLineBarMotionEventCBUpdateSelectionIterate(
	MPView *v,
	const gint x, const gint y,
	const GdkModifierType state
)
{
	gint	start_obj_num = 0,
		start_string_num = 0,
		obj_num,
		string_num,
		line_num;
	MPVString *ms;
	MPVObject *o;

	/* Get the new MPVObject, MPVString and line number that the
	 * pointer is currently over
	 */
	while(MPViewTranslateCoordinatesToLine(
	    v->toplevel,
	    y,
	    start_obj_num,
	    start_string_num,
	    &obj_num,
	    &string_num,
	    &line_num
	))
	{
	    o = MPViewGetObject(v, obj_num);
	    ms = MPVObjectGetString(o, string_num);
	    if((o != NULL) && (ms != NULL))
	    {
		gint	i,
			cursor_pos,
			cursor_line_end_pos;
		CAString	*ca_string = ms->ca_string,
				*ca_string_ptr = ca_string;

		/* Seek to the pointer to the line number excluding
		 * the count of the first line
		 */
		for(i = 0; i < line_num; i++)
		    ca_string_ptr = CAStringNextLine(ca_string_ptr);

		/* Calculate the cursor position */
		cursor_pos = CAStringLengthPointers(
		    ca_string,
		    ca_string_ptr
		);

		/* Calculate the position of the end of the line that
		 * the cursor is on
		 */
		ca_string_ptr = CAStringNextLine(ca_string_ptr);
		cursor_line_end_pos = CAStringLengthPointers(
		    ca_string,
		    ca_string_ptr
		);

		/* If this is the first iteration then reselect every
		 * line from the drag origin to the cursor position
		 */
		if(start_obj_num == 0)
		{
		    MPVObject *o;
		    MPVString *ms;

		    MPViewUnselectAll(v->toplevel);
		    MPViewSelectObjectsRange(
			v->toplevel,
			v->last_drag_obj_num,
			v->last_drag_string_num,
			v->last_drag_cursor_pos,
			obj_num,
			string_num,
			cursor_pos
		    );

		    /* Select the first line */
		    o = MPViewGetObject(v, v->last_drag_obj_num);
		    ms = MPVObjectGetString(o, v->last_drag_string_num);
		    if(ms != NULL)
		    {
			const gint cursor_pos = v->last_drag_cursor_pos;
			gint cursor_line_end_pos;
			CAString	*ca_string = ms->ca_string,
					*ca_string_ptr = CAStringSeekToIndex(
					    ca_string,
					    cursor_pos
					);
			ca_string_ptr = CAStringNextLine(ca_string_ptr);
	                cursor_line_end_pos = CAStringLengthPointers(
	                    ca_string,
	                    ca_string_ptr
	                );
			MPVStringSelect(
			    ms,
			    cursor_pos,
			    cursor_line_end_pos,
			    FALSE		/* Do not clear the existing selection */
			);
		    }
		}

		/* Set the has changes flag on the MPVObject */
		o->flags |= MPV_OBJECT_HAS_SELECTION;

		/* Select this line */
		MPVStringSelect(
		    ms,
		    cursor_pos,
		    cursor_line_end_pos,
		    FALSE			/* Do not clear the existing selection */
		);

		/* Update the cursor position */
		MPViewSetCursorPosition(
		    v->toplevel,
		    obj_num,
		    string_num,
		    cursor_pos
		);
	    }

	    start_obj_num = obj_num;
	    start_string_num = string_num + 1;
	}
}

/*
 *	Line Bar GtkDrawingArea "motion_notify_event" callback.
 */
gint MPViewLineBarMotionEventCB(
	GtkWidget *widget, GdkEventMotion *motion, gpointer data
)
{
	GtkAdjustment	*hadj,
			*vadj;
	MPView *v = MP_VIEW(data);
	if((widget == NULL) || (motion == NULL) || (v == NULL))
	    return(FALSE);

	hadj = v->hadjustment;
	vadj = v->vadjustment;

	/* Request for further motion events if this is a hint and
	 * the pointer button 1 is pressed
	 */
	if(motion->is_hint &&
	   ((v->flags & MP_VIEW_DRAG_SELECT) ||
	    (v->flags & MP_VIEW_DRAG_SCROLL)
	   )
	)
	{
	    gint x, y;
	    GdkModifierType mask;
	    gdk_window_get_pointer(
		motion->window, &x, &y, &mask
	    );
	}

	/* Dragging select? */
	if(v->flags & MP_VIEW_DRAG_SELECT)
	{
	    const gint	x = (gint)motion->x,
			y = (gint)motion->y;

	    /* Update the selection */
	    MPViewLineBarMotionEventCBUpdateSelectionIterate(
		v,
		x, y,
		motion->state
	    );

	    /* Check if the motion has moved outside of the
	     * GtkDrawingArea in which case we should start or stop
	     * scrolling
	     */
	    if((y < 0) || (y >= (gint)vadj->page_size))
	    {
		if(v->scroll_toid == 0)
		    v->scroll_toid = gtk_timeout_add(
			180l,
			MPViewLineBarScrollTOCB, v
		    );
	    }
	    /* Inside the GtkDrawingArea */
	    else
	    {
		/* Remove the scroll timeout callback if it was active */
		if(v->scroll_toid != 0)
		{
		    gtk_timeout_remove(v->scroll_toid);
		    v->scroll_toid = 0;
		}
	    }
	}
	/* Dragging scroll? */
	else if(v->flags & MP_VIEW_DRAG_SCROLL)
	{
	    const gint dy = (gint)motion->y - v->last_motion_y;
	    if(dy != 0)
		GTK_ADJUSTMENT_SET_VALUE(
		    vadj,
		    vadj->value - (gfloat)dy
		);
	}

	/* Update the last motion values */
	v->last_motion_x = (gint)motion->x;
	v->last_motion_y = (gint)motion->y;

	return(TRUE);
}


/*
 *	View GtkDrawingArea "realize" signal callback.
 */
void MPViewRealizeCB(GtkWidget *widget, gpointer data)
{
	GdkWindow *window;
	MPView *v = MP_VIEW(data);
	if((widget == NULL) || (v == NULL))
	    return;

	window = widget->window;

	if(v->normal_font == NULL)
	    v->normal_font = gdk_font_load(
"-adobe-helvetica-medium-r-normal-*-12-*-*-*-p-*-iso8859-1"
	    );
	if(v->bold_font == NULL)
	    v->bold_font = gdk_font_load(
"-adobe-helvetica-bold-r-normal-*-12-*-*-*-p-*-iso8859-1"
	    );
	if(v->underline_font == NULL)
	    v->underline_font = gdk_font_load(
"-adobe-helvetica-medium-o-normal-*-12-*-*-*-p-*-iso8859-1"
	    );
	if(v->heading_font == NULL)
	    v->heading_font = gdk_font_load(
"-adobe-helvetica-bold-r-normal-*-14-*-*-*-p-*-iso8859-1"
	    );

	if(v->line_break_bitmap == NULL)
	{
	    const gint	width = 8,
			height = 8;
	    const guint8 data[] = {
		0x20,
		0x20,
		0x20,
		0x24,
		0x26,
		0x3F,
		0x06,
		0x04
	    };
	    v->line_break_bitmap = gdk_bitmap_create_from_data(
		window,
		(const gchar *)data,
		width, height
	    );
	}
	if(v->string_end_bitmap == NULL)
	{
	    const gint	width = 8,
			height = 8;
	    const guint8 data[] = {
		0x24,
		0x24,
		0xff,
		0x24,
		0x24,
		0xff,
		0x24,
		0x24
	    };
	    v->string_end_bitmap = gdk_bitmap_create_from_data(
		window,
		(const gchar *)data,
		width, height
	    );
	}

	if(v->colormap == NULL)
	{
	    GdkColormap *colormap = GDK_COLORMAP_REF(gdk_window_get_colormap(window));
	    GtkStyle *style = gtk_widget_get_style(widget);
	    if(style != NULL)
	    {
		GtkStateType state = GTK_STATE_NORMAL;
		/* Set default colors as needed */
		if(v->normal_color == NULL)
		    v->normal_color = g_memdup(
			&style->text[state],
			sizeof(GdkColor)
		    );
		if(v->bold_color == NULL)
		    v->bold_color = g_memdup(
			&style->text[state],
			sizeof(GdkColor)
		    );
		if(v->underline_color == NULL)
		    v->underline_color = g_memdup(
			&style->text[state],
			sizeof(GdkColor)
		    );
		if(v->heading_color == NULL)
		    v->heading_color = g_memdup(
			&style->text[state],
			sizeof(GdkColor)
		    );
		if(v->bg_color == NULL)
		    v->bg_color = g_memdup(
			&style->base[state],
			sizeof(GdkColor)
		    );
	    }
	    GDK_COLORMAP_ALLOC_COLOR(colormap, v->normal_color);
	    GDK_COLORMAP_ALLOC_COLOR(colormap, v->bold_color);
	    GDK_COLORMAP_ALLOC_COLOR(colormap, v->underline_color);
	    GDK_COLORMAP_ALLOC_COLOR(colormap, v->heading_color);
	    GDK_COLORMAP_ALLOC_COLOR(colormap, v->bg_color);
	    v->colormap = colormap;
	}

	if(v->gc == NULL)
	    v->gc = gdk_gc_new(window);

	gdk_window_set_cursor(window, v->text_cursor);

	v->flags |= MP_VIEW_REALIZED;
}

/*
 *	View GtkDrawingArea "configure_event" signal callback.
 */
gint MPViewConfigureEventCB(
	GtkWidget *widget, GdkEventConfigure *configure, gpointer data
)
{
	MPView *v = MP_VIEW(data);
	if((widget == NULL) || (configure == NULL) || (v == NULL))
	    return(FALSE);

	MPViewResize(
	    v->toplevel,
	    configure->width, configure->height
	);
	MPViewQueueDraw(v->toplevel);

	return(TRUE);
}

/*
 *	View GtkDrawingArea "expose_event" signal callback.
 */
gint MPViewExposeEventCB(
	GtkWidget *widget, GdkEventExpose *expose, gpointer data
)
{
	MPView *v = MP_VIEW(data);
	if((widget == NULL) || (expose == NULL) || (v == NULL))
	    return(FALSE);

	MPViewDraw(v->toplevel);

	return(TRUE);
}

/*
 *	View GtkDrawingArea "focus_in_event" or "focus_out_event"
 *	signal callback.
 */
gint MPViewFocusEventCB(
	GtkWidget *widget, GdkEventFocus *focus, gpointer data
)
{
	MPView *v = MP_VIEW(data);
	if((widget == NULL) || (focus == NULL) || (v == NULL))
	    return(FALSE);

	switch((gint)focus->type)
	{
	  case GDK_FOCUS_CHANGE:
	    if(focus->in)
		GTK_WIDGET_SET_FLAGS(widget, GTK_HAS_FOCUS);
	    else
		GTK_WIDGET_UNSET_FLAGS(widget, GTK_HAS_FOCUS);
	    MPViewQueueDraw(v->toplevel);
	    break;
	}

	return(TRUE);
}

/*
 *	View GtkDrawingArea "key_press_event" or "key_release_event"
 *	signal callback.
 */
gint MPViewKeyEventCB(
	GtkWidget *widget, GdkEventKey *key, gpointer data
)
{
	gboolean press;
	guint state;
	MPView *v = MP_VIEW(data);
	if((widget == NULL) || (key == NULL) || (v == NULL))
	    return(FALSE);

	if(v->freeze_count > 0)
	    return(FALSE);

#define REPORT_CHANGE	{			\
 if(v->changed_cb != NULL)			\
  v->changed_cb(				\
   v,						\
   v->changed_data				\
  );						\
}
#define STOP_KEY_SIGNAL_EMIT(_w_)	{	\
 if(!key->send_event) {				\
  gtk_signal_emit_stop_by_name(			\
   GTK_OBJECT(_w_),				\
   press ?					\
    "key_press_event" : "key_release_event"	\
  );						\
 }						\
}
#define MOVE_TO_FOCUS_OBJECT(_x_coeff_,_y_coeff_) { \
 if(MPViewGetObjectVisibility(			\
  v->toplevel,					\
  v->focus_obj_num				\
 ) != GDK_VISIBILITY_UNOBSCURED)		\
  MPViewMoveToObject(				\
   v->toplevel,					\
   v->focus_obj_num,				\
   (_x_coeff_), (_y_coeff_)			\
  );						\
}
#define MOVE_TO_CURSOR(_xc_,_yc_)	{	\
 /* Must updates the cursor's coordiantes */	\
 MPViewUpdateCursorCoordinates(v->toplevel);	\
 /* Scroll to the cursor if it is not fully	\
  * visible, check each dimension individually	\
  */						\
 if(MPViewGetCursorVisibilityX(v->toplevel) !=	\
     GDK_VISIBILITY_UNOBSCURED			\
 )						\
  MPViewMoveToCursorX(				\
   v->toplevel,					\
   (_xc_)					\
  );						\
						\
 MPViewUpdateCursorCoordinates(v->toplevel);	\
 if(MPViewGetCursorVisibilityY(v->toplevel) !=	\
     GDK_VISIBILITY_UNOBSCURED			\
 )						\
  MPViewMoveToCursorY(				\
   v->toplevel,					\
   (_yc_)					\
  );						\
}

	press = (key->type == GDK_KEY_PRESS) ? TRUE : FALSE;
	state = key->state;

	switch(key->keyval)
	{
	  case GDK_ISO_Left_Tab:
	    if(press)
	    {
		if(TRUE)
		{
		    if(v->focus_string_num > 0)
		    {
			v->focus_string_num--;
			MPViewUnselectAll(v->toplevel);
			MPViewSetCursorPosition(
			    v->toplevel,
			    v->focus_obj_num,
			    v->focus_string_num,
			    0
			);
			MOVE_TO_CURSOR(0.0f, 0.0f);
		    }
		    else if(v->focus_obj_num > 0)
		    {
			MPVObject *o;

			v->focus_obj_num--;
			o = MPViewGetObject(
			    v,
			    v->focus_obj_num
			);
			if(o != NULL)
			{
			    v->focus_string_num = g_list_length(o->strings_list) - 1;
			    MPViewUnselectAll(v->toplevel);
			    MPViewSetCursorPosition(
				v->toplevel,
				v->focus_obj_num,
				v->focus_string_num,
				0
			    );
			}
			MOVE_TO_CURSOR(0.0f, 0.0f);
		    }
		    else
		    {
			break;
		    }
		}
		STOP_KEY_SIGNAL_EMIT(widget);
	    }
	    break;

	  case GDK_Tab:
	    if(press)
	    {
		if(TRUE)
		{
		    const gint nobjs = g_list_length(v->objects_list);
		    MPVObject *o = MPViewGetObject(
			v,
			v->focus_obj_num
		    );
		    if(o != NULL)
		    {
			const gint nstrings = g_list_length(o->strings_list);
			if(v->focus_string_num < (nstrings - 1))
			{
			    v->focus_string_num++;
			    MPViewUnselectAll(v->toplevel);
			    MPViewSetCursorPosition(
				v->toplevel,
				v->focus_obj_num,
				v->focus_string_num,
				0
			    );
			    MOVE_TO_CURSOR(0.0f, 0.0f);
			}
			else if(v->focus_obj_num < (nobjs - 1))
			{
			    v->focus_obj_num++;
			    v->focus_string_num = 0;
			    MPViewUnselectAll(v->toplevel);
			    MPViewSetCursorPosition(
				v->toplevel,
				v->focus_obj_num,
				v->focus_string_num,
				0
			    );
			    MOVE_TO_CURSOR(0.0f, 0.0f);
			}
			else
			{
			    break;
			}
		    }
		}

		STOP_KEY_SIGNAL_EMIT(widget);
	    }
	    break;

	  case GDK_Up:
	  case GDK_KP_Up:
	    if(press)
	    {
		MPVObject *o = MPViewGetObject(
		    v,
		    v->focus_obj_num
		);
		if(o != NULL)
		{
		    MPVString *ms = MPVObjectGetString(
			o,
			v->focus_string_num
		    );
		    if(ms != NULL)
		    {
			/* Scroll only? */
			if(state & GDK_CONTROL_MASK)
			{
			    GtkAdjustment *adj = v->vadjustment;
			    GTK_ADJUSTMENT_SET_VALUE(
				adj,
				adj->value - adj->step_increment
			    );
			}
			/* Switch to the previous object? */
			else if(ms->cursor_pos == 0)
			{
			    if(v->focus_string_num > 0)
			    {
				v->focus_string_num--;
				MPViewSetCursorPosition(
				    v->toplevel,
				    v->focus_obj_num,
				    v->focus_string_num,
				    -1
				);
				MOVE_TO_CURSOR(1.0f, 0.0f);
			    }
			    else if(v->focus_obj_num > 0)
			    {
				v->focus_obj_num--;
				o = MPViewGetObject(
				    v,
				    v->focus_obj_num
				);
				if(o != NULL)
				    v->focus_string_num = g_list_length(o->strings_list) - 1;
				MPViewSetCursorPosition(
				    v->toplevel,
				    v->focus_obj_num,
				    v->focus_string_num,
				    -1
				);
				MOVE_TO_CURSOR(1.0f, 0.0f);
			    }

			    /* Clear the selection */
			    if(!(state & GDK_SHIFT_MASK))
				MPViewUnselectAll(v->toplevel);
			}
			else
			{
			    /* Move the cursor one line up */
			    const gint prev_cursor_pos = ms->cursor_pos;
			    gint cur_line_offset;
			    CAString	*ca_string = ms->ca_string,
					*cas_cur = CAStringSeekToIndex(
					    ca_string,
					    ms->cursor_pos
					),
					*cas_cur_line = CAStringCurrentLineStart(
					    ca_string,
					    cas_cur
					),
					*cas_prev_line;

			    cur_line_offset = (gint)(cas_cur - cas_cur_line);
			    cas_prev_line = CAStringPreviousLine(
				ca_string,
				cas_cur
			    );
			    if(cas_prev_line == cas_cur_line)
			    {
				ms->cursor_pos = 0;
			    }
			    else
			    {
				gchar c;
				gint i;
				CAString *ca_string_ptr = cas_prev_line;
				for(i = 0; i < cur_line_offset; i++)
				{
				    c = CA_STRING_CHARACTER(ca_string_ptr);
				    if((c == '\0') ||
				       (c == CA_STRING_CHARACTER_LINE_END) ||
				       (c == CA_STRING_CHARACTER_LINE_BREAK)
				    )
					break;

				    ca_string_ptr++;
				}
				ms->cursor_pos = (gint)(ca_string_ptr - ca_string);
			    }
			    if(ms->cursor_pos > ms->length)
				ms->cursor_pos = ms->length;
			    if(ms->cursor_pos < 0)
				ms->cursor_pos = 0;

			    /* Update the selection up? */
			    if(state & GDK_SHIFT_MASK)
			    {
				if(o->flags & MPV_OBJECT_HAS_SELECTION)
				{
				    if(ms->cursor_pos > ms->selection_end_pos)
					ms->selection_end_pos = ms->cursor_pos;
				    else
					ms->selection_start_pos = ms->cursor_pos;
				}
				else
				{
				    ms->selection_end_pos = prev_cursor_pos;
				    ms->selection_start_pos = ms->cursor_pos;
				}
				MPViewSelect(
				    v->toplevel,
				    v->focus_obj_num,
				    v->focus_string_num,
				    ms->selection_start_pos,
				    ms->selection_end_pos
				);
			    }
			    else
			    {
				MPViewUnselectAll(v->toplevel);
			    }
			    MOVE_TO_CURSOR(0.0f, 0.0f);
			}
		    }

		    STOP_KEY_SIGNAL_EMIT(widget);
		}
	    }
	    break;

	  case GDK_Down:
	  case GDK_KP_Down:
	    if(press)
	    {
		MPVObject *o = MPViewGetObject(
		    v,
		    v->focus_obj_num
		);
		if(o != NULL)
		{
		    MPVString *ms = MPVObjectGetString(
			o,
			v->focus_string_num
		    );
		    if(ms != NULL)
		    {
			const gint nobjs = g_list_length(v->objects_list);

			/* Scroll only? */
			if(state & GDK_CONTROL_MASK)
			{
			    GtkAdjustment *adj = v->vadjustment;
			    GTK_ADJUSTMENT_SET_VALUE(
				adj,
				adj->value + adj->step_increment
			    );
			}
			/* Switch to the next object? */
			else if(ms->cursor_pos == ms->length)
			{
			    const gint nstrings = g_list_length(o->strings_list);
			    if(v->focus_string_num < (nstrings - 1))
			    {
				v->focus_string_num++;
				MPViewSetCursorPosition(
				    v->toplevel,
				    v->focus_obj_num,
				    v->focus_string_num,
				    0
				);
				MOVE_TO_CURSOR(0.0f, 1.0f);
			    }
			    else if(v->focus_obj_num < (nobjs - 1))
			    {
				v->focus_obj_num++;
				v->focus_string_num = 0;
				MPViewSetCursorPosition(
				    v->toplevel,
				    v->focus_obj_num,
				    v->focus_string_num,
				    0
				);
				MOVE_TO_CURSOR(0.0f, 1.0f);
			    }

			    /* Clear the selection */
			    if(!(state & GDK_SHIFT_MASK))
				MPViewUnselectAll(v->toplevel);
			}
			else
			{
			    /* Move the cursor down one line */
			    const gint prev_cursor_pos = ms->cursor_pos;
			    gint cur_line_offset;
			    CAString	*ca_string = ms->ca_string,
					*cas_cur = CAStringSeekToIndex(
					    ca_string,
					    ms->cursor_pos
					),
					*cas_cur_line = CAStringCurrentLineStart(
					    ca_string,
					    cas_cur
					),
					*cas_next_line;
			    cur_line_offset = (gint)(cas_cur - cas_cur_line);
			    cas_next_line = CAStringNextLine(cas_cur);
			    if(CA_STRING_CHARACTER(cas_next_line) == '\0')
			    {
				ms->cursor_pos = ms->length;
			    }
			    else
			    {
				gchar c;
				gint i;
				CAString *ca_string_ptr = cas_next_line;
				for(i = 0; i < cur_line_offset; i++)
				{
				    c = CA_STRING_CHARACTER(ca_string_ptr);
				    if((c == '\0') ||
				       (c == CA_STRING_CHARACTER_LINE_END) ||
				       (c == CA_STRING_CHARACTER_LINE_BREAK)
				    )
					break;

				    ca_string_ptr++;
				}
				ms->cursor_pos = (gint)(ca_string_ptr - ca_string);
			    }
			    if(ms->cursor_pos > ms->length)
				ms->cursor_pos = ms->length;
			    if(ms->cursor_pos < 0)
				ms->cursor_pos = 0;

			    /* Update the selection down? */
			    if(state & GDK_SHIFT_MASK)
			    {
				if(o->flags & MPV_OBJECT_HAS_SELECTION)
				{
				    if(ms->cursor_pos < ms->selection_start_pos)
					ms->selection_start_pos = ms->cursor_pos;
				    else
					ms->selection_end_pos = ms->cursor_pos;
				}
				else
				{
				    ms->selection_start_pos = prev_cursor_pos;
				    ms->selection_end_pos = ms->cursor_pos;
				}
				MPViewSelect(
				    v->toplevel,
				    v->focus_obj_num,
				    v->focus_string_num,
				    ms->selection_start_pos,
				    ms->selection_end_pos
				);
			    }
			    else
			    {
				MPViewUnselectAll(v->toplevel);
			    }
			    MOVE_TO_CURSOR(0.0f, 1.0f);
			}
		    }

		    STOP_KEY_SIGNAL_EMIT(widget);
		}
	    }
	    break;

	  case GDK_Left:
	  case GDK_KP_Left:
	    if(press)
	    {
		MPVObject *o = MPViewGetObject(
		    v,
		    v->focus_obj_num
		);
		if(o != NULL)
		{
		    MPVString *ms = MPVObjectGetString(
			o,
			v->focus_string_num
		    );
		    if(ms != NULL)
		    {
			/* Scroll only? */
			if(state & GDK_CONTROL_MASK)
			{
			    GtkAdjustment *adj = v->hadjustment;
			    GTK_ADJUSTMENT_SET_VALUE(
				adj,
				adj->value - adj->step_increment
			    );
			}
			/* Switch to the previous object? */
			else if(ms->cursor_pos == 0)
			{
			    if(v->focus_string_num > 0)
			    {
				v->focus_string_num--;
				MPViewSetCursorPosition(
				    v->toplevel,
				    v->focus_obj_num,
				    v->focus_string_num,
				    -1
				);
				MOVE_TO_CURSOR(1.0f, 0.0f);
			    }
			    else if(v->focus_obj_num > 0)
			    {
				v->focus_obj_num--;
				o = MPViewGetObject(
				    v,
				    v->focus_obj_num
				);
				if(o != NULL)
				    v->focus_string_num = g_list_length(o->strings_list) - 1;
				MPViewSetCursorPosition(
				    v->toplevel,
				    v->focus_obj_num,
				    v->focus_string_num,
				    -1
				);
				MOVE_TO_CURSOR(1.0f, 0.0f);
			    }

			    /* Clear the selection */
			    if(!(state & GDK_SHIFT_MASK))
				MPViewUnselectAll(v->toplevel);
			}
			else
			{
			    /* Move the cursor left */
			    ms->cursor_pos--;
			    if(ms->cursor_pos > ms->length)
				ms->cursor_pos = ms->length;
			    if(ms->cursor_pos < 0)
				ms->cursor_pos = 0;

			    /* Update the selection to the left? */
			    if(state & GDK_SHIFT_MASK)
			    {
				if(o->flags & MPV_OBJECT_HAS_SELECTION)
				{
				    if(ms->cursor_pos < ms->selection_start_pos)
					ms->selection_start_pos = ms->cursor_pos;
				    else if(ms->cursor_pos != ms->selection_start_pos)
					ms->selection_end_pos = ms->cursor_pos;
				}
				else
				{
				    ms->selection_end_pos = ms->cursor_pos + 1;
				    ms->selection_start_pos = ms->cursor_pos;
				}
				MPViewSelect(
				    v->toplevel,
				    v->focus_obj_num,
				    v->focus_string_num,
				    ms->selection_start_pos,
				    ms->selection_end_pos
				);
			    }
			    else
			    {
				MPViewUnselectAll(v->toplevel);
			    }
			    MOVE_TO_CURSOR(0.0f, 0.0f);
			}
		    }

		    STOP_KEY_SIGNAL_EMIT(widget);
		}
	    }
	    break;

	  case GDK_Right:
	  case GDK_KP_Right:
	    if(press)
	    {
		MPVObject *o = MPViewGetObject(
		    v,
		    v->focus_obj_num
		);
		if(o != NULL)
		{
		    MPVString *ms = MPVObjectGetString(
			o,
			v->focus_string_num
		    );
		    if(ms != NULL)
		    {
			const gint nobjs = g_list_length(v->objects_list);

			/* Scroll only? */
			if(state & GDK_CONTROL_MASK)
			{
			    GtkAdjustment *adj = v->hadjustment;
			    GTK_ADJUSTMENT_SET_VALUE(
				adj,
				adj->value + adj->step_increment
			    );
			}
			/* Switch to the next object? */
			else if(ms->cursor_pos == ms->length)
			{
			    const gint nstrings = g_list_length(o->strings_list);
			    if(v->focus_string_num < (nstrings - 1))
			    {
				v->focus_string_num++;
				MPViewSetCursorPosition(
				    v->toplevel,
				    v->focus_obj_num,
				    v->focus_string_num,
				    0
				);
				MOVE_TO_CURSOR(0.0f, 1.0f);
			    }
			    else if(v->focus_obj_num < (nobjs - 1))
			    {
				v->focus_obj_num++;
				v->focus_string_num = 0;
				MPViewSetCursorPosition(
				    v->toplevel,
				    v->focus_obj_num,
				    v->focus_string_num,
				    0
				);
				MOVE_TO_CURSOR(0.0f, 1.0f);
			    }

			    /* Clear the selection */
			    if(!(state & GDK_SHIFT_MASK))
				MPViewUnselectAll(v->toplevel);
			}
			else
			{
			    /* Move the cursor to the right */
			    ms->cursor_pos++;
			    if(ms->cursor_pos > ms->length)
				ms->cursor_pos = ms->length;
			    if(ms->cursor_pos < 0)
				ms->cursor_pos = 0;

			    /* Update the selection to the right? */
			    if(state & GDK_SHIFT_MASK)
			    {
				if(o->flags & MPV_OBJECT_HAS_SELECTION)
				{
				    if(ms->cursor_pos > ms->selection_end_pos)
					ms->selection_end_pos = ms->cursor_pos;
				    else if(ms->cursor_pos != ms->selection_end_pos)
					ms->selection_start_pos = ms->cursor_pos;
				}
				else
				{
				    ms->selection_end_pos = ms->cursor_pos;
				    ms->selection_start_pos = ms->cursor_pos - 1;
				}
				MPViewSelect(
				    v->toplevel,
				    v->focus_obj_num,
				    v->focus_string_num,
				    ms->selection_start_pos,
				    ms->selection_end_pos
				);
			    }
			    else
			    {
				MPViewUnselectAll(v->toplevel);
			    }
			    MOVE_TO_CURSOR(1.0f, 1.0f);
			}
		    }

		    STOP_KEY_SIGNAL_EMIT(widget);
		}
	    }
	    break;

	  case GDK_Page_Up:
	  case GDK_KP_Page_Up:
	    if(press)
	    {
		GtkAdjustment *adj = v->vadjustment;
		if(state & GDK_CONTROL_MASK)
		{
		    GTK_ADJUSTMENT_SET_VALUE(
			adj,
			adj->value - adj->page_increment
		    );
		}
		else
		{
		    GdkFont *font = v->normal_font;
		    const gint	font_height = (font != NULL) ?
			MAX((font->ascent + font->descent), 1) : 10,
				lines_per_page = MAX(
			((gint)adj->page_size / font_height),
			1
		    );

		    /* Count the number of lines to scroll */
		    gint line_count = lines_per_page;

		    GdkEventKey e;

		    e.type = GDK_KEY_PRESS;
		    e.window = key->window;
		    e.send_event = TRUE;	/* Is send event */
		    e.time = key->time;
		    e.state = key->state;
		    e.keyval = GDK_Up;		/* Up */
		    e.length = key->length;
		    e.string = key->string;

		    while(line_count >= 0)
		    {
			(void)MPViewKeyEventCB(widget, &e, data);
			line_count--;
		    }
		}
	    }
	    break;

	  case GDK_Page_Down:
	  case GDK_KP_Page_Down:
	    if(press)
	    {
		GtkAdjustment *adj = v->vadjustment;
		if(state & GDK_CONTROL_MASK)
		{
		    GTK_ADJUSTMENT_SET_VALUE(
			adj,
			adj->value + adj->page_increment
		    );
		}
		else
		{
		    GdkFont *font = v->normal_font;
		    const gint	font_height = (font != NULL) ?
			MAX((font->ascent + font->descent), 1) : 10,
				lines_per_page = MAX(
			((gint)adj->page_size / font_height),
			1
		    );

		    /* Count the number of lines to scroll */
		    gint line_count = lines_per_page;

		    GdkEventKey e;

		    e.type = GDK_KEY_PRESS;
		    e.window = key->window;
		    e.send_event = TRUE;	/* Is send event */
		    e.time = key->time;
		    e.state = key->state;
		    e.keyval = GDK_Down;	/* Down */
		    e.length = key->length;
		    e.string = key->string;

		    while(line_count >= 0)
		    {
			(void)MPViewKeyEventCB(widget, &e, data);
			line_count--;
		    }
		}
	    }
	    break;

	  case GDK_Home:
	  case GDK_KP_Home:
	    if(press)
	    {
		MPVObject *o = MPViewGetObject(
		    v,
		    v->focus_obj_num
		);
		if(o != NULL)
		{
		    MPVString *ms = MPVObjectGetString(
			o,
			v->focus_string_num
		    );
		    if(ms != NULL)
		    {
			/* Scroll only? */
			if(state & GDK_CONTROL_MASK)
			{
			    GtkAdjustment *adj = v->vadjustment;
			    GTK_ADJUSTMENT_SET_VALUE(
				adj,
				adj->lower
			    );
			}
			/* Switch to the previous object? */
			else if(ms->cursor_pos == 0)
			{
			    if(v->focus_string_num > 0)
			    {
				v->focus_string_num--;
				MPViewSetCursorPosition(
				    v->toplevel,
				    v->focus_obj_num,
				    v->focus_string_num,
				    -1
				);
				MOVE_TO_CURSOR(1.0f, 0.0f);
			    }
			    else if(v->focus_obj_num > 0)
			    {
				v->focus_obj_num--;
				o = MPViewGetObject(
				    v,
				    v->focus_obj_num
				);
				if(o != NULL)
				    v->focus_string_num = g_list_length(o->strings_list) - 1;
				MPViewSetCursorPosition(
				    v->toplevel,
				    v->focus_obj_num,
				    v->focus_string_num,
				    -1
				);
				MOVE_TO_CURSOR(1.0f, 0.0f);
			    }

			    /* Clear the selection */
			    if(!(state & GDK_SHIFT_MASK))
				MPViewUnselectAll(v->toplevel);
			}
			else
			{
			    /* Move the cursor home */
			    const gint prev_cursor_pos = ms->cursor_pos;

			    ms->cursor_pos = 0;

			    /* Update the selection to the left? */
			    if(state & GDK_SHIFT_MASK)
			    {
				if(o->flags & MPV_OBJECT_HAS_SELECTION)
				{
				    if(ms->cursor_pos < ms->selection_start_pos)
					ms->selection_start_pos = ms->cursor_pos;
				    else if(ms->cursor_pos != ms->selection_start_pos)
					ms->selection_end_pos = ms->cursor_pos;
				}
				else
				{
				    ms->selection_start_pos = ms->cursor_pos;
				    ms->selection_end_pos = prev_cursor_pos;
				}
				MPViewSelect(
				    v->toplevel,
				    v->focus_obj_num,
				    v->focus_string_num,
				    ms->selection_start_pos,
				    ms->selection_end_pos
				);
			    }
			    else
			    {
				MPViewUnselectAll(v->toplevel);
			    }
			    MPViewQueueDraw(v->toplevel);
			    MOVE_TO_CURSOR(0.0f, 0.0f);
			}
		    }

		    STOP_KEY_SIGNAL_EMIT(widget);
		}
	    }
	    break;

	  case GDK_End:
	  case GDK_KP_End:
	    if(press)
	    {
		MPVObject *o = MPViewGetObject(
		    v,
		    v->focus_obj_num
		);
		if(o != NULL)
		{
		    MPVString *ms = MPVObjectGetString(
			o,
			v->focus_string_num
		    );
		    if(ms != NULL)
		    {
			const gint nobjs = g_list_length(v->objects_list);

			/* Scroll only? */
			if(state & GDK_CONTROL_MASK)
			{
			    GtkAdjustment *adj = v->vadjustment;
			    GTK_ADJUSTMENT_SET_VALUE(
				adj,
				adj->upper
			    );
			}
			/* Switch to the next object? */
			else if(ms->cursor_pos == ms->length)
			{
			    const gint nstrings = g_list_length(o->strings_list);
			    if(v->focus_string_num < (nstrings - 1))
			    {
				v->focus_string_num++;
				MPViewSetCursorPosition(
				    v->toplevel,
				    v->focus_obj_num,
				    v->focus_string_num,
				    0
				);
				MOVE_TO_CURSOR(0.0f, 1.0f);
			    }
			    else if(v->focus_obj_num < (nobjs - 1))
			    {
				v->focus_obj_num++;
				v->focus_string_num = 0;
				MPViewSetCursorPosition(
				    v->toplevel,
				    v->focus_obj_num,
				    v->focus_string_num,
				    0
				);
				MOVE_TO_CURSOR(0.0f, 1.0f);
			    }

			    /* Clear the selection */
			    if(!(state & GDK_SHIFT_MASK))
				MPViewUnselectAll(v->toplevel);
			}
			else
			{
			    /* Move the cursor to the end */
			    const gint prev_cursor_pos = ms->cursor_pos;

			    ms->cursor_pos = ms->length;

			    /* Update the selection to the right? */
			    if(state & GDK_SHIFT_MASK)
			    {
				if(o->flags & MPV_OBJECT_HAS_SELECTION)
				{
				    if(ms->cursor_pos > ms->selection_end_pos)
					ms->selection_end_pos = ms->cursor_pos;
				    else if(ms->cursor_pos != ms->selection_end_pos)
					ms->selection_start_pos = ms->cursor_pos;
				}
				else
				{
				    ms->selection_start_pos = prev_cursor_pos;
				    ms->selection_end_pos = ms->cursor_pos;
				}
				MPViewSelect(
				    v->toplevel,
				    v->focus_obj_num,
				    v->focus_string_num,
				    ms->selection_start_pos,
				    ms->selection_end_pos
				);
			    }
			    else
			    {
				MPViewUnselectAll(v->toplevel);
			    }
			    MPViewQueueDraw(v->toplevel);
			    MOVE_TO_CURSOR(1.0f, 1.0f);
			}
		    }

		    STOP_KEY_SIGNAL_EMIT(widget);
		}
	    }
	    break;

	  case GDK_BackSpace:
	    if(press && !(v->flags & MP_VIEW_READ_ONLY))
	    {
		MPVObject *o = MPViewGetObject(
		    v,
		    v->focus_obj_num
		);
		if(o != NULL)
		{
		    MPVString *ms = MPVObjectGetString(
			o,
			v->focus_string_num
		    );
		    if(ms != NULL)
		    {
			if(CAStringIsEmpty(ms->ca_string))
			{
			    /* Check if all the strings on the
			     * MPVObject are empty
			     */
			    GList *glist;
			    for(glist = o->strings_list;
				glist != NULL;
				glist = g_list_next(glist)
			    )
			    {
				if(!MPVStringIsEmpty(MPV_STRING(glist->data)))
				    break;
			    }
			    if(glist == NULL)
			    {
				/* Delete the MPVObject */
				v->objects_list = g_list_remove(
				    v->objects_list,
				    o
				);
				MPVObjectDelete(o);
				v->focus_obj_num--;
				REPORT_CHANGE
			    }
			}
			else if(o->flags & MPV_OBJECT_HAS_SELECTION)
			{
			    /* Delete the selected characters */
			    MPViewRemoveSelectedCharacters(v->toplevel);
			    REPORT_CHANGE
			}
			else if(ms->cursor_pos > -1)
			{
			    /* Delete the previous character */
			    if(ms->cursor_pos > 0)
			    {
				ms->cursor_pos--;
				MPVStringRemoveCharacter(
				    ms,
				    ms->cursor_pos
				);
				REPORT_CHANGE
			    }
			}
			MPViewQueueResize(v->toplevel);
			MPViewQueueDraw(v->toplevel);
			MOVE_TO_CURSOR(0.0f, 0.0f);
		    }
		}
	    }
	    break;

	  case GDK_Delete:
	    if(press && !(v->flags & MP_VIEW_READ_ONLY))
	    {
		MPViewDelete(v->toplevel);
	    }
	    break;
	}

#undef STOP_KEY_SIGNAL_EMIT
#undef MOVE_TO_FOCUS_OBJECT
#undef MOVE_TO_CURSOR
#undef REPORT_CHANGE

	return(TRUE);
}

/*
 *	View GtkDrawingArea "button_press_event" or
 *	"button_release_event" signal callback.
 */
gint MPViewButtonEventCB(
	GtkWidget *widget, GdkEventButton *button, gpointer data
)
{
	gint 		obj_num,
			string_num,
			cursor_pos;
	MPView *v = MP_VIEW(data);
	if((widget == NULL) || (button == NULL) || (v == NULL))
	    return(FALSE);

	if(v->freeze_count > 0)
	    return(FALSE);

#define MOVE_TO_CURSOR(_xc_,_yc_)	{	\
 /* Must updates the cursor's coordiantes */	\
 MPViewUpdateCursorCoordinates(v->toplevel);	\
 /* Scroll to the cursor if it is not fully	\
  * visible, check each dimension individually	\
  */						\
 if(MPViewGetCursorVisibilityX(v->toplevel) !=	\
     GDK_VISIBILITY_UNOBSCURED			\
 )						\
  MPViewMoveToCursorX(				\
   v->toplevel,					\
   (_xc_)					\
  );						\
						\
 MPViewUpdateCursorCoordinates(v->toplevel);	\
 if(MPViewGetCursorVisibilityY(v->toplevel) !=	\
     GDK_VISIBILITY_UNOBSCURED			\
 )						\
  MPViewMoveToCursorY(				\
   v->toplevel,					\
   (_yc_)					\
  );						\
}

	switch((gint)button->type)
	{
	  case GDK_BUTTON_PRESS:
	    if(!GTK_WIDGET_HAS_FOCUS(widget))
		gtk_widget_grab_focus(widget);
	    switch(button->button)
	    {
	      case GDK_BUTTON1:
		if(button->state & GDK_CONTROL_MASK)
		{

		}
		else
		{
		    MPViewUnselectAll(v->toplevel);
		}
		if(MPViewTranslateCoordinatesToIndex(
		    v->toplevel,
		    (gint)button->x, (gint)button->y,
		    &obj_num,
		    &string_num,
		    &cursor_pos
		))
		{
		    if(button->state & GDK_SHIFT_MASK)
		    {
			/* Get the previous cursor position */
			MPViewGetCursorPosition(
			    v->toplevel,
			    &v->last_drag_obj_num,
			    &v->last_drag_string_num,
			    &v->last_drag_cursor_pos,
			    NULL		/* No cursor geometry return */
			);
			MPViewSelectObjectsRange(
			    v->toplevel,
			    v->last_drag_obj_num,
			    v->last_drag_string_num,
			    v->last_drag_cursor_pos,
			    obj_num,
			    string_num,
			    cursor_pos
			);
		    }

		    /* Set the focus object, string, and cursor position */
		    MPViewSetCursorPosition(
			v->toplevel,
			obj_num,
			string_num,
			cursor_pos
		    );
		    MOVE_TO_CURSOR(0.5f, 0.5f);

		    /* Focus cell, set selection, and mark dragging */
		    if(!(v->flags & MP_VIEW_DRAG_SELECT))
		    {
			v->flags |= MP_VIEW_DRAG_SELECT;
			v->last_motion_x = (gint)button->x;
			v->last_motion_y = (gint)button->y;
			v->last_drag_obj_num = obj_num;
			v->last_drag_string_num = string_num;
			v->last_drag_cursor_pos = cursor_pos;
			gdk_pointer_grab(
			    button->window,
			    FALSE,		/* Not owner events */
			    GDK_BUTTON_PRESS_MASK |
				GDK_BUTTON_RELEASE_MASK |
				GDK_POINTER_MOTION_MASK |
				GDK_POINTER_MOTION_HINT_MASK,
			    NULL,		/* Do not confine to a
						 * GdkWindow */
			    NULL,		/* No GdkCursor */
			    button->time
			);
		    }
		}
		else
		{
		    /* Clear the focus object, string, and cursor position */
		    MPViewSetCursorPosition(
			v->toplevel,
			-1,
			-1,
			-1
		    );
		}
		break;

	      case GDK_BUTTON2:
		if(v->flags & MP_VIEW_READ_ONLY)
		    break;
		MPViewUnselectAll(v->toplevel);
		if(MPViewTranslateCoordinatesToIndex(
		    v->toplevel,
		    (gint)button->x, (gint)button->y,
		    &obj_num,
		    &string_num,
		    &cursor_pos
		))
		{

		    /* Set the focus object, string, and cursor position */
		    MPViewSetCursorPosition(
			v->toplevel,
			obj_num,
			string_num,
			cursor_pos
		    );
		    MOVE_TO_CURSOR(0.5f, 0.5f);

		    /* Paste */
		    MPViewPaste(v->toplevel);
		}
		break;

	      case GDK_BUTTON4:
		if(v->vadjustment != NULL)
		{
		    GtkAdjustment *adj = v->vadjustment;
		    const gfloat inc = MAX(
			(0.25f * adj->page_size),
			adj->step_increment
		    );
		    GTK_ADJUSTMENT_SET_VALUE(
			adj,
			adj->value - inc
		    );
		}
		break;

	      case GDK_BUTTON5:
		if(v->vadjustment != NULL)
		{
		    GtkAdjustment *adj = v->vadjustment;
		    const gfloat inc = MAX(
			(0.25f * adj->page_size),
			adj->step_increment
		    );
		    GTK_ADJUSTMENT_SET_VALUE(
			adj,
			adj->value + inc
		    );
		}
		break;
	    }
	    break;

	  case GDK_BUTTON_RELEASE:
	    switch(button->button)
	    {
	      case GDK_BUTTON1:
		if(v->flags & MP_VIEW_DRAG_SELECT)
		{
		    v->flags &= ~MP_VIEW_DRAG_SELECT;
		    v->last_motion_x = 0;
		    v->last_motion_y = 0;
		    v->last_drag_obj_num = -1;
		    v->last_drag_string_num = -1;
		    v->last_drag_cursor_pos = -1;
		    gdk_pointer_ungrab(button->time);
		}
		/* Remove the scroll timeout callback if it was active */
		if(v->scroll_toid != 0)
		{
		    gtk_timeout_remove(v->scroll_toid);
		    v->scroll_toid = 0;
		}
		/* Copy the selected regions to the selection */
		MPViewCopyString(v->toplevel);
		break;
	    }
	    break;
	}

#undef MOVE_TO_CURSOR

	return(TRUE);
}

/*
 *	Called by MPViewButtonEventCB(), MPViewMotionEventCB(), and
 *	others to update the selection.
 *
 *	The MPView's selection, cursor position, and last drag values
 *	will be updated.
 *
 *	The x and y specifies the current pointer coordinates
 *	relative to the GtkDrawingArea.
 */
void MPViewMotionEventCBUpdateSelectionIterate(
	MPView *v,
	const gint x, const gint y,
	const GdkModifierType state
)
{
	gint	obj_num,
		string_num,
		cursor_pos;

	/* Get the new MPVObject, MPVString and cursor positiong
	 * that the pointer is currently over
	 */
	if(MPViewTranslateCoordinatesToIndex(
	    v->toplevel,
	    x, y,
	    &obj_num,
	    &string_num,
	    &cursor_pos
	))
	{
	    if(state & GDK_CONTROL_MASK)
	    {

	    }
	    else
	    {
		MPViewUnselectAll(v->toplevel);
	    }

	    /* Select all between the previous cursor position and the
	     * new cursor position
	     */
	    MPViewSelectObjectsRange(
		v->toplevel,
		v->last_drag_obj_num,
		v->last_drag_string_num,
		v->last_drag_cursor_pos,
		obj_num,
		string_num,
		cursor_pos
	    );

	    /* Update the cursor position */
	    MPViewSetCursorPosition(
		v->toplevel,
		obj_num,
		string_num,
		cursor_pos
	    );
	}
	/* Not over any object so get the cursor relative to the
	 * original MPVObject and MPVString
	 */
	else if(MPViewTranslateObjectCoordinatesToIndex(
	    v->toplevel,
	    v->last_drag_obj_num,
	    v->last_drag_string_num,
	    x, y,
	    &cursor_pos
	))
	{
	    if(state & GDK_CONTROL_MASK)
	    {

	    }
	    else
	    {
		MPViewUnselectAll(v->toplevel);
	    }

	    if(cursor_pos >= v->last_drag_cursor_pos)
		MPViewSelect(
		    v->toplevel,
		    v->last_drag_obj_num,
		    v->last_drag_string_num,
		    v->last_drag_cursor_pos,
		    cursor_pos
		);
	    else
		MPViewSelect(
		    v->toplevel,
		    v->last_drag_obj_num,
		    v->last_drag_string_num,
		    cursor_pos,
		    v->last_drag_cursor_pos
		);

	    /* Update the cursor position */
	    MPViewSetCursorPosition(
		v->toplevel,
		v->last_drag_obj_num,
		v->last_drag_string_num,
		cursor_pos
	    );
	}
}

/*
 *	View GtkDrawingArea "motion_notify_event" callback.
 */
gint MPViewMotionEventCB(
	GtkWidget *widget, GdkEventMotion *motion, gpointer data
)
{
	GtkAdjustment	*hadj,
			*vadj;
	MPView *v = MP_VIEW(data);
	if((widget == NULL) || (motion == NULL) || (v == NULL))
	    return(FALSE);

	hadj = v->hadjustment;
	vadj = v->vadjustment;

	/* Request for further motion events if this is a hint and
	 * the pointer button 1 is pressed
	 */
	if(motion->is_hint &&
	   ((v->flags & MP_VIEW_DRAG_SELECT) ||
	    (v->flags & MP_VIEW_DRAG_SCROLL)
	   )
	)
	{
	    gint x, y;
	    GdkModifierType mask;
	    gdk_window_get_pointer(
		motion->window, &x, &y, &mask
	    );
	}

	/* Dragging select? */
	if(v->flags & MP_VIEW_DRAG_SELECT)
	{
	    const gint	x = (gint)motion->x,
			y = (gint)motion->y;

	    /* Update the selection */
	    MPViewMotionEventCBUpdateSelectionIterate(
		v,
		x, y,
		motion->state
	    );

	    /* Check if the motion has moved outside of the
	     * GtkDrawingArea in which case we should start or stop
	     * scrolling
	     */
	    if((x < 0) || (x >= (gint)hadj->page_size) ||
	       (y < 0) || (y >= (gint)vadj->page_size)
	    )
	    {
		if(v->scroll_toid == 0)
		    v->scroll_toid = gtk_timeout_add(
			180l,
			MPViewScrollTOCB, v
		    );
	    }
	    /* Inside the GtkDrawingArea */
	    else
	    {
		/* Remove the scroll timeout callback if it was active */
		if(v->scroll_toid != 0)
		{
		    gtk_timeout_remove(v->scroll_toid);
		    v->scroll_toid = 0;
		}
	    }
	}
	/* Dragging scroll? */
	else if(v->flags & MP_VIEW_DRAG_SCROLL)
	{
	    const gint	dx = (gint)motion->x - v->last_motion_x,
			dy = (gint)motion->y - v->last_motion_y;
	    if(dx != 0)
		GTK_ADJUSTMENT_SET_VALUE(
		    hadj,
		    hadj->value - (gfloat)dx
		);
	    if(dy != 0)
		GTK_ADJUSTMENT_SET_VALUE(
		    vadj,
		    vadj->value - (gfloat)dy
		);
	}

	/* Update the last motion values */
	v->last_motion_x = (gint)motion->x;
	v->last_motion_y = (gint)motion->y;

	return(TRUE);
}

/*
 *	MPView GtkAdjustment "value_changed" signal callback.
 */
void MPViewAdjustmentValueChangedCB(
	GtkAdjustment *adj, gpointer data
)
{
	MPView *v = MP_VIEW(data);
	if((adj == NULL) || (v == NULL))
	    return;

	if(v->freeze_count > 0)
	    return;

	MPViewQueueDraw(v->toplevel);
}

/*
 *	Resize idle callback.
 */
gint MPViewResizeIdleCB(gpointer data)
{
	MPView *v = MP_VIEW(data);
	if(v == NULL)
	    return(FALSE);

	v->resize_idle_id = 0;

	MPViewResize(v->toplevel, -1, -1);

	return(FALSE);
}

/*
 *	Line Bar scroll timeout callback.
 */
gint MPViewLineBarScrollTOCB(gpointer data)
{
	gint x, y;
	GdkModifierType mask;
	GtkAdjustment *vadj;
	GtkWidget *w;
	MPView *v = MP_VIEW(data);
	if(v == NULL)
	    return(FALSE);

	w = v->line_bar_da;
	vadj = v->vadjustment;

	gdk_window_get_pointer(w->window, &x, &y, &mask);

	/* Update the selection */
	if(v->flags & MP_VIEW_DRAG_SELECT)
	{
	    /* Update the selection */
	    MPViewLineBarMotionEventCBUpdateSelectionIterate(
		v,
		x, y,
		mask
	    );
	}

	/* Scroll if the coordinate is out of the GtkDrawingArea */
	if(y < 0)
	    GTK_ADJUSTMENT_SET_VALUE(
		vadj,
		vadj->value + (gfloat)y
	    );
	else if(y >= (gint)vadj->page_size)
	    GTK_ADJUSTMENT_SET_VALUE(
		vadj,
		vadj->value + ((gfloat)y - vadj->page_size)
	    );

	return(TRUE);
}

/*
 *	Scroll timeout callback.
 */
gint MPViewScrollTOCB(gpointer data)
{
	gint x, y;
	GdkModifierType mask;
	GtkAdjustment	*hadj,
			*vadj;
	GtkWidget *w;
	MPView *v = MP_VIEW(data);
	if(v == NULL)
	    return(FALSE);

	w = v->da;
	hadj = v->hadjustment;
	vadj = v->vadjustment;

	gdk_window_get_pointer(w->window, &x, &y, &mask);

	/* Update the selection */
	if(v->flags & MP_VIEW_DRAG_SELECT)
	{
	    /* Update the selection */
	    MPViewMotionEventCBUpdateSelectionIterate(
		v,
		x, y,
		mask
	    );
	}

	/* Scroll if the coordinate is out of the GtkDrawingArea */
	if(x < 0)
	    GTK_ADJUSTMENT_SET_VALUE(
		hadj,
		hadj->value + (gfloat)x
	    );
	else if(x >= (gint)hadj->page_size)
	    GTK_ADJUSTMENT_SET_VALUE(
		hadj,
		hadj->value + ((gfloat)x - hadj->page_size)
	    );

	if(y < 0)
	    GTK_ADJUSTMENT_SET_VALUE(
		vadj,
		vadj->value + (gfloat)y
	    );
	else if(y >= (gint)vadj->page_size)
	    GTK_ADJUSTMENT_SET_VALUE(
		vadj,
		vadj->value + ((gfloat)y - vadj->page_size)
	    );

	return(TRUE);
}
