#ifndef MANPAGE_VIEW_PRIVATE_H
#define MANPAGE_VIEW_PRIVATE_H

#include <gtk/gtk.h>
#include "ca_string.h"
#include "manpage_view.h"


/*
 *	MPVString:
 *
 *	A 2 byte per character string, the first byte is the
 *	character value and the second byte is the attribute.
 */
struct _MPVString {

        CAString	*ca_string;

        gint		length,

                        cursor_pos,

                        selection_start_pos,
                        selection_end_pos;

        MPGeometry	geometry;
};


/*
 *	MPVObject Types:
 */
typedef enum {
        MPV_OBJECT_LEFT_PARAGRAPH,
        MPV_OBJECT_ITEM,
        MPV_OBJECT_HEADING
} MPVObjectType;

/*
 *	MPVObject Flags:
 */
typedef enum {
        MPV_OBJECT_HAS_SELECTION	= (1 << 0)
} MPVObjectFlags;

/*
 *	MPVObject:
 */
struct _MPVObject {

        MPVObjectType	type;
        MPVObjectFlags	flags;
        MPGeometry	geometry;

        GList		*strings_list;		/* GList of MPVString * strings */

        gint            focus_string_num;
};


/*
 *	MPView Flags:
 */
typedef enum {
        MP_VIEW_REALIZED		= (1 << 1),
        MP_VIEW_READ_ONLY		= (1 << 2),
	MP_VIEW_SHOW_OBJECT_BOUNDS	= (1 << 3),
	MP_VIEW_SHOW_STRING_MARKERS	= (1 << 4),
        MP_VIEW_DRAG_SELECT		= (1 << 8),
        MP_VIEW_DRAG_SCROLL		= (1 << 9)
} MPViewFlags;


/*
 *	MPView:
 */
struct _MPView {

        GtkWidget	*toplevel;
        gint		freeze_count;
        GdkColormap	*colormap;
        GdkGC		*gc;
        MPViewFlags	flags;
        guint		resize_idle_id,
                        scroll_toid;

	GdkAtom		ca_string_target_type,
			mpv_object_target_type;

        GdkFont		*normal_font,
                        *bold_font,
                        *underline_font,
                        *heading_font;

        GdkColor	*normal_color,
                        *bold_color,
                        *underline_color,
                        *heading_color,
                        *bg_color;

	GdkBitmap	*line_break_bitmap,
			*string_end_bitmap;

        GdkCursor	*text_cursor,
                        *line_bar_cursor;

        gint		x_margin,		/* Edge spacing */
                        y_margin,

                        x_spacing,		/* Spacing between MPVObjects */
                        y_spacing,

			object_x_margin,	/* MPVObject margin */
			object_y_margin;

        GtkAdjustment	*vadjustment,
                        *hadjustment;

        GtkWidget	*line_bar_da,		/* GtkDrawingArea */
                        *da,			/* GtkDrawingArea */
                        *vscroll_bar,		/* Vertical GtkScrollBar */
                        *hscroll_bar;		/* Horizontal GtkScrollBar */

        GdkPixmap       *line_bar_pixmap,
                        *pixmap;

        GList		*objects_list;		/* GList * of MPVObjects *
                                                 * objects */
        gint		focus_obj_num,
                        focus_string_num;

        /* Last cursor geometry set in MPViewDraw() and
         * MPViewUpdateCursorCoordinates() relative to the
         * GtkDrawingArea with scroll adjustments applied
         */
        MPGeometry	last_cursor;

        /* Drag values used when MP_VIEW_DRAG_SELECT is set */
        gint		last_motion_x,
                        last_motion_y,
                        last_drag_obj_num,
                        last_drag_string_num,
                        last_drag_cursor_pos;

        void		(*changed_cb)(
                MPView *,
                gpointer
        );
        gpointer	changed_data;
};


/* MPVObjects (manpage_view.c) */
extern MPVObject *MPVObjectNew(void);
extern void MPVObjectDelete(MPVObject *o);

extern MPVString *MPVObjectGetString(
        MPVObject *o,
        const gint string_num
);
extern void MPVObjectRemoveSelectedCharacters(MPVObject *o);


/* MPView Get MPVObject (manpage_view.c) */
extern MPVObject *MPViewGetObject(
        MPView *v,
        const gint obj_num
);


/* MPView Callbacks (manpage_view_cb.c) */
extern void MPViewDestroyCB(gpointer data);
extern void MPViewLineBarRealizeCB(GtkWidget *widget, gpointer data);
extern gint MPViewLineBarConfigureEventCB(
        GtkWidget *widget, GdkEventConfigure *configure, gpointer data
);
extern gint MPViewLineBarExposeEventCB(
        GtkWidget *widget, GdkEventExpose *expose, gpointer data
);
extern gint MPViewLineBarFocusEventCB(
        GtkWidget *widget, GdkEventFocus *focus, gpointer data
);
extern gint MPViewLineBarKeyEventCB(
        GtkWidget *widget, GdkEventKey *key, gpointer data
);
extern gint MPViewLineBarButtonEventCB(
        GtkWidget *widget, GdkEventButton *button, gpointer data
);
extern void MPViewLineBarMotionEventCBUpdateSelectionIterate(
        MPView *v,
        const gint x, const gint y,
        const GdkModifierType state
);
extern gint MPViewLineBarMotionEventCB(
        GtkWidget *widget, GdkEventMotion *motion, gpointer data
);
extern void MPViewRealizeCB(GtkWidget *widget, gpointer data);
extern gint MPViewConfigureEventCB(
        GtkWidget *widget, GdkEventConfigure *configure, gpointer data
);
extern gint MPViewExposeEventCB(
        GtkWidget *widget, GdkEventExpose *expose, gpointer data
);
extern gint MPViewFocusEventCB(
        GtkWidget *widget, GdkEventFocus *focus, gpointer data
);
extern gint MPViewKeyEventCB(
        GtkWidget *widget, GdkEventKey *key, gpointer data
);
extern gint MPViewButtonEventCB(
        GtkWidget *widget, GdkEventButton *button, gpointer data
);
extern void MPViewMotionEventCBUpdateSelectionIterate(
        MPView *v,
        const gint x, const gint y,
        const GdkModifierType state
);
extern gint MPViewMotionEventCB(
        GtkWidget *widget, GdkEventMotion *motion, gpointer data
);
extern void MPViewAdjustmentValueChangedCB(
        GtkAdjustment *adj, gpointer data
);
extern gint MPViewResizeIdleCB(gpointer data);
extern gint MPViewLineBarScrollTOCB(gpointer data);
extern gint MPViewScrollTOCB(gpointer data);


/* MPView */
extern MPView *MPViewGetWidgetData(
        GtkWidget *w,
        const gchar *func_name
);


#endif	/* MANPAGE_VIEW_PRIVATE_H */
