#include <stdlib.h>
#include <sys/stat.h>
#include <gtk/gtk.h>

#include "../include/disk.h"

#include "cfg.h"

#include "guiutils.h"
#include "fprompt.h"
#include "cdialog.h"
#include "csd.h"
#include "fsd.h"
#include "fb.h"
#include "pdialog.h"

#include "me_types.h"
#include "editor.h"
#include "editor_fio.h"
#include "viewer.h"
#include "viewer_fio.h"
#include "about_dlg.h"
#include "manedit.h"
#include "me_cb.h"
#include "me_emit.h"
#include "me_help.h"

#include "me_cfg_list.h"
#include "config.h"


gint METimeoutCB(gpointer data);

gint MEReconfiguredIdleCB(gpointer data);

void MEHelpContentsCB(GtkWidget *widget, gpointer data);
void MEHelpTutorialCB(GtkWidget *widget, gpointer data);
void MEHelpXMLReferenceCB(GtkWidget *widget, gpointer data);

void MEAboutCB(GtkWidget *widget, gpointer data);
void MENewEditorCB(GtkWidget *widget, gpointer data);
void MENewViewerCB(GtkWidget *widget, gpointer data);

void MEOptionsWinApplyCB(
        GtkWidget *cfg_win, CfgList *cfg_list, gpointer data
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)


/*
 *	Main timeout callback.
 */
gint METimeoutCB(gpointer data)
{
	gboolean need_break, still_processing;
	gint i, windows_mapped;
	GtkCTreeNode *branch;
	GtkCTreeRow *branch_row;
	editor_struct *editor;
	viewer_struct *viewer;
	editor_item_struct *item;
	me_core_struct *core = ME_CORE(data);
	if(core == NULL)
	    return(FALSE);

	/* Need to close all windows? */
	if(core->need_close_all_windows) 
	{
	    /* Check if any dialogs are in query, break query as needed
	     * and return. Let next call to this function handle rest of
	     * closing all windows
	     */
	    if(CDialogIsQuery())
	    {
		CDialogBreakQuery();
		return(TRUE);
	    }
	    if(FileBrowserIsQuery())
	    {
		FileBrowserBreakQuery();
		return(TRUE);
	    }
	    if(CSDIsQuery())
	    {
		CSDBreakQuery();
		return(TRUE);
	    }
	    if(FSDIsQuery())
	    {
		FSDBreakQuery();
		return(TRUE);
	    }
	    if(PDialogIsQuery())
	    {
		PDialogBreakQuery();
		return(TRUE);
	    }
	    /* All dialogs not in query */


	    /* Reset need_close_all_windows back to FALSE and then check
	     * a few things.
	     */
	    core->need_close_all_windows = FALSE;

	    /* Check if any windows are processing */
	    still_processing = FALSE;

	    /* Check each editor to see if any are still processing */
	    for(i = 0; i < core->total_editors; i++)
	    {
		editor = core->editor[i];
		if(editor == NULL)
		    continue;
		
		if(editor->freeze_count > 0)
		{
		    still_processing = TRUE;
		    break;
		}
	    }
	    /* Check each viewer to see if any are still processing */
	    for(i = 0; i < core->total_viewers; i++)
	    {
		viewer = core->viewer[i];
		if(viewer == NULL)
		    continue;

		if(viewer->processing)
		{
		    still_processing = TRUE;
		    break;
		}
	    }
	    /* One or more windows still processing? */
	    if(still_processing)
	    {
		/* Return now, next call to timeout won't request to
		 * close all windows.
		 */
		return(TRUE);
	    }


	    /* Go through each editor and check for changes */
	    for(i = 0; i < core->total_editors; i++)
	    {
		editor = core->editor[i];
		if(editor == NULL)
		    continue;

		if(TRUE)
		{
		    need_break = FALSE;
		    branch = EditorGetFirstToplevelNode(editor);
		    while(branch != NULL)
		    {
			item = EditorBranchGetData(
			    GTK_CTREE(editor->layout_ctree),
			    branch
			);
			if(item->has_changes)
			{
			    need_break = TRUE;
			    break;
			}

			branch_row = GTK_CTREE_ROW(branch);
			branch = (branch_row != NULL) ?
			    branch_row->sibling : NULL;
		    }

		    if(need_break)
			break;
		}
	    }
	    if(i < core->total_editors)
	    {
		/* Found some changed data that was not saved */
		gint response;

		CDialogSetTransientFor(NULL);
		response = CDialogGetResponse(
"Save Changes",
"There are unsaved changes on one or more of the editors,\n\
are you sure you want to exit and discard those changes?",
		    NULL,
		    CDIALOG_ICON_WARNING,
		    CDIALOG_BTNFLAG_YES | CDIALOG_BTNFLAG_NO,
		    CDIALOG_BTNFLAG_NO
		);
		CDialogSetTransientFor(NULL);
		switch(response)
		{
		  case CDIALOG_RESPONSE_YES:
		  case CDIALOG_RESPONSE_YES_TO_ALL: 
		  case CDIALOG_RESPONSE_OK:
		    core->need_close_all_windows = TRUE;
		    break;
		}
	    }
	    else
	    {
		/* No changed data found, can continue to close all
		 * windows.
		 */
		core->need_close_all_windows = TRUE;
	    }

/* Go through other types of windows to check for changes? */


	    /* Confirmed to close all windows? */
	    if(core->need_close_all_windows)
	    {
		/* Unmap all the windows */
		for(i = 0; i < core->total_editors; i++)
		{
		    editor = core->editor[i];
		    if(editor == NULL)
			continue;

		    EditorUnmap(editor);
		}
		for(i = 0; i < core->total_viewers; i++)
		{
		    viewer = core->viewer[i];
		    if(viewer == NULL)
			continue;
	    
		    ViewerUnmap(viewer);
		}
	    }
	}


	/* Delete any unmapped windows */
	windows_mapped = 0;
	for(i = 0; i < core->total_editors; i++)
	{
	    editor = core->editor[i];
	    if(editor == NULL)
		continue;

	    /* Is this editor mapped? */
	    if(EditorIsMapped(editor))
	    {
		windows_mapped++;
	    }
	    else
	    {
		core->editor[i] = NULL;
		EditorDelete(editor);
	    }
	}
	for(i = 0; i < core->total_viewers; i++)
	{
	    viewer = core->viewer[i];
	    if(viewer == NULL)
		continue;

	    /* Is this viewer mapped? */
	    if(ViewerIsMapped(viewer))
	    {
		windows_mapped++;
            }
            else
            {
                core->viewer[i] = NULL;
                ViewerDelete(viewer);
            }
	}

	/* Any one windows still mapped? */
	if(windows_mapped > 0)
	{
	    /* Keep running */
	}
	else
	{
	    /* All windows have been unmapped, shut down
	     *
	     * Break out of the main GTK loop
	     */
	    gtk_main_quit();

	    /* Do not call this timeout again */
	    core->main_toid = 0;
	    return(FALSE);
	}

	return(TRUE);
}


/*
 *	Reconfigured idle callback.
 */
gint MEReconfiguredIdleCB(gpointer data)
{
	me_core_struct *core = ME_CORE(data);
	if(core == NULL)
	    return(FALSE);

	core->reconfigured_idle_id = 0;

	MEEmitReconfigured(core);

	return(FALSE);
}


/*
 *	Help Contents callback.
 */
void MEHelpContentsCB(GtkWidget *widget, gpointer data)
{
	GtkWidget *toplevel;
	me_core_struct *core = ME_CORE(data);
	if(core == NULL)
	    return;

	toplevel = (widget != NULL) ?
	    gtk_widget_get_toplevel(widget) : NULL;

	MEHelp(
	    core,
	    NULL,
	    toplevel
	);
}

/*
 *	Help Tutorial callback.
 */
void MEHelpTutorialCB(GtkWidget *widget, gpointer data)
{
	GtkWidget *toplevel;
	me_core_struct *core = ME_CORE(data);
	if(core == NULL)
	    return;

	toplevel = (widget != NULL) ?
	    gtk_widget_get_toplevel(widget) : NULL;

	MEHelp(
	    core,
	    "Tutorial",
	    toplevel
	);
}

/*
 *	Help XML Reference callback.
 */
void MEHelpXMLReferenceCB(GtkWidget *widget, gpointer data)
{
	GtkWidget *toplevel;
	me_core_struct *core = ME_CORE(data);
	if(core == NULL)
	    return;

	toplevel = (widget != NULL) ?
	    gtk_widget_get_toplevel(widget) : NULL;

	MEHelp(
	    core,
	    "XML Reference",
	    toplevel
	);
}

/*
 *	About callback.
 */
void MEAboutCB(GtkWidget *widget, gpointer data)
{
	about_dialog_struct *d;
	me_core_struct *core = ME_CORE(data);
	if(core == NULL)
	    return;

	d = core->about_dialog;
	if(d == NULL)
	    core->about_dialog = d = AboutDialogNew(core);
	if(d == NULL)
	    return;

	AboutDialogMap(d);
}

/*
 *	New Editor callback.
 */
void MENewEditorCB(GtkWidget *widget, gpointer data)
{
	gint i;
	editor_struct *editor;
	me_core_struct *core = ME_CORE(data);
	if(core == NULL)
	    return;

	for(i = 0; i < core->total_editors; i++)
	{
	    if(core->editor[i] == NULL)
		break;
	}
	if(i < core->total_editors)
	{
	    core->editor[i] = editor = EditorNew(core);
	}
	else
	{
	    i = core->total_editors;
	    core->total_editors = i + 1;
	    core->editor = (editor_struct **)g_realloc(
		core->editor,
		core->total_editors * sizeof(editor_struct *)
	    );
	    if(core->editor == NULL)
	    {
		core->total_editors = 0;
		return;
	    }
	    core->editor[i] = editor = EditorNew(core);
	}
	if(editor == NULL)
	    return;

	EditorMap(editor);
}

/*
 *	New Viewer callback.
 */
void MENewViewerCB(GtkWidget *widget, gpointer data)
{
	gint i;
	CfgList *cfg_list;
	viewer_struct *viewer;
	me_core_struct *core = ME_CORE(data);
	if(core == NULL)
	    return;

	cfg_list = core->cfg_list;

	for(i = 0; i < core->total_viewers; i++)
	{
	    if(core->viewer[i] == NULL)
		break;
	}
	if(i < core->total_viewers)
	{
	    core->viewer[i] = viewer = ViewerNew(core, -1);
	}
	else
	{
	    i = core->total_viewers;
	    core->total_viewers = i + 1;
	    core->viewer = (viewer_struct **)g_realloc(
		core->viewer,
		core->total_viewers * sizeof(viewer_struct *)
	    );
	    if(core->viewer == NULL)
	    {
		core->total_viewers = 0;
		return;
	    }
	    core->viewer[i] = viewer = ViewerNew(core, -1);
	}
	if(viewer == NULL)
	    return;

	ViewerMap(viewer);
	ViewerSetPage(
	    viewer,
	    (viewer_page)ME_GET_I(ME_CFG_PARM_VIEWER_STARTUP_PAGE)
	);
}


/*
 *	Options Window CfgWin "apply" signal callback.
 */
void MEOptionsWinApplyCB(
        GtkWidget *cfg_win, CfgList *cfg_list, gpointer data
)
{
	me_core_struct *core = ME_CORE(data);
	if(core == NULL)
	    return;

	MEQueueEmitReconfigured(core);
}
