#include <stdlib.h>
#include <ctype.h>
#include <gtk/gtk.h>
#include <gdk/gdkx.h>

#include "../include/disk.h"
#include "../include/string.h"

#include "cfg.h"
#include "cfg_fio.h"

#include "guiutils.h"
#include "tool_bar.h"
#include "cfg_win.h"

#include "editor.h"
#include "editor_fio.h"
#include "viewer.h"
#include "viewer_fio.h"
#include "about_dlg.h"
#include "manedit.h"
#include "me_cb.h"
#include "me_op.h"

#define NEED_ME_CFG_LIST_SOURCE
#include "me_cfg_list.h"
#include "config.h"

#include "images/icon_manedit_20x20.xpm"
#include "images/icon_manedit_editor_20x20.xpm"
#include "images/icon_manedit_viewer_20x20.xpm"
#include "images/icon_manpage_heading_20x20.xpm"
#include "images/icon_manpage_section_20x20.xpm"
#include "images/icon_folder_closed_20x20.xpm"
#include "images/icon_folder_opened_20x20.xpm"
#include "images/icon_manual_closed_20x20.xpm"
#include "images/icon_manual_closed_32x32.xpm"
#include "images/icon_manual_opened_20x20.xpm"
#include "images/icon_options_20x20.xpm"

static CfgList *MENewDefaultConfigurationList(
	const gchar *home_path
);

me_core_struct *MENew(
	gint argc, gchar **argv,
	gint *status_rtn
);
void MEDelete(
	me_core_struct *core,
	const gboolean save_cfg
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)


/*
 *	Creates a new CfgList with default values.
 */
static CfgList *MENewDefaultConfigurationList(
	const gchar *home_path
) 
{
	gchar		*tmp_dir;
	CfgStringList *string_list;
	CfgColor c;
	const CfgList src_cfg_list[] = ME_CONFIGURATION_LIST;

	/* Create a new configuration list with all the parameters but
	 * no associated values
	 */
	CfgItem *cfg_list = CFGItemListCopyList(src_cfg_list);
	if(cfg_list == NULL)
	    return(NULL);

	if(home_path == NULL)
	    home_path = g_getenv(ENV_VAR_NAME_HOME);
	if(home_path == NULL)
#if defined(_WIN32)
	    home_path = "\\";
#else
	    home_path = "/";
#endif

	/* Tempory Directory */
	tmp_dir = STRDUP(g_getenv(ENV_VAR_NAME_TMPDIR));
	if(tmp_dir == NULL)
#if defined(P_tmpdir)
	    tmp_dir = STRDUP(P_tmpdir);
#elif defined(_WIN32)
	    tmp_dir = STRDUP("C:\\TEMP");
#else
	    tmp_dir = STRDUP("/tmp");
#endif

	/* Begin setting default configuration values */

	ME_SET_S(
	    ME_CFG_PARM_PATH_GLOBAL_DATA,
	    ME_PATH_GLOBAL_DATA
	);
	string_list = CFGStringListNew(NULL);
	string_list->list = g_list_append(
	    string_list->list,
	    g_strdup("/usr/kerberos/man")
	);
	string_list->list = g_list_append(
	    string_list->list,
	    g_strdup("/usr/man")
	);
	string_list->list = g_list_append(
	    string_list->list,
	    g_strdup("/usr/X11R6/man")
	);
	string_list->list = g_list_append(
	    string_list->list,
	    g_strdup("/usr/local/man")
	);
	string_list->list = g_list_append(
	    string_list->list,
	    g_strdup("/usr/share/man")
	);
	ME_SET_STRING_LIST(
	    ME_CFG_PARM_PATHS_MANPAGES,
	    string_list
	);
	CFGStringListDelete(string_list);
	ME_SET_S(
	    ME_CFG_PARM_PATH_TMP,
	    tmp_dir
	);
	ME_SET_S(
	    ME_CFG_PARM_PATH_LAST_OPEN,
	    home_path
	);
	ME_SET_S(
	    ME_CFG_PARM_PATH_LAST_SAVE_AS,
	    home_path
	);

	ME_SET_S(
	    ME_CFG_PARM_CMD_MANPAGE_INTERPRETER,
	    ME_CMD_MANPAGE_INTERPRETER
	);

	ME_SET_B(ME_CFG_PARM_STARTUP_EDITOR, TRUE);
	ME_SET_B(ME_CFG_PARM_STARTUP_VIEWER, FALSE);
	ME_SET_B(ME_CFG_PARM_RECORD_WIN_POS, TRUE);
	ME_SET_B(ME_CFG_PARM_SHOW_TOOLTIPS, TRUE);

	/* Editor */
	ME_SET_I(ME_CFG_PARM_EDITOR_X, 0);
	ME_SET_I(ME_CFG_PARM_EDITOR_Y, 0);
	ME_SET_I(ME_CFG_PARM_EDITOR_WIDTH, ME_EDITOR_DEF_WIDTH);
	ME_SET_I(ME_CFG_PARM_EDITOR_HEIGHT, ME_EDITOR_DEF_HEIGHT);
	ME_SET_I(ME_CFG_PARM_EDITOR_PANED_POSITION, ME_EDITOR_DEF_PANED_POSITION);
	ME_SET_I(ME_CFG_PARM_EDITOR_TOOL_BAR_BUTTON_DECALS, TOOL_BAR_BUTTON_DECALS_PICTURES_AND_TEXT);
	ME_SET_I(ME_CFG_PARM_EDITOR_TOOL_BAR_RELIEF, GTK_RELIEF_NONE);

	ME_SET_B(ME_CFG_PARM_EDITOR_SYNTAX_HIGHLIGHTING, FALSE);

	/* Viewer */
	ME_SET_I(ME_CFG_PARM_VIEWER_X, 0);
	ME_SET_I(ME_CFG_PARM_VIEWER_Y, 0);
	ME_SET_I(ME_CFG_PARM_VIEWER_WIDTH, ME_VIEWER_DEF_WIDTH);
	ME_SET_I(ME_CFG_PARM_VIEWER_HEIGHT, ME_VIEWER_DEF_HEIGHT);
	ME_SET_I(ME_CFG_PARM_VIEWER_TOOL_BAR_BUTTON_DECALS, TOOL_BAR_BUTTON_DECALS_PICTURES_AND_TEXT);
	ME_SET_I(ME_CFG_PARM_VIEWER_TOOL_BAR_RELIEF, GTK_RELIEF_NONE);
	ME_SET_I(ME_CFG_PARM_VIEWER_STARTUP_PAGE, VIEWER_PAGE_MANPAGE);

	/* Editor Fonts */
	ME_SET_S(
	    ME_CFG_PARM_FONT_EDIT_STD,
	    ME_DEF_EDITABLE_NORMAL_FONT_NAME
	);

	/* ManPage Fonts */
	ME_SET_S(
	    ME_CFG_PARM_FONT_MANPAGE_STD,
	    ME_DEF_MANPAGE_NORMAL_FONT_NAME
	);
	ME_SET_S(
	    ME_CFG_PARM_FONT_MANPAGE_BOLD,
	    ME_DEF_MANPAGE_BOLD_FONT_NAME
	);
	ME_SET_S(
	    ME_CFG_PARM_FONT_MANPAGE_UNDERLINE,
	    ME_DEF_MANPAGE_UNDERLINE_FONT_NAME
	);

	/* Editor Colors */
	c.a = 1.0f;
	c.r = 0.0f;
	c.g = 0.0f;
	c.b = 0.0f;
	ME_SET_COLOR(
	    ME_CFG_PARM_COLOR_EDITOR_FG,
	    &c
	);
	c.a = 1.0f;
	c.r = 1.0f;
	c.g = 1.0f;
	c.b = 1.0f;
	ME_SET_COLOR(
	    ME_CFG_PARM_COLOR_EDITOR_BG,
	    &c
	);
	c.a = 1.0f;
	c.r = 0.8f;
	c.g = 0.0f;
	c.b = 0.8f;
	ME_SET_COLOR(
	    ME_CFG_PARM_COLOR_EDITOR_DELIMINATOR,
	    &c
	);
	c.a = 1.0f;
	c.r = 0.0f;
	c.g = 0.4f;
	c.b = 0.8f;
	ME_SET_COLOR(
	    ME_CFG_PARM_COLOR_EDITOR_TAG,
	    &c
	);
	c.a = 1.0f;
	c.r = 0.8f;
	c.g = 0.4f;
	c.b = 0.0f;
	ME_SET_COLOR(
	    ME_CFG_PARM_COLOR_EDITOR_SYMBOL,
	    &c
	);

	/* ManPage Colors */
	c.a = 1.0f;
	c.r = 0.0f;
	c.g = 0.0f;
	c.b = 0.0f;
	ME_SET_COLOR(
	    ME_CFG_PARM_COLOR_MANPAGE_FG,
	    &c
	);
	c.a = 1.0f;
	c.r = 1.0f;
	c.g = 1.0f;
	c.b = 0.8f;
	ME_SET_COLOR(
	    ME_CFG_PARM_COLOR_MANPAGE_BG,
	    &c
	);
	c.a = 1.0f;
	c.r = 0.0f;
	c.g = 0.0f;
	c.b = 0.0f;
	ME_SET_COLOR(
	    ME_CFG_PARM_COLOR_MANPAGE_BOLD,
	    &c
	);
	c.a = 1.0f;
	c.r = 0.0f;
	c.g = 0.0f;
	c.b = 0.0f;
	ME_SET_COLOR(
	    ME_CFG_PARM_COLOR_MANPAGE_UNDERLINE,
	    &c
	);

	g_free(tmp_dir);

	return(cfg_list);
}


/*
 *	Creates a new ManEdit core.
 */
me_core_struct *MENew(
	gint argc, gchar **argv, 
	gint *status_rtn
)
{
	gboolean startup_viewer = FALSE;
	gint	i,
		section_num = -2,		/* Any */
		editor_num = -1,
		viewer_num = -1;
	const gchar	*s,
			*arg,
			*filename = NULL;
	GdkWindow *window = (GdkWindow *)GDK_ROOT_PARENT();
	CfgList *cfg_list;
	editor_struct *editor = NULL;
	viewer_struct *viewer = NULL;
	me_pixmaps_list_struct *pixmaps_list;
	me_core_struct *core = ME_CORE(g_malloc0(
	    sizeof(me_core_struct)
	));
	if(core == NULL)
	{
	    *status_rtn = 3;
	    return(NULL);
	}

	/* Get the path to the configuration file */
	s = g_getenv(ENV_VAR_NAME_HOME);
	if(s == NULL)
	    s = "/";
	s = PrefixPaths(s, MEDIT_RCFILE_LOCAL);
	if(s == NULL)
	    s = MEDIT_RCFILE_LOCAL;
	core->cfg_file = STRDUP(s);

	/* Create a new default CfgList */
	core->cfg_list = cfg_list = MENewDefaultConfigurationList(
	    NULL				/* Default home directory */
	);

/*
	core->untitled_count = 0;

	core->main_toid = 0;
	core->reconfigured_idle_id = 0;
 */

	/* Parse arguments */
	for(i = 1; i < argc; i++)
	{
	    arg = argv[i];
	    if(arg == NULL)
		continue;

	    /* Start up with viewer? */
	    if(!g_strcasecmp(arg, "--viewer") ||
	       !g_strcasecmp(arg, "-viewer") ||
	       !g_strcasecmp(arg, "--view") ||
	       !g_strcasecmp(arg, "-view") ||
	       !g_strcasecmp(arg, "--manview") ||
	       !g_strcasecmp(arg, "-manview")
	    )
	    {
		startup_viewer = TRUE;
	    }
	    /* Configuration file? */
	    else if(!g_strcasecmp(arg, "--config") ||
		    !g_strcasecmp(arg, "-config") ||
		    !g_strcasecmp(arg, "--rcfile") ||
		    !g_strcasecmp(arg, "-rcfile") ||
		    !g_strcasecmp(arg, "-f")
	    )
	    {
		i++;
		arg = (i < argc) ? argv[i] : NULL;
		if(arg != NULL)
		{
		    g_free(core->cfg_file);
		    core->cfg_file = STRDUP(arg);
		}
		else
		{
		    g_printerr(
			"%s: Requires argument.\n",
			argv[i - 1]
		    );
		}
	    }
	    /* Filename to load on startup? */
	    else if((*arg != '-') && (*arg != '+'))
	    {
		if(isdigit(*arg))
		    section_num = ATOI(arg);

		filename = arg;
	    }
	}


	/* Load the GdkCursors */
	core->busy_cursor = gdk_cursor_new(GDK_WATCH);
	core->text_cursor = gdk_cursor_new(GDK_XTERM);


	/* Load the pixmaps list */
	pixmaps_list = &core->pixmaps_list;
	if(window != NULL)
	{
	    guint8 **data;
	    GdkBitmap **mask;
	    GdkPixmap **pixmap;

#define DO_LOAD_PIXMAP {				\
 *pixmap = gdk_pixmap_create_from_xpm_d(		\
  window,						\
  mask,							\
  NULL,							\
  (gchar **)data					\
 );							\
}
	    pixmap = &pixmaps_list->manedit_20x20;
	    mask = &pixmaps_list->manedit_20x20_mask;
	    data = (guint8 **)icon_manedit_20x20_xpm;
	    DO_LOAD_PIXMAP

	    pixmap = &pixmaps_list->editor_20x20;
	    mask = &pixmaps_list->editor_20x20_mask;
	    data = (guint8 **)icon_manedit_editor_20x20_xpm;
	    DO_LOAD_PIXMAP

	    pixmap = &pixmaps_list->viewer_20x20;
	    mask = &pixmaps_list->viewer_20x20_mask;
	    data = (guint8 **)icon_manedit_viewer_20x20_xpm;
	    DO_LOAD_PIXMAP


	    pixmap = &pixmaps_list->folder_closed_20x20;
	    mask = &pixmaps_list->folder_closed_20x20_mask;
	    data = (guint8 **)icon_folder_closed_20x20_xpm;
	    DO_LOAD_PIXMAP

	    pixmap = &pixmaps_list->folder_opened_20x20;
	    mask = &pixmaps_list->folder_opened_20x20_mask;
	    data = (guint8 **)icon_folder_opened_20x20_xpm;
	    DO_LOAD_PIXMAP


	    pixmap = &pixmaps_list->manual_closed_20x20;
	    mask = &pixmaps_list->manual_closed_20x20_mask;
	    data = (guint8 **)icon_manual_closed_20x20_xpm;
	    DO_LOAD_PIXMAP

	    pixmap = &pixmaps_list->manual_closed_32x32;
	    mask = &pixmaps_list->manual_closed_32x32_mask;
	    data = (guint8 **)icon_manual_closed_32x32_xpm;
	    DO_LOAD_PIXMAP

	    pixmap = &pixmaps_list->manual_opened_20x20;
	    mask = &pixmaps_list->manual_opened_20x20_mask;
	    data = (guint8 **)icon_manual_opened_20x20_xpm;
	    DO_LOAD_PIXMAP


	    pixmap = &pixmaps_list->manpage_heading_20x20;
	    mask = &pixmaps_list->manpage_heading_20x20_mask;
	    data = (guint8 **)icon_manpage_heading_20x20_xpm;
	    DO_LOAD_PIXMAP

	    pixmap = &pixmaps_list->manpage_section_20x20;
	    mask = &pixmaps_list->manpage_section_20x20_mask;
	    data = (guint8 **)icon_manpage_section_20x20_xpm;
	    DO_LOAD_PIXMAP


	    pixmap = &pixmaps_list->options_20x20;
	    mask = &pixmaps_list->options_20x20_mask;
	    data = (guint8 **)icon_options_20x20_xpm;
	    DO_LOAD_PIXMAP

#undef DO_LOAD_PIXMAP
	}

	/* Open the configuration */
	if(core->cfg_file != NULL)
	    (void)CFGFileOpen(
		core->cfg_file,
		core->cfg_list
	    );


	/* Update the global tooltips state */
	GUISetGlobalTipsState(ME_GET_B(ME_CFG_PARM_SHOW_TOOLTIPS));


	/* Update the GtkRcStyles */
	MEUpdateRcStyles(core);


	/* Startup with an editor? */
	if(!startup_viewer &&
	   ME_GET_B(ME_CFG_PARM_STARTUP_EDITOR)
	)
	    editor = EditorNew(core);
	else
	    editor = NULL;
	if(editor != NULL)
	{
	    if(core->total_editors < 0)
		core->total_editors = 0;

	    editor_num = core->total_editors;
	    core->total_editors++;
	    core->editor = (editor_struct **)g_realloc(
		core->editor,
		core->total_editors * sizeof(editor_struct *)
	    );
	    if(core->editor == NULL)
	    {
		core->total_editors = 0;
		EditorDelete(editor);
		editor = NULL;
	    }
	    else
	    {
		core->editor[editor_num] = editor;
		EditorMap(editor);
		if(filename != NULL)
		    (void)EditorFileOpen(
			editor,
			filename,
			NULL,			/* Append */
			FALSE			/* Not as a template */
		    );
	    }
	}

	/* Startup with a viewer? */
	if((startup_viewer) ||
	   ME_GET_B(ME_CFG_PARM_STARTUP_VIEWER)
	)
	    viewer = ViewerNew(
		core,
		(core->total_editors > 0) ? 0 : -1
	    );
	else
	    viewer = NULL;
	if(viewer != NULL)
	{
	    if(core->total_viewers < 0)
		core->total_viewers = 0;
	      
	    viewer_num = core->total_viewers;
	    core->total_viewers++;
	    core->viewer = (viewer_struct **)g_realloc(
		core->viewer,
		core->total_viewers * sizeof(viewer_struct *)
	    );
	    if(core->viewer == NULL)
	    {
		core->total_viewers = 0;
		ViewerDelete(viewer);
		viewer = NULL;
	    }
	    else
	    {
		core->viewer[viewer_num] = viewer;
		ViewerMap(viewer);

		/* Section specified? */
		if(section_num != 0)
		    ViewerSetSection(viewer, section_num);

		/* Open a manual page on startup? */
		if(filename != NULL)
		{
		    gchar *s;

		    ViewerSetBusy(viewer, TRUE);

		    /* Add the section to the beginning of the search
		     * string */
		    switch(section_num)
		    {
		      case -1:			/* Exact */
			/* filename must be an absolute path */
			s = STRDUP(filename);
			break;

		      case -2:			/* Any */
			s = STRDUP(filename);
			break;

		      default:
			s = g_strdup_printf(
			    "%i %s",
			    section_num,
			    filename
			);
			break;
		    }
		    ViewerTextDelete(viewer, 0, -1);
		    ViewerTextInsertPosition(viewer, 0);
		    ViewerOpenFile(viewer, s, filename);
		    g_free(s);
		    ViewerSetBusy(viewer, FALSE);
		}
		else
		{
		    ViewerSetPage(
			viewer,
			(viewer_page)ME_GET_I(ME_CFG_PARM_VIEWER_STARTUP_PAGE)
		    );
		}
	    }
	}

	/* If no windows were created then create a new editor */
	if((editor == NULL) && (viewer == NULL))
	{
	    editor = EditorNew(core);
	    if(editor != NULL)
	    {
		editor_num = MAX(core->total_editors, 0);
		core->total_editors = editor_num + 1;
		core->editor = (editor_struct **)g_realloc(
		    core->editor,
		    core->total_editors * sizeof(editor_struct *)
		);
		if(core->editor == NULL)
		{
		    core->total_editors = 0;
		    EditorDelete(editor);
		    editor = NULL;
		}
		else
		{
		    core->editor[editor_num] = editor;
		    EditorMap(editor);
		    if(filename != NULL)
			(void)EditorFileOpen(
			    editor,
			    filename,
			    NULL,		/* Append */
			    FALSE		/* Not as a template */
			);
		}
	    }
	}


	/* Call crash check recovery function to prompt and recover any
	 * files saved by an emergency save in a previous run of this
	 * program.
	 */
	MECrashRecovery(core);


	/* Set the timeout callback */
	core->main_toid = gtk_timeout_add(
	    1000l,
	    METimeoutCB, core
	);


	*status_rtn = 0;

	return(core);
}

/*
 *	Deletes the ManEdit core.
 *
 *	If save_cfg is TRUE then the CfgList will be saved.
 */
void MEDelete(
	me_core_struct *core,
	const gboolean save_cfg
)
{
	gint i;
	me_pixmaps_list_struct *pixmaps_list;

	if(core == NULL)
	    return;

	/* Remove the timeout callback */
	if(core->main_toid != 0)
	{
	    gtk_timeout_remove(core->main_toid);
	    core->main_toid = 0;
	}

	/* Remove the timeout callback */
	if(core->reconfigured_idle_id != 0)
	{
	    gtk_idle_remove(core->reconfigured_idle_id);
	    core->reconfigured_idle_id = 0;
	}

	/* Delete all the editors */
	for(i = 0; i < core->total_editors; i++)
	    EditorDelete(core->editor[i]);
	g_free(core->editor);
	core->editor = NULL;
	core->total_editors = 0;

	/* Delete all the viewers */
	for(i = 0; i < core->total_viewers; i++)
	    ViewerDelete(core->viewer[i]);
	g_free(core->viewer);
	core->viewer = NULL;
	core->total_viewers = 0;

	/* Save the configuration */
	if(save_cfg && !STRISEMPTY(core->cfg_file))
	    (void)CFGFileSave(
		core->cfg_file,
		core->cfg_list
	    );

	/* Delete the About Dialog */
	AboutDialogDelete(core->about_dialog);
	core->about_dialog = NULL;

	/* Delete the Options Window */
	CfgWinUnref(core->options_window);
	core->options_window = NULL;


	/* Destroy all the GdkCursors */
	GDK_CURSOR_DESTROY(core->busy_cursor);
	GDK_CURSOR_DESTROY(core->text_cursor);


	/* Unref all pixmaps */
	pixmaps_list = &core->pixmaps_list;

	GDK_PIXMAP_UNREF(pixmaps_list->manedit_20x20);
	GDK_BITMAP_UNREF(pixmaps_list->manedit_20x20_mask);

	GDK_PIXMAP_UNREF(pixmaps_list->editor_20x20);
	GDK_BITMAP_UNREF(pixmaps_list->editor_20x20_mask);

	GDK_PIXMAP_UNREF(pixmaps_list->viewer_20x20);
	GDK_BITMAP_UNREF(pixmaps_list->viewer_20x20_mask);

	GDK_PIXMAP_UNREF(pixmaps_list->folder_closed_20x20);
	GDK_BITMAP_UNREF(pixmaps_list->folder_closed_20x20_mask);

	GDK_PIXMAP_UNREF(pixmaps_list->folder_opened_20x20);
	GDK_BITMAP_UNREF(pixmaps_list->folder_opened_20x20_mask);

	GDK_PIXMAP_UNREF(pixmaps_list->manual_closed_20x20);
	GDK_BITMAP_UNREF(pixmaps_list->manual_closed_20x20_mask);

	GDK_PIXMAP_UNREF(pixmaps_list->manual_closed_32x32);
	GDK_BITMAP_UNREF(pixmaps_list->manual_closed_32x32_mask);

	GDK_PIXMAP_UNREF(pixmaps_list->manual_opened_20x20);
	GDK_BITMAP_UNREF(pixmaps_list->manual_opened_20x20_mask);

	GDK_PIXMAP_UNREF(pixmaps_list->manpage_heading_20x20);
	GDK_BITMAP_UNREF(pixmaps_list->manpage_heading_20x20_mask);

	GDK_PIXMAP_UNREF(pixmaps_list->manpage_section_20x20);
	GDK_BITMAP_UNREF(pixmaps_list->manpage_section_20x20_mask);

	GDK_PIXMAP_UNREF(pixmaps_list->options_20x20);
	GDK_BITMAP_UNREF(pixmaps_list->options_20x20_mask);


	/* Unref all the GtkRcStyles */
	GTK_RC_STYLE_UNREF(core->editable_rcstyle);
	GTK_RC_STYLE_UNREF(core->manpage_rcstyle);

	/* Unref all the GdkFonts */
	GDK_FONT_UNREF(core->editable_font);
	GDK_FONT_UNREF(core->manpage_normal_font);
	GDK_FONT_UNREF(core->manpage_bold_font);
	GDK_FONT_UNREF(core->manpage_underline_font);

	/* Delete the CfgList */
	CFGItemListDeleteList(core->cfg_list);
	g_free(core->cfg_file);

	g_free(core);
}
