#include <gtk/gtk.h>

#include "cfg.h"

#include "guiutils.h"

#include "me_types.h"
#include "editor.h"
#include "editor_cb.h"
#include "viewer.h"
#include "viewer_cb.h"
#include "manedit.h"
#include "me_cb.h"
#include "me_op.h"
#include "me_emit.h"

#include "me_cfg_list.h"


void MEEmitReconfigured(me_core_struct *core);
void MEQueueEmitReconfigured(me_core_struct *core);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)


/*
 *	Emits the reconfigured signal.
 */
void MEEmitReconfigured(me_core_struct *core)
{
	gint i;
	CfgList *cfg_list;

	if(core == NULL)
	    return;

	cfg_list = core->cfg_list;

	/* Update the global tooltips state */
	GUISetGlobalTipsState(ME_GET_B(ME_CFG_PARM_SHOW_TOOLTIPS));

	/* Update the GtkRcStyles and GdkFonts */
	MEUpdateRcStyles(core);

	for(i = 0; i < core->total_editors; i++)
	    EditorReconfiguredCB(core->editor[i]);

	for(i = 0; i < core->total_viewers; i++)
	    ViewerReconfiguredCB(core->viewer[i]);
}

/*
 *	Queues a reconfigured signal emit.
 */
void MEQueueEmitReconfigured(me_core_struct *core)
{
	if(core == NULL)
	    return;

	if(core->reconfigured_idle_id != 0)
	    return;

	core->reconfigured_idle_id = gtk_idle_add_priority(
	    G_PRIORITY_LOW,
	    MEReconfiguredIdleCB, core
	);
}
