/*
			       ManPage Viewer
 */

#ifndef VIEWER_H
#define VIEWER_H

#include <gtk/gtk.h>
#include "me_types.h"


typedef struct _viewer_struct			viewer_struct;
#define VIEWER(p)				((viewer_struct *)(p))

typedef struct _ViewerIndexItem			ViewerIndexItem;
#define VIEWER_INDEX_ITEM(p)			((ViewerIndexItem *)(p))


/*
 *	Operation IDs.
 */
typedef enum {
	VIEWER_OPID_OPEN,
	VIEWER_OPID_CLEAR,
	VIEWER_OPID_STOP
} viewer_opid;

/*
 *	Viewer Pages:
 */
typedef enum {
	VIEWER_PAGE_MANPAGE,
	VIEWER_PAGE_INDEX
} viewer_page;


/*
 *	Viewer Index Item Types:
 */
typedef enum {
	VIEWER_INDEX_ITEM_TYPE_MANPAGE,
	VIEWER_INDEX_ITEM_TYPE_DIRECTORY,
	VIEWER_INDEX_ITEM_TYPE_OTHER_FILE
} ViewerIndexItemType;


/*
 *	Viewer Index Item:
 */
struct _ViewerIndexItem {
	ViewerIndexItemType	type;
	gchar		*path;
};


/*
 *	Viewer:
 */
struct _viewer_struct {

	GtkWidget	*toplevel;
	GtkAccelGroup	*accelgrp;
	gint		freeze_count,
			busy_count,
			stop_count;
	gboolean	processing;
	GdkColormap	*colormap;
	me_core_struct	*core;
	gint		editor_num;

	GtkWidget	*main_vbox,
			*menu_bar_dock,
			*menu_bar,
			*tool_bar_dock,
			*tool_bar,
			*main_notebook;

	viewer_page	current_page;

	GtkWidget	*view_text;		/* View GtkText */

	/* View Text GdkColors */
	GdkColor	view_normal_text_color,
			view_bold_text_color,
			view_underline_text_color;

	/* Index listing of available manual pages, each branch has
	 * an item data of type ViewerIndexItem.
	 * There are 4 columns; name, lines, read, short description
	 */
	GtkWidget	*index_ctree;

	/* Last selected GtkCTreeNode on the Index GtkCTree */
	GtkCTreeNode	*last_selected_index_node;

	/* Buttons on index ctree page */
	GtkWidget	*index_parent_btn,
			*index_view_btn,
			*index_refresh_btn,
			*index_find_in_pages_combo;

	/* Combos on manpage view panel */
	GtkWidget	*section_pulistbox,
			*manpage_combo,
			*find_combo;

	/* Main GtkMenuItems */
	GtkWidget	*manual_mh,
			*open_mi,
			*clear_mi,
			*close_mi,
			*exit_mi,

			*edit_mh,
			*edit_copy_mi,
			*edit_select_all_mi,
			*edit_unselect_all_mi,

			*view_mh,
			*view_stop_mi,
			*view_view_mi,
			*view_index_mi,

			*settings_mh,
			*settings_options_mi,

			*windows_mh,
			*windows_new_editor,
			*windows_new_viewer;

	/* View Right Click GtkMenu */
	GtkWidget	*view_menu,
			*view_copy_mi,
			*view_select_all_mi,
			*view_unselect_all_mi;

	/* Index GtkCTree Right Click GtkMenu */
	GtkWidget	*index_menu,
			*index_expand_mi,
			*index_parent_mi,
			*index_view_mi,
			*index_refresh_mi,
			*index_stop_mi;

	/* Status Bar */
	GtkWidget       *status_bar_dock,
			*status_bar_toplevel,   /* An hbox */
			*status_bar_progress,
			*status_bar_label,
			*status_bar_cursor_label;

	/* Last progress position value from 0.0 to 1.0 */
	gfloat		status_bar_progress_pos_last;

	/* Current manual page path */
	gchar		*cur_manpage_path;

	/* Name of current manpage, this is to match up to see if
	 * newly loaded manpage is the same. Note that this is not
	 * the file name, rather it should be the name from the
	 * manual page's header */
	gchar		*cur_manpage_name;

	/* Last scroll positions */
	gfloat		last_scroll_hpos,	/* Not used */
			last_scroll_vpos;

	/* Last manual page output file opened path */
	gchar		*last_open_path;

};


/* Index GtkCTree Item */
extern ViewerIndexItem *ViewerIndexItemNew(void);
extern void ViewerIndexItemDelete(ViewerIndexItem *item);

/* Section */
extern void ViewerSetSection(
	viewer_struct *v,
	const gint section_num
);
extern gint ViewerCurrentSection(viewer_struct *v);

/* View Text */
extern void ViewerTextInsertPosition(
	viewer_struct *v,
	const gint start_pos
);
extern void ViewerTextInsert(
	viewer_struct *v,
	const gchar *buf, const gint buf_len,
	gint (*progress_cb)(const gulong, const gulong, gpointer),
	gpointer progress_data
);
extern void ViewerTextDelete(
	viewer_struct *v,
	const gint start_pos, const gint end_pos
); 

/* Index GtkCTree Load */
extern void ViewerIndexLoad(
	viewer_struct *v,
	GList *paths_list
);
extern void ViewerIndexRefresh(viewer_struct *v);

/* Find */
extern gboolean ViewerDoFind(
	viewer_struct *v,
	GtkText *text,
	gchar *haystack,
	gchar *needle,
	gint haystack_len,
	gint start_pos,
	gboolean case_sensitive,
	gboolean move_to,
	gboolean *search_wrapped
);

/* Record Scroll Position */
extern void ViewerViewTextRecordScrollPositions(viewer_struct *v);

/* Viewer */
extern viewer_struct *ViewerNew(
	me_core_struct *core,
	const gint editor_num
);
extern void ViewerRecordConfiguration(viewer_struct *v);
extern void ViewerUpdateTitle(viewer_struct *v);
extern void ViewerUpdateMenus(viewer_struct *v);
extern void ViewerSetPage(
	viewer_struct *v,
	viewer_page page
);
extern void ViewerSetStatusPosition(
	viewer_struct *v,
	const gint row, const gint column
);
extern void ViewerSetStatusMessage(
	viewer_struct *v,
	const gchar *msg
);
extern void ViewerSetStatusProgress(
	viewer_struct *v,
	const gfloat val
);
extern void ViewerSetBusy(
	viewer_struct *v,
	const gboolean busy
);
extern gboolean ViewerIsMapped(viewer_struct *v);
extern void ViewerMap(viewer_struct *v);
extern void ViewerUnmap(viewer_struct *v);
extern void ViewerDelete(viewer_struct *v);


#endif	/* VIEWER_H */
