#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "../include/string.h"
#include "../include/disk.h"

#include "cfg.h"

#include "guiutils.h"
#include "tool_bar.h"
#include "pulist.h"
#include "cfg_win.h"
#include "cdialog.h"
#include "fb.h"

#include "editor.h"
#include "viewer.h"
#include "viewer_cb.h"
#include "viewer_fio.h"
#include "manedit.h"
#include "options_win.h"

#include "me_cfg_list.h"
#include "config.h"

#include "mansuffix.h"


gint ViewerDeleteEventCB(GtkWidget *widget, GdkEvent *event, gpointer data);
void ViewerCloseCB(GtkWidget *widget, gpointer data);
void ViewerExitCB(GtkWidget *widget, gpointer data);
gint ViewerKeyEventCB(
	GtkWidget *widget, GdkEventKey *key, gpointer data
);

static gint ViewerPageChangedIdleCB(gpointer data);
void ViewerSwitchPageCB(
	GtkNotebook *notebook, GtkNotebookPage *page, guint page_num,
	gpointer data
);
void ViewerPageToggleCB(GtkWidget *widget, gpointer data);

gint ViewerViewKeyEventCB(
	GtkWidget *widget, GdkEventKey *key, gpointer data
);
gint ViewerViewButtonEventCB(
	GtkWidget *widget, GdkEventButton *button, gpointer data
);

void ViewerIndexBranchExpandCB(GtkWidget *widget, gpointer data);

void ViewerToolBarItemActivateCB(
	ToolBarItem *item, gint id, gpointer data
);

void ViewerOpenCB(GtkWidget *widget, gpointer data);
void ViewerClearCB(GtkWidget *widget, gpointer data);
void ViewerStopCB(GtkWidget *widget, gpointer data);
void ViewerIndexGotoParentCB(GtkWidget *widget, gpointer data);
void ViewerIndexViewCB(GtkWidget *widget, gpointer data);
void ViewerIndexRefreshCB(GtkWidget *widget, gpointer data);
void ViewerCopyCB(GtkWidget *widget, gpointer data);
void ViewerSelectAllCB(GtkWidget *widget, gpointer data);
void ViewerUnselectAllCB(GtkWidget *widget, gpointer data);
void ViewerOptionsCB(GtkWidget *widget, gpointer data);

gint ViewerMenuMapCB(GtkWidget *widget, GdkEvent *event, gpointer data);

void ViewerManPageActivateCB(GtkWidget *widget, gpointer data);
static void ViewerFind(
	viewer_struct *v,
	GtkCombo *find_combo,
	const gboolean case_sensitive
);
void ViewerFindActivateCB(GtkWidget *widget, gpointer data);

static gboolean ViewerFindInPagesGrep(
	const gchar *path,
	const gchar *expression,
	const gboolean case_sensitive
);
static void ViewerDoFindInPagesCB(
	viewer_struct *v, GtkCombo *find_combo, gboolean case_sensitive
);
void ViewerFindInPagesActivateCB(GtkWidget *widget, gpointer data);

void ViewerIndexCTreeSelectCB(
	GtkCTree *ctree, GtkCTreeNode *branch, gint column, gpointer data
);
void ViewerIndexCTreeUnselectCB(
	GtkCTree *ctree, GtkCTreeNode *branch, gint column, gpointer data
);
void ViewerIndexCTreeExpandCB(
	GtkCTree *ctree, GtkCTreeNode *node, gpointer data
);

void ViewerReconfiguredCB(viewer_struct *v);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)

#define ISCR(c)         (((c) == '\n') || ((c) == '\r'))
#define ISBLANK(c)      (((c) == ' ') || ((c) == '\t'))


/*
 *	Toplevel GtkWindow "delete_event" signal callback.
 */
gint ViewerDeleteEventCB(GtkWidget *widget, GdkEvent *event, gpointer data)
{
	viewer_struct *v = VIEWER(data);
	if(v == NULL)
	    return(TRUE);

	if(v->freeze_count > 0)
	    return(TRUE);

	v->freeze_count++;

	/* Revord the Viewer's values to the CfgList */
	ViewerRecordConfiguration(v);

	ViewerUnmap(v);

	v->freeze_count--;

	return(TRUE);
}


/*
 *	Close callback.
 */
void ViewerCloseCB(GtkWidget *widget, gpointer data)
{
	ViewerDeleteEventCB(widget, NULL, data);
	return;
}


/*
 *	Close all windows callback.
 */
void ViewerExitCB(GtkWidget *widget, gpointer data)
{
	me_core_struct *core;
	viewer_struct *v = VIEWER(data);
	if(v == NULL)
	    return;

	if(v->freeze_count > 0)
	    return;

	core = v->core;
	core->need_close_all_windows = TRUE; 
}


/*
 *	Toplevel GtkWindow "key_press_event" or "key_release_event"
 *	signal callback.
 */
gint ViewerKeyEventCB(
	GtkWidget *widget, GdkEventKey *key, gpointer data
)
{
	gint status = FALSE;
	gboolean press;
	me_core_struct *core;
	viewer_struct *v = VIEWER(data);
	if((key == NULL) || (v == NULL))
	    return(status);

	if(v->freeze_count > 0)
	    return(status);

	press = (key->type == GDK_KEY_PRESS) ? TRUE : FALSE;
	core = v->core;

	switch(key->keyval)
	{
	  case GDK_F3:
	    if(press)
	    {
		ViewerOpenCB(widget, data);
	    }
	    status = TRUE;
	    break;
	}

	return(status);
}


/*
 *	GtkNoteBook page changed idle callback.
 */
static gint ViewerPageChangedIdleCB(gpointer data)
{
	GtkCTree *ctree;
	viewer_struct *v = VIEWER(data);
	if(v == NULL)
	    return(FALSE);

	if(v->freeze_count > 0)
	    return(FALSE);

	/* Do not freeze the viewer or else the user can not abort the
	 * operation
	 */

	switch(v->current_page)
	{
	  case VIEWER_PAGE_MANPAGE:
	    break;

	  case VIEWER_PAGE_INDEX:
	    /* If the index is empty then load it */
	    ctree = GTK_CTREE(v->index_ctree);
	    if(gtk_ctree_node_nth(ctree, 0) == NULL)
		ViewerIndexRefresh(v);
	    break;
	}

	return(FALSE);
}

/*
 *	GtkNotebook "switch_page" signal callback.
 */
void ViewerSwitchPageCB(
	GtkNotebook *notebook, GtkNotebookPage *page, guint page_num,
	gpointer data
)
{
	viewer_struct *v = VIEWER(data);
	if((notebook == NULL) || (v == NULL))
	    return;

	if(v->freeze_count > 0)
	    return;

	v->freeze_count++;

	/* Record the new page */
	v->current_page = (viewer_page)page_num;

	ViewerUpdateTitle(v);
	ViewerUpdateMenus(v);

	/* Add the timeout to check for page changed, cannot handle
	 * page changed now since this callback is called just before
	 * the page is actually changed
	 */
	gtk_idle_add_priority(
	    G_PRIORITY_LOW,
	    ViewerPageChangedIdleCB, v
	);

	v->freeze_count--;
}

/*
 *	Page toggle widgets callback.
 */
void ViewerPageToggleCB(GtkWidget *widget, gpointer data)
{
	viewer_struct *v = VIEWER(data);
	if((widget == NULL) || (v == NULL))
	    return;

	if(v->freeze_count > 0)
	    return;

	if(widget == v->view_view_mi)
	{
	    gtk_notebook_set_page(
		GTK_NOTEBOOK(v->main_notebook),
		VIEWER_PAGE_MANPAGE
	    );
	}
	else if(widget == v->view_index_mi)
	{
	    gtk_notebook_set_page(
		GTK_NOTEBOOK(v->main_notebook),
		VIEWER_PAGE_INDEX
	    );
	}
}


/*
 *	View GtkText "key_press_event" or "key_release_event" signal
 *	callback.
 */
gint ViewerViewKeyEventCB(
	GtkWidget *widget, GdkEventKey *key, gpointer data   
)
{
	gint status = FALSE;
	gboolean press;
	GtkAdjustment	*hadj,
			*vadj;
	GtkEditable *editable;
	GtkText *text;
	viewer_struct *v = VIEWER(data);
	if((widget == NULL) || (key == NULL) || (v == NULL))
	    return(status);

	if(v->freeze_count > 0)
	    return(status);

	press = (key->type == GDK_KEY_PRESS) ? TRUE : FALSE;
	editable = GTK_EDITABLE(widget);
	text = GTK_TEXT(widget);
	hadj = text->hadj;
	vadj = text->vadj;

	/* Important, for some obscure reason the GtkText widget
	 * when set not editable and recieving a key event to insert
	 * text will cause it to internally call gtk_main_quit() and
	 * thus exit or mess up the application.
	 *
	 * Here we play a trick, when the key is pressed the GtkText
	 * widget is set editable but when the key release is recieved
	 * the GtkText is set back not editable.
	 */
	if(press)
	    gtk_editable_set_editable(editable, TRUE);
	else
	    gtk_editable_set_editable(editable, FALSE);

	switch(key->keyval)
	{
	  case GDK_Home:
	  case GDK_KP_Home:
	    if(press)
	    {
		GTK_ADJUSTMENT_SET_VALUE(
		    vadj,
		    vadj->lower
		);
	    }
	    status = TRUE;
	    break;

	  case GDK_End:
	  case GDK_KP_End:
	    if(press)
	    {
		GTK_ADJUSTMENT_SET_VALUE(
		    vadj,
		    vadj->upper
		);
	    }
	    status = TRUE;
	    break;

	  case GDK_Return:
	  case GDK_space:
	  case GDK_n:
	    if(press)
	    {
		GTK_ADJUSTMENT_SET_VALUE(
		    vadj,
		    vadj->value + vadj->page_size
		);
	    }
	    status = TRUE;
	    break;

	  case GDK_BackSpace:
	  case GDK_b:
	  case GDK_p:
	    if(press)
	    {
		GTK_ADJUSTMENT_SET_VALUE(
		    vadj,
		    vadj->value - vadj->page_size
		);
	    }
	    status = TRUE;
	    break;
	}

	return(status);
}

/*
 *	View GtkText "button_press_event" or "button_release_event"
 *	signal callback.
 */
gint ViewerViewButtonEventCB(
	GtkWidget *widget, GdkEventButton *button, gpointer data
)
{
	gint status = FALSE;
	GtkText *text;
	viewer_struct *v = VIEWER(data);
	if((widget == NULL) || (button == NULL) || (v == NULL))
	    return(status);

	if(!GTK_IS_TEXT(widget))
	    return(status);

	text = GTK_TEXT(widget);

	switch((gint)button->type)
	{
	  case GDK_BUTTON_PRESS:
	    switch(button->button)
	    {
	      case GDK_BUTTON3:		/* Popup menu */
		/* Map the popup menu */
		gtk_menu_popup(
		    GTK_MENU(v->view_menu),
		    NULL, NULL, NULL, v,
		    button->button, button->time
		);
		/* Need to mark the text widget's button as 0 or
		 * else it will keep marking.
		 */
		text->button = 0;
		gtk_grab_remove(widget);
		gtk_signal_emit_stop_by_name(
		    GTK_OBJECT(widget), "button_press_event"
		);
		status = TRUE;
		break;

	      case GDK_BUTTON4:		/* Scroll up */
		if(text->vadj != NULL)
		{
		    GtkAdjustment *adj = text->vadj;
		    const gfloat inc = MAX(
			(0.25f * adj->page_size),
			adj->step_increment
		    );
		    gfloat v = adj->value - inc;  
		    if(v > (adj->upper - adj->page_size))
			v = adj->upper - adj->page_size;
		    if(v < adj->lower)
			v = adj->lower;
		    gtk_adjustment_set_value(adj, v);
		}
		/* Need to mark the GtkText button as 0 or else it will
		 * keep marking
		 */
		text->button = 0;
		gtk_grab_remove(widget);
		gtk_signal_emit_stop_by_name(
		    GTK_OBJECT(widget), "button_press_event"
		);
		status = TRUE;
		break;

	      case GDK_BUTTON5:		/* Scroll down */
		if(text->vadj != NULL)
		{
		    GtkAdjustment *adj = text->vadj;
		    const gfloat inc = MAX(
			(0.25f * adj->page_size),
			adj->step_increment
		    );
		    gfloat v = adj->value + inc;
		    if(v > (adj->upper - adj->page_size))
			v = adj->upper - adj->page_size;
		    if(v < adj->lower)
			v = adj->lower;
		    gtk_adjustment_set_value(adj, v);
		}
		/* Need to mark the GtkText button as 0 or else it will
		 * keep marking
		 */
		text->button = 0;
		gtk_grab_remove(widget);
		gtk_signal_emit_stop_by_name(
		    GTK_OBJECT(widget), "button_press_event"
		);
		status = TRUE;
		break;
	    }
	    break;

	  case GDK_BUTTON_RELEASE:
	    switch(button->button)
	    {
	      case GDK_BUTTON3:		/* Popup menu */
		/* Need to mark the GtkText button as 0 or else it will
		 * keep marking
		 */
		text->button = 0;
		gtk_grab_remove(widget);
		gtk_signal_emit_stop_by_name(
		    GTK_OBJECT(widget), "button_release_event"
		);
		status = TRUE;
		break;

	      case GDK_BUTTON4:		/* Scroll up */
		/* Need to mark the GtkText button as 0 or else it will
		 * keep marking
		 */
		text->button = 0;
		gtk_grab_remove(widget);
		gtk_signal_emit_stop_by_name(
		    GTK_OBJECT(widget), "button_release_event"
		);
		status = TRUE;
		break;

	      case GDK_BUTTON5:		/* Scroll down */
		/* Need to mark the GtkText button as 0 or else it will
		 * keep marking
		 */
		text->button = 0;
		gtk_grab_remove(widget);
		gtk_signal_emit_stop_by_name(
		    GTK_OBJECT(widget), "button_release_event"
		);
		status = TRUE;
		break;
	    }
	    break;
	}

	return(status);
}


/*
 *	Branch expand/collapse callback (not to handle a signal when
 *	a branch has expanded or collapsed).
 */
void ViewerIndexBranchExpandCB(GtkWidget *widget, gpointer data)
{
	GtkCTree *ctree;
	GtkCTreeNode *branch;
	GtkCTreeRow *branch_row;
	viewer_struct *v = VIEWER(data);
	if(v == NULL)
	    return;

	ctree = GTK_CTREE(v->index_ctree);
	branch = v->last_selected_index_node;
	if(branch == NULL)
	    return;

	branch_row = GTK_CTREE_ROW(branch);
	if(branch_row == NULL)
	    return;

	if(branch_row->is_leaf)
	    return;

	if(branch_row->expanded)
	    gtk_ctree_collapse(ctree, branch);
	else
	    gtk_ctree_expand(ctree, branch);
}


/*
 *	ToolBarItem "activate" signal callback.
 */
void ViewerToolBarItemActivateCB(
	ToolBarItem *item, gint id, gpointer data
)
{
	GtkWidget *widget;
	viewer_struct *v = VIEWER(data);
	if((item == NULL) || (v == NULL))
	    return;

	if(v->freeze_count > 0)
	    return;

	widget = ToolBarItemGetWidget(item);

	switch((viewer_opid)id)
	{
	  case VIEWER_OPID_OPEN:
	    ViewerOpenCB(widget, data);
	    break;

	  case VIEWER_OPID_CLEAR:
	    ViewerClearCB(widget, data);
	    break;

	  case VIEWER_OPID_STOP:
	    ViewerStopCB(widget, data);
	    break;
	}
}


/*
 *	Open file callback.
 */
void ViewerOpenCB(GtkWidget *widget, gpointer data)
{
	GtkWidget *toplevel;
	GtkText *text;
	me_core_struct *core;
	viewer_struct *v = VIEWER(data);
	if(v == NULL)
	    return;

	if(v->freeze_count > 0)
	    return;

	v->freeze_count++;

	toplevel = v->toplevel;
	text = GTK_TEXT(v->view_text);
	core = v->core;

	if(TRUE)
	{
	    gboolean response;
	    gchar **paths_list;
	    gint	npaths = 0,
			nftypes = 0;
	    fb_type_struct	**ftypes_list = NULL,
				*ftype_rtn = NULL;

	    /* Query the user to open a manual page */
	    FileBrowserTypeListNew(
		&ftypes_list, &nftypes,
		MANPAGE_SUFFIX,
		"Manual page"
	    );
	    FileBrowserTypeListNew(
		&ftypes_list, &nftypes,
		".mpt",
		"Manual page template"
	    );
	    FileBrowserTypeListNew(
		&ftypes_list, &nftypes,
		"*.*", "All files"
	    );

	    FileBrowserSetTransientFor(toplevel);
	    response = FileBrowserGetResponse(
		"Open Manual Page",
		"Open", "Cancel",
		v->last_open_path,
		ftypes_list, nftypes,
		&paths_list, &npaths,
		&ftype_rtn
	    );
	    FileBrowserSetTransientFor(NULL);

	    if(response && (npaths > 0))
	    {
		gchar *path = STRDUP(paths_list[npaths - 1]);
		if(path != NULL)
		{
		    gint status;
		    gchar *msg;

		    ViewerSetBusy(v, TRUE);
/* Reset scroll positions when user is calling open callback */
		    v->last_scroll_hpos = 0.0f;
		    v->last_scroll_vpos = 0.0f;
/*		    ViewerViewTextRecordScrollPositions(v); */
		    ViewerTextDelete(v, 0, -1);
		    ViewerTextInsertPosition(v, 0);
		    status = ViewerOpenFile(
			v,
			path,
			path
		    );
		    ViewerSetBusy(v, FALSE);
		    switch(status)
		    {
		      case 0:			/* Success */
			break;

		      case -2:			/* No such file */
			msg = g_strdup_printf(
"Unable to find the manual page file:\n\
\n\
    %s",
			    path
			);
			CDialogSetTransientFor(toplevel);
			CDialogGetResponse(
			    "Open Manual Page Failed",
			    msg,
			    NULL,
			    CDIALOG_ICON_WARNING,
			    CDIALOG_BTNFLAG_OK,
			    CDIALOG_BTNFLAG_OK
			);
			g_free(msg);
			CDialogSetTransientFor(NULL);
			break;

		      default:			/* Other error */
			CDialogSetTransientFor(toplevel);
			CDialogGetResponse(
			    "Open Manual Page Failed",
"Error occured while opening manual page.",
			    NULL,
			    CDIALOG_ICON_ERROR,
			    CDIALOG_BTNFLAG_OK,
			    CDIALOG_BTNFLAG_OK
			);
			CDialogSetTransientFor(NULL);
			break;
		    }

		    g_free(path);
		}
	    }

	    /* Delete the file types list */
	    FileBrowserDeleteTypeList(ftypes_list, nftypes);
	}


	/* Update menus, note that a successful load may have updated
	 * menus already but it is safe to do it again here
	 */
	ViewerUpdateMenus(v);

	v->freeze_count--;
}


/*
 *	Clear callback.
 */
void ViewerClearCB(GtkWidget *widget, gpointer data)
{
	GtkCombo *combo;
	GtkCTreeNode *branch, *prev_branch;
	GtkCTreeRow *branch_row;
	GtkCTree *ctree;
	viewer_struct *v = VIEWER(data);
	if(v == NULL)
	    return;

	if(v->freeze_count > 0)
	    return;

	v->freeze_count++;
	ViewerSetBusy(v, TRUE);

	ViewerSetStatusMessage(v, "Clearing contents...");

	switch(v->current_page)
	{
	  case VIEWER_PAGE_MANPAGE:
	    g_free(v->cur_manpage_path);
	    v->cur_manpage_path = NULL;
	    g_free(v->cur_manpage_name);
	    v->cur_manpage_name = NULL;
	    combo = GTK_COMBO(v->manpage_combo);
	    gtk_entry_set_text(GTK_ENTRY(combo->entry), "");
	    ViewerTextDelete(v, 0, -1);
	    ViewerUpdateTitle(v);
	    ViewerUpdateMenus(v);
	    break;

	  case VIEWER_PAGE_INDEX:
	    ctree = GTK_CTREE(v->index_ctree);

	    /* Remove all branches on index ctree */
	    gtk_clist_freeze((GtkCList *)ctree);
	    branch = gtk_ctree_node_nth(ctree, 0);
	    while(branch != NULL)
	    {
		prev_branch = branch;	/* Record current branch */

		/* Get branch row data */
		branch_row = GTK_CTREE_ROW(branch);
		/* Get next sibling branch */
		branch = ((branch_row == NULL) ?
		    NULL : branch_row->sibling
		);

		/* Remove current branch */
		gtk_ctree_remove_node(ctree, prev_branch);
	    }
	    gtk_clist_thaw((GtkCList *)ctree);

	    ViewerUpdateMenus(v);
	    break;
	}
	ViewerSetStatusMessage(v, "Clear done");

	v->freeze_count--;
	ViewerSetBusy(v, FALSE);
}

/*
 *	Stop callback.
 */
void ViewerStopCB(GtkWidget *widget, gpointer data)
{
	viewer_struct *v = VIEWER(data);
	if(v == NULL)
	    return;

	v->stop_count++;
	if(v->stop_count < 1)
	    v->stop_count = 1;
}

/*
 *	Viewer index ctree, goto parent item callback.
 */
void ViewerIndexGotoParentCB(GtkWidget *widget, gpointer data)
{
	GtkCTree *ctree;
	GtkCTreeNode *branch;
	GtkCTreeRow *branch_row;
	viewer_struct *v = VIEWER(data);
	if(v == NULL)
	    return;

	if(v->freeze_count > 0)
	    return;

	ctree = GTK_CTREE(v->index_ctree);

	/* Get the selected branch */
	branch = v->last_selected_index_node;
	if(branch == NULL)
	    return;

	branch_row = GTK_CTREE_ROW(branch);

	branch = branch_row->parent;
	if(branch != NULL)
	    gtk_ctree_select(ctree, branch);
}

/*
 *	Goto selected index ctree branch callback.
 */
void ViewerIndexViewCB(GtkWidget *widget, gpointer data)
{
	GtkCTree *ctree;
	GtkCTreeNode *branch;
	ViewerIndexItem *item;
	viewer_struct *v = VIEWER(data);
	if(v == NULL)
	    return;

	if(v->freeze_count > 0)
	    return;

	ctree = GTK_CTREE(v->index_ctree);

	/* Get the selected branch */
	branch = v->last_selected_index_node;
	if(branch == NULL)
	{
	    v->freeze_count--;
	    return;
	}

	item = VIEWER_INDEX_ITEM(gtk_ctree_node_get_row_data(
	    ctree,
	    branch
	));
	if(item == NULL)
	{
	    v->freeze_count--;
	    return;
	}

	if(item->type == VIEWER_INDEX_ITEM_TYPE_MANPAGE)
	{
	    if(item->path != NULL)
	    {
		gchar	*path = g_strdup(item->path),
			*name = STRDUP(g_basename(path));

		ViewerSetBusy(v, TRUE);

		ViewerViewTextRecordScrollPositions(v);

		ViewerTextDelete(v, 0, -1);
		ViewerTextInsertPosition(v, 0);

		/* Open the manual page and switch to the view page,
		 * do not freeze or else the page will not change
		 */
		ViewerOpenFile(
		    v,
		    path,
		    name
		);

		g_free(path);
		g_free(name);

		ViewerSetBusy(v, FALSE);
	    }
	}
}


/*
 *	Index Refresh.
 */
void ViewerIndexRefreshCB(GtkWidget *widget, gpointer data)
{
	viewer_struct *v = VIEWER(data);
	if(v == NULL)
	    return;

	if(v->freeze_count > 0)
	    return;

	ViewerIndexRefresh(v);
}


/*
 *	Copy.
 */
void ViewerCopyCB(GtkWidget *widget, gpointer data)
{
	GtkEditable *editable;
	GtkText *text;
	viewer_struct *v = VIEWER(data);
	if((widget == NULL) || (v == NULL))
	    return;

	text = GTK_TEXT(v->view_text);
	editable = GTK_EDITABLE(text);
	gtk_text_freeze(text);
	gtk_editable_copy_clipboard(editable);
	gtk_text_thaw(text);

	ViewerUpdateMenus(v);
	ViewerSetStatusMessage(v, "Selection coppied");
}

/*
 *	Select All.
 */
void ViewerSelectAllCB(GtkWidget *widget, gpointer data)
{
	GtkEditable *editable;
	GtkText *text;
	viewer_struct *v = VIEWER(data);
	if((widget == NULL) || (v == NULL))
	    return;

	text = GTK_TEXT(v->view_text);
	editable = GTK_EDITABLE(text);
	gtk_text_freeze(text);
	gtk_editable_select_region(editable, 0, -1);
	gtk_text_thaw(text);
}

/*
 *	Unselect All.
 */
void ViewerUnselectAllCB(GtkWidget *widget, gpointer data)
{
	GtkEditable *editable;
	GtkText *text;
	viewer_struct *v = VIEWER(data);
	if((widget == NULL) || (v == NULL))
	    return;

	text = GTK_TEXT(v->view_text);
	editable = GTK_EDITABLE(text);
	gtk_text_freeze(text);
	gtk_editable_select_region(editable, -1, -1);
	gtk_text_thaw(text);
}


/*
 *	Options.
 */
void ViewerOptionsCB(GtkWidget *widget, gpointer data)
{
	GtkWidget	*toplevel,
			*cfg_win;
	me_core_struct *core;
	viewer_struct *v = VIEWER(data);
	if((widget == NULL) || (v == NULL))
	    return;

	toplevel = v->toplevel;
	core = v->core;

	cfg_win = core->options_window;
	if(cfg_win == NULL)
	    core->options_window = cfg_win = MEOptionsWinNew(core);

	if(GTK_WIDGET_MAPPED(cfg_win))
	{
	    gtk_widget_hide(cfg_win);
	    gtk_widget_show(cfg_win);
	    return;
	}

	CfgWinGetValues(cfg_win, FALSE);
	CfgWinSetHasChanges(cfg_win, FALSE);
	MEOptionsWinSelectPage(
	    cfg_win,
	    "Viewer"
	);
	CfgWinUpdateMenus(cfg_win);

	if(cfg_win->window != NULL)
	{
	    const gint  x_offset = 0,
			y_offset = 0;
	    gint        x, y, 
			width1, height1,
			width2, height2;
	    GdkWindow   *window1 = toplevel->window, 
			*window2 = cfg_win->window; 

	    gdk_window_get_root_origin(window1, &x, &y);
	    gdk_window_get_size(window1, &width1, &height1);
	    gdk_window_get_size(window2, &width2, &height2);
	    gtk_widget_set_uposition(
		cfg_win,
		((width2 > 0) ?
		    (x + (width1 / 2) - (width2 / 2)) : (x + 20)
		) + x_offset,
		((height2 > 0) ?
		    (y + (height1 / 2) - (height2 / 2)) : (y + 20)
		) + y_offset
	    );
	}

	gtk_widget_show(cfg_win);
}


/*
 *	Right-click menu mapping.
 */
gint ViewerMenuMapCB(GtkWidget *widget, GdkEvent *event, gpointer data)
{
	GdkEventButton *button;
	viewer_struct *v = VIEWER(data);
	if((v == NULL) || (event == NULL) || (widget == NULL))
	    return(TRUE);

	if(event->type == GDK_BUTTON_PRESS)
	    button = (GdkEventButton *)event;
	else
	    button = NULL;

	/* View GtkText */
	if(widget == v->view_text)
	{

	}
	/* Index GtkCTree */
	else if(widget == v->index_ctree)
	{
	    gint row, column;   
	    GtkCTree *ctree = GTK_CTREE(widget);
	    GtkCList *clist = GTK_CLIST(ctree);
	    if(button != NULL)
	    {
		switch(button->button)
		{
		  case GDK_BUTTON3:
		    if(gtk_clist_get_selection_info(
			clist,
			(gint)button->x, (gint)button->y,
			&row, &column
		    ))
		    {
			GtkCTreeNode *node = gtk_ctree_node_nth(
			    ctree,
			    row
			);
			if((node != NULL) &&
			   (node != v->last_selected_index_node)
			)
			{
			    gtk_clist_freeze(clist);
			    gtk_clist_unselect_all(clist);
			    gtk_ctree_select(ctree, node);
			    gtk_clist_thaw(clist);
			}
		    }

		    /* Map the Index Right Click GtkMenu */
		    gtk_menu_popup(
			GTK_MENU(v->index_menu),
			NULL, NULL,
			NULL, NULL,
			button->button, button->time
		    );
		    break;
		}
	    }

	    /* Double click? */
	    if(event->type == GDK_2BUTTON_PRESS)
	    {
		ViewerIndexViewCB(widget, data);
	    }
	}

	return(TRUE);
}

/*
 *	Manpage GtkEntry "activate" signal callback.
 */
void ViewerManPageActivateCB(GtkWidget *widget, gpointer data)
{
	gint section_num;
	gchar *s, *name, *search_string;
	GtkWidget *toplevel;
	GtkCombo *combo;
	viewer_struct *v = VIEWER(data);
	if(v == NULL)
	    return;

	if(v->freeze_count > 0)
	    return;

	v->freeze_count++;

	toplevel = v->toplevel;
	section_num = ViewerCurrentSection(v);
	combo = GTK_COMBO(v->manpage_combo);
	search_string = STRDUP(gtk_entry_get_text(
	    GTK_ENTRY(combo->entry)
	));

	/* Get the name (which is just the search string) */
	name = STRDUP(search_string);
	if(STRISEMPTY(search_string))
	{
	    g_free(search_string);
	    g_free(name);
	    v->freeze_count--;
	    return;
	}

	/* Add the section to the beginning of the search string */
	switch(section_num)
	{
	  case -1:				/* Exact */
	    /* For exact matches, make sure search_string starts with
	     * a full path
	     */
	    if(!g_path_is_absolute(search_string))
	    {
		CDialogSetTransientFor(toplevel);
		CDialogGetResponse(
"Invalid Path",
"When specifying an exact manual page, the path must be\n\
an absolute path to the manual page file.",
		    NULL,
		    CDIALOG_ICON_WARNING,
		    CDIALOG_BTNFLAG_OK,
		    CDIALOG_BTNFLAG_OK
		);
		CDialogSetTransientFor(NULL);
		g_free(search_string);
		g_free(name);
		v->freeze_count--;
		return;
	    }
	    break;

	  case -2:				/* Any */
	    break;

	  default:
	    /* Prepend the section to the search string */
	    s = g_strdup_printf(
		"%i %s",
		section_num,
		search_string
	    );
	    g_free(search_string);
	    search_string = s;
	    break;
	}

	/* Call load file procedure, this will run the conversion
	 * program with the given search string and output any
	 * matched results
	 */
	ViewerSetBusy(v, TRUE);
	ViewerViewTextRecordScrollPositions(v);
	ViewerTextDelete(v, 0, -1);
	ViewerTextInsertPosition(v, 0);
	ViewerOpenFile(v, search_string, name);
	ViewerSetBusy(v, FALSE);

	g_free(search_string);
	g_free(name);

	v->freeze_count--;
}

/*
 *	Procedure to find string in the currently displayed manual
 *	page on the viewer page.
 */
static void ViewerFind(
	viewer_struct *v,
	GtkCombo *find_combo,
	const gboolean case_sensitive
)
{
	gboolean got_match, search_wrapped;
	gint start_pos;
	gchar *haystack = NULL, *needle = NULL;
	GtkWidget *toplevel;
	GtkEditable *editable;
	GtkText *text; 

	if((v == NULL) || (find_combo == NULL))
	    return;

	toplevel = v->toplevel;
	text = GTK_TEXT(v->view_text);
	if(text == NULL)
	    return;

	editable = GTK_EDITABLE(text);

	/* Get needle buffer (coppied) */
	needle = STRDUP(gtk_entry_get_text(
	    GTK_ENTRY(find_combo->entry)
	));

	/* Get haystack from text */
	haystack = gtk_editable_get_chars(editable, 0, -1);

	/* Set starting position plus one */
	start_pos = gtk_editable_get_position(editable) + 1;

	/* Do find */
	got_match = ViewerDoFind(
	    v, text,
	    haystack, needle,
	    (gint)gtk_text_get_length(text), start_pos,
	    case_sensitive,
	    TRUE,                   /* Scroll to matched position */
	    &search_wrapped
	);
	/* Did we get a match? */
	if(got_match)
	{
	    /* Search wrapped? */
	    if(search_wrapped)
	    {
		CDialogSetTransientFor(toplevel);
		CDialogGetResponse(
"Search Wrapped",
"Search wrapped.",
		    NULL,
		    CDIALOG_ICON_INFO,
		    CDIALOG_BTNFLAG_OK,
		    CDIALOG_BTNFLAG_OK
		);
		CDialogSetTransientFor(NULL);
	    }
	}   
	else
	{
	    CDialogSetTransientFor(toplevel);
	    CDialogGetResponse(
"String Not Found",
"The string was not found.",
		NULL,
		CDIALOG_ICON_SEARCH,
		CDIALOG_BTNFLAG_OK,
		CDIALOG_BTNFLAG_OK
	    );
	    CDialogSetTransientFor(NULL);
	}

	g_free(haystack);
	g_free(needle);
}

/*
 *	Find GtkEntry "activate" signal callback.
 */
void ViewerFindActivateCB(GtkWidget *widget, gpointer data)  
{
	viewer_struct *v = VIEWER(data);
	if(v == NULL)
	    return;

	if(v->freeze_count > 0)
	    return;

	v->freeze_count++;
	ViewerSetBusy(v, TRUE);

	ViewerFind(
	    v,
	    GTK_COMBO(v->find_combo),
	    FALSE
	);

	ViewerUpdateMenus(v);

	v->freeze_count--;
	ViewerSetBusy(v, FALSE);
}

/*
 *	Checks if the string given by needle exists in the contents of
 *	the file given by path.
 */
static gboolean ViewerFindInPagesGrep(
	const gchar *path,
	const gchar *expression,
	const gboolean case_sensitive
)
{
	FILE *fp;
	gboolean got_match;
	gint c;
	glong last_line_start_pos;
	const gchar *expression_ptr;

	if((path == NULL) || (expression == NULL))
	    return(FALSE);

	/* Open the file for reading */
	fp = fopen(path, "rb");
	if(fp == NULL)
	    return(FALSE);

	/* Search through the file */
	got_match = FALSE;
	expression_ptr = expression;
	last_line_start_pos = 0l;

	while(TRUE)
	{
	    /* Get the next character */
	    c = fgetc(fp);
	    if(c == EOF)
		break;

	    if(!case_sensitive)
		c = toupper(c);

	    /* If this character marks the end of the line then record
	     * the last line starting position as one character ahead
	     * of the current position
	     */
	    if(ISCR(c))
		last_line_start_pos = ftell(fp);

	    /* This character matches current character of expression?
	     * If so then increment the expression pointer. Otherwise
	     * reset the expression pointer back to the start of the
	     * expression
	     */
	    if((c == *expression_ptr) && (c != '\0'))
		expression_ptr++;
	    else
		expression_ptr = expression;

	    /* Matched entire expression? Test if the expression pointer
	     * has been incremented all the way to the null terminating
	     * character which suggests that all characters in the
	     * expression were matched
	     */
	    if(*expression_ptr == '\0')
	    {
		got_match = TRUE;
		break;
	    }
	}

	/* Close the file */
	fclose(fp);

	return(got_match);
}

/*
 *	Finds the string from the currently selected manual page
 *	on the index ctree.
 */
static void ViewerDoFindInPagesCB(
	viewer_struct *v, GtkCombo *find_combo, gboolean case_sensitive
)
{
	gboolean got_match;
	gint pass, row_num, start_row;
	gchar *needle = NULL;
	GList *glist;
	GtkCList *clist;
	GtkCTree *ctree;
	GtkCTreeNode *end_node;

	if((v == NULL) || (find_combo == NULL))
	    return;

	ctree = GTK_CTREE(v->index_ctree);
	clist = GTK_CLIST(ctree);

	/* Get needle buffer (coppied) */
	needle = STRDUP(gtk_entry_get_text(
	    GTK_ENTRY(find_combo->entry)
	));
	if(needle == NULL)
	    return;

	if(!case_sensitive)
	    strtoupper(needle);


	/* Skip if already processing */
	if(v->processing)
	{
	    g_free(needle);
	    return;
	}

	/* Mark as processing and reset stop count to begin search */
	v->processing = TRUE;
	v->stop_count = 0;
	ViewerSetBusy(v, TRUE);
	ViewerUpdateMenus(v);


	/* Set starting row plus one and the end node */
	start_row = 0;
	end_node = NULL;
	glist = clist->selection_end;
	if((glist != NULL) ? (glist->data != NULL) : FALSE)
	{
	    GList *glist2 = clist->row_list;
	    GtkCTreeRow *ctree_row_ptr = GTK_CTREE_ROW(
		(GtkCTreeNode *)glist->data
	    );
	    GtkCListRow *clist_row_ptr = (GtkCListRow *)ctree_row_ptr;

	    /* Set the end node as the selected node, this will be
	     * encountered on the find second pass farther below
	     */
	    end_node = (GtkCTreeNode *)glist->data;

	    /* Iterate through the row pointers, to find which row pointer
	     * matches the selected row pointer obtained from the selected
	     * tree node. start_row will be incremented to be one past the
	     * selected row index
	     */
	    while(glist2 != NULL)
	    {
		if((gpointer)clist_row_ptr == glist2->data)
		{
		    start_row++;
		    break;
		}

		start_row++;
		glist2 = glist2->next;
	    }
	}


	/* Iterate for two passes */
	got_match = FALSE;
	pass = 0;
	while(pass < 2)
	{
	    GtkCTreeNode *node;
	    GtkCTreeRow *ctree_row_ptr;
	    ViewerIndexItem *item;

	    row_num = MAX(start_row, 0);

	    for(row_num = MAX(start_row, 0);
		row_num < clist->rows;
		row_num++
	    )
	    {
		if(v->stop_count > 0)
		    break;

		node = gtk_ctree_node_nth(ctree, row_num);
		if((node == NULL) || (node == end_node))
		    break;

		ctree_row_ptr = GTK_CTREE_ROW(node);
		if(ctree_row_ptr == NULL)
		    continue;

		/* Expand this node as needed */
		if(!ctree_row_ptr->expanded && !ctree_row_ptr->is_leaf)
		{
		    /* Expand and skip to next node */
		    gtk_ctree_expand(ctree, node);
		    continue;
		}

		/* Get the item */
		item = VIEWER_INDEX_ITEM(gtk_ctree_node_get_row_data(
		    ctree,
		    node
		));
		if(item == NULL)
		    continue;

		ViewerSetStatusMessage(v, item->path);

		/* Check if the string needle exists in the contents of
		 * the file referenced by the item data structure
		 */
		got_match = ViewerFindInPagesGrep(
		    item->path,
		    needle,
		    case_sensitive
		);
		if(got_match)
		{
		    gtk_ctree_select(ctree, node);
		    break;
		}

		/* Update status bar progress activity */
		ViewerSetStatusProgress(v, -1.0);
	    }

	    if(got_match)
		break;

	    /* Reset start row back to the beginning */
	    start_row = 0;

	    /* Go on to second pass */
	    pass++;
	}

	/* Delete coppied search string */
	g_free(needle);
	needle = NULL;

	v->processing = FALSE;
	ViewerSetBusy(v, FALSE);
	ViewerSetStatusProgress(v, 0.0f);
	if(!got_match)
	    ViewerSetStatusMessage(v, "No manual page found");
	ViewerUpdateMenus(v);
}

/*
 *	Find In Pages GtkEntry "activate" signal callback.
 */
void ViewerFindInPagesActivateCB(GtkWidget *widget, gpointer data)
{
	viewer_struct *v = VIEWER(data);
	if(v == NULL)
	    return;

	ViewerDoFindInPagesCB(
	    v,
	    (GtkCombo *)v->index_find_in_pages_combo,
	    FALSE
	);
}


/*
 *	Viewer index ctree select callback.
 */
void ViewerIndexCTreeSelectCB(
	GtkCTree *ctree, GtkCTreeNode *branch, gint column, gpointer data
)
{
	static gboolean reenterant = FALSE;
	gint row;
	GtkCList *clist;
	viewer_struct *v = VIEWER(data);
	if((ctree == NULL) || (v == NULL))
	    return;

	if(reenterant)
	    return;
	else
	    reenterant = TRUE;

	if((gpointer)ctree == (gpointer)v->index_ctree)
	{
	    clist = (GtkCList *)ctree;

	    ViewerSetBusy(v, TRUE);

	    /* A branch selected previously? */
	    if(v->last_selected_index_node != NULL)
	    {
		/* Unselect signal will have the values applied, we don't
		 * need to apply values here.
		 */
	    }

	    /* Update selected index branch */
	    v->last_selected_index_node = branch;

#if 0 
	    /* Set DND drag icon */
	    ViewerDNDSetIcon(
		v,
		gtk_ctree_node_get_row_data(ctree, branch)
	    );
#endif

	    /* Scroll if row not visibile */
	    row = gtk_clist_find_row_from_data(
		clist,
		gtk_ctree_node_get_row_data(ctree, branch)
	    );
	    if(row > -1)
	    {
		if(gtk_clist_row_is_visible(clist, row) !=
		    GTK_VISIBILITY_FULL
		)
		    gtk_clist_moveto(
			clist,
			row, 0,         /* Row, column */
			0.5, 0.0        /* Row, column */
		    );
	    }



/* Fetch values for newly selected branch? */

/* Notify about selection */

	    /* Update menus */
	    ViewerUpdateMenus(v);

	    ViewerSetBusy(v, FALSE);
	}

	reenterant = FALSE;        
}

/*
 *	Viewer index ctree unselect callback.
 */
void ViewerIndexCTreeUnselectCB(
	GtkCTree *ctree, GtkCTreeNode *branch, gint column, gpointer data
)
{
	static gboolean reenterant = FALSE;
	viewer_struct *v = VIEWER(data);
	if((ctree == NULL) || (v == NULL))
	    return;

	if(reenterant)
	    return;
	else
	    reenterant = TRUE;

	if((gpointer)ctree == (gpointer)v->index_ctree)
	{
	    /* Apply values? */

	    if(branch == v->last_selected_index_node)
		v->last_selected_index_node = NULL;

	    /* Clear values? */


	    /* Update menus */
	    ViewerUpdateMenus(v);
	}

	reenterant = FALSE;
}

/*
 *	Viewer index GtkCTree branch expand (and collapse) callback.
 */
void ViewerIndexCTreeExpandCB(
	GtkCTree *ctree, GtkCTreeNode *node, gpointer data
)
{
	static gboolean reenterant = FALSE;
	viewer_struct *v = VIEWER(data);
	if((ctree == NULL) || (v == NULL))
	    return;

	if(reenterant)
	    return;
	else
	    reenterant = TRUE;

	if((gpointer)ctree == (gpointer)v->index_ctree)
	{


	}

	reenterant = FALSE;
}


/*
 *	Reconfigured callback.
 */
void ViewerReconfiguredCB(viewer_struct *v)
{
	CfgList *cfg_list;
	me_core_struct *core;

	if(v == NULL)
	    return;

	core = v->core;
	cfg_list = core->cfg_list;

	/* Update the Tool Bar */
	ToolBarSetButtonDecals(
	    v->tool_bar,
	    (ToolBarButtonDecals)ME_GET_I(ME_CFG_PARM_VIEWER_TOOL_BAR_BUTTON_DECALS)
	);
	ToolBarSetRelief(
	    v->tool_bar,
	    (GtkReliefStyle)ME_GET_I(ME_CFG_PARM_VIEWER_TOOL_BAR_RELIEF)
	);

	/* Update the colors */
	if(v->colormap != NULL)
	{
	    GdkColormap *colormap = v->colormap;
	    GdkColor *ct;
	    CfgColor *cs;

	    ct = &v->view_normal_text_color;
	    cs = ME_GET_COLOR(ME_CFG_PARM_COLOR_MANPAGE_FG);
	    if(cs != NULL)
	    {
		GDK_COLOR_SET_COEFF(ct, cs->r, cs->g, cs->b);
		GDK_COLORMAP_FREE_COLOR(colormap, ct);
		GDK_COLORMAP_ALLOC_COLOR(colormap, ct);
	    }

	    ct = &v->view_bold_text_color;
	    cs = ME_GET_COLOR(ME_CFG_PARM_COLOR_MANPAGE_BOLD);
	    if(cs != NULL)
	    {
		GDK_COLOR_SET_COEFF(ct, cs->r, cs->g, cs->b);
		GDK_COLORMAP_FREE_COLOR(colormap, ct);
		GDK_COLORMAP_ALLOC_COLOR(colormap, ct);
	    }

	    ct = &v->view_underline_text_color;
	    cs = ME_GET_COLOR(ME_CFG_PARM_COLOR_MANPAGE_UNDERLINE);
	    if(cs != NULL)
	    {
		GDK_COLOR_SET_COEFF(ct, cs->r, cs->g, cs->b);
		GDK_COLORMAP_FREE_COLOR(colormap, ct);
		GDK_COLORMAP_ALLOC_COLOR(colormap, ct);
	    }
	}

	/* Update the View GtkText background */
	gtk_widget_modify_style(
	    v->view_text,
	    core->manpage_rcstyle
	);
}
