#include <stdio.h>
#include <string.h>
#include <gtk/gtk.h>

#include "url.h"

#include "guiutils.h"
#include "cdialog.h"

#include "editor.h"
#include "editor_dnd.h"
#include "editor_preview.h"
#include "viewer.h"
#include "viewer_fio.h"
#include "viewer_dnd.h"

#include "manedit.h"
#include "config.h"


void ViewerIndexCTreeDNDDataRequestCB(
	GtkWidget *widget, GdkDragContext *dc,
	GtkSelectionData *selection_data, guint info,
	guint t,
	gpointer data
);
void ViewerViewTextDNDDataRecievedCB(
	GtkWidget *widget,
	GdkDragContext *dc,
	gint x, gint y,
	GtkSelectionData *selection_data,
	guint info, guint t,
	gpointer data
);
void ViewerDNDDataDeleteCB(
	GtkWidget *widget, GdkDragContext *dc, gpointer data
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)


/*
 *	Index GtkCTree "drag_data_request" signal callback.
 */
void ViewerIndexCTreeDNDDataRequestCB(
	GtkWidget *widget, GdkDragContext *dc,
	GtkSelectionData *selection_data, guint info,
	guint t,
	gpointer data
)
{
	gint buf_len;
	guint8 *buf;
	GList *url_list;
	GtkCTree *ctree;
	GtkCTreeNode *node;
	ViewerIndexItem *item;
	viewer_struct *v = VIEWER(data);
	if((widget == NULL) || (v == NULL) || (dc == NULL))
	    return;

	ctree = GTK_CTREE(v->index_ctree);
	node = v->last_selected_index_node;
	if(node == NULL)
	    return;

	item = VIEWER_INDEX_ITEM(gtk_ctree_node_get_row_data(
	    ctree,
	    node
	));
	if(item == NULL)
	    return;

	if(item->path == NULL)
	    return;

	url_list = NULL;
	url_list = g_list_append(
	    url_list,
	    URLNewWithValues(
		0,
		"file",
		NULL,
		NULL,
		NULL,
		0,
		item->path,
		NULL
	    )
	);
	buf = URLEncode(url_list, &buf_len);
	g_list_foreach(
	    url_list,
	    (GFunc)URLDelete,
	    NULL
	);
	g_list_free(url_list);

	/* Send out the data */
	gtk_selection_data_set(
	    selection_data,
	    GDK_SELECTION_TYPE_STRING,
	    8,					/* 8 bits per character */
	    buf, buf_len
	);

	g_free(buf);
}


/*
 *	View GtkText "drag_data_received" signal callback.
 */
void ViewerViewTextDNDDataRecievedCB(
	GtkWidget *widget, GdkDragContext *dc,
	gint x, gint y,
	GtkSelectionData *selection_data, guint info,
	guint t,
	gpointer data
)
{
	gboolean need_delete = FALSE;
	gint viewer_num;
	me_core_struct *core;
	viewer_struct *v = VIEWER(data);
	if((widget == NULL) || (dc == NULL) || (v == NULL))
	    return;

	if(selection_data == NULL)
	    return;
	if(selection_data->length < 0)
	    return;

	core = v->core;

	/* Get this viewer's index */
	for(viewer_num = 0; viewer_num < core->total_viewers; viewer_num++)
	{
	    if(v == core->viewer[viewer_num])
		break;
	}
	if(viewer_num >= core->total_viewers)
	    viewer_num = -1;

	/* Editor Branch */
	if(info == MEDIT_DND_TYPE_INFO_EDITOR_BRANCH_CMD)
	{
	    gboolean	is_branch_file = FALSE,
			is_branch_header = FALSE,
			is_branch_section = FALSE;
	    GList *nodes_list;
	    editor_struct *src_editor = NULL;

	    /* Parse the data */
	    nodes_list = EditorDNDParseCmd(
		(const gchar *)selection_data->data,
		&src_editor
	    );
	    if(src_editor != NULL)
	    {
		GList *glist;
		editor_item_struct *item;

		/* Check the item type */
		for(glist = nodes_list;
		    glist != NULL;
		    glist = g_list_next(glist)
		)
		{
		    item = EditorBranchGetData(
			GTK_CTREE(src_editor->layout_ctree),
			(GtkCTreeNode *)glist->data
		    );
		    if(item == NULL)
			continue;

		    switch(item->type)
		    {
		      case EDITOR_ITEM_FILE:
			is_branch_file = TRUE;
			break;
		
		      case EDITOR_ITEM_HEADER:
			is_branch_header = TRUE;
			break;

		      case EDITOR_ITEM_SECTION:  
			is_branch_section = TRUE;
			break;
		    }
		}

		/* Handle by the branch item type */
		if(is_branch_file)
		{
		    /* Unset delete, we don't want to delete anything
		     * for this operation.
		     */
		    need_delete = FALSE;

		    /* Handle only the first branch */
		    if(nodes_list != NULL)
		    {
			/* Call editor's preview procedure on source
			 * editor so that it first saves the loaded
			 * data before loading it into the viewer.
			 */
			EditorPreview(
			    src_editor,
			    (GtkCTreeNode *)nodes_list->data,
			    viewer_num
			);
		    }
		}
		/* Header or section? */
		else if(is_branch_header || is_branch_section)
		{
		    /* Unset delete, we don't want to delete anything
		     * for this operation.
		     */
		    need_delete = FALSE;

		    /* Handle only the first branch */
		    if(nodes_list != NULL)
		    {
			/* Get the source node's toplevel node */
			GtkCTreeNode *src_toplevel_node = EditorGetToplevelNode(
			    (GtkCTreeNode *)nodes_list->data
			);

			/* Call editor's preview procedure on source
			 * editor so that it first saves the loaded
			 * data before loading it into the viewer.
			 */
			EditorPreview(
			    src_editor,
			    src_toplevel_node,
			    viewer_num
			);
		    }
		}
	    }

	    /* Delete the nodes list */
	    g_list_free(nodes_list);
	}
	/* text/plain, text/uri-list, and STRING */
	else if((info == MEDIT_DND_TYPE_INFO_TEXT_PLAIN) ||
		(info == MEDIT_DND_TYPE_INFO_TEXT_URI_LIST) ||
		(info == MEDIT_DND_TYPE_INFO_STRING)
	)
	{
	    GList *url_list = URLDecode(
		selection_data->data, selection_data->length
	    );
	    if(url_list != NULL)
	    {
		GList *glist;
		url_struct *url;

		for(glist = url_list;
		    glist != NULL;
		    glist = g_list_next(glist)
		)
		{
		    url = URL(glist->data);
		    if(url == NULL)
			continue;

		    if((url->protocol == NULL) || (url->path == NULL))
			continue;

		    if(g_strcasecmp(url->protocol, "file"))
			continue;

		    /* Open the manual page */
		    ViewerSetBusy(v, TRUE);
		    ViewerViewTextRecordScrollPositions(v);
		    ViewerTextDelete(v, 0, -1);
		    ViewerTextInsertPosition(v, 0);
		    ViewerOpenFile(
			v,
			url->path,
			url->path
		    );
		    ViewerSetBusy(v, FALSE);

		    /* Only load one */
		    break;
		}

		g_list_foreach(
		    url_list,
		    (GFunc)URLDelete,
		    NULL
		);
		g_list_free(url_list);
	    }
	}
}

/*
 *      Viewer DND data delete callback.
 *
 *      This deletes the selected item on viewer's layout ctree.
 */
void ViewerDNDDataDeleteCB(
	GtkWidget *widget, GdkDragContext *dc, gpointer data
)
{
	return;
}
